
module Make

# 55 "asllib/Parser.mly"
          (Config : ParserConfig.CONFIG)
# 7 "asllib/Parser.ml"

= struct
  
  module MenhirBasics = struct
    
    exception Error
    
    let _eRR =
      fun _s ->
        raise Error
    
    type token = Tokens.token
    
  end
  
  include MenhirBasics
  
# 63 "asllib/Parser.mly"
  

open AST
open ASTUtils
open Desugar

let version = V1

let t_bit =
  T_Bits (E_Literal (L_Int Z.one) |> add_dummy_annotation ~version, [])

let zero = E_Literal (L_Int Z.zero) |> add_dummy_annotation ~version

let make_ldi_vars (xs, ty) =
  let make_one x =
    S_Decl (LDK_Var, LDI_Var x, Some ty, None)
    |> add_dummy_annotation ~version
  in
  List.map make_one xs |> stmt_from_list |> desc

let make_ty_decl_subtype (x, s) =
  let name, _fields = s.desc in
  let ty = ASTUtils.add_pos_from s (T_Named name) in
  D_TypeDecl (x, ty, Some s.desc)

let prec =
  let open AST in
  function
  | `BOR | `BAND | `IMPL | `BEQ -> 1
  | `EQ_OP | `NEQ -> 2
  | `PLUS | `MINUS | `OR | `XOR | `AND | `BV_CONCAT -> 3
  | `MUL | `DIV | `DIVRM | `RDIV | `MOD | `SHL | `SHR -> 4
  | `POW -> 5
  | `GT | `GEQ | `LT | `LEQ -> 0 (* Non assoc *)

let check_not_same_prec loc op op' =
  if prec op = prec op' then Error.(fatal_from loc CannotParse)

let check_not_binop_same_prec op e =
  match e.desc with
  | E_Binop (op', _, _) when op != op' -> check_not_same_prec e op op'
  | _ -> ()

let e_binop (e1, op, e2) =
  let () = check_not_binop_same_prec op e1
  and () = check_not_binop_same_prec op e2
  in
  E_Binop (op, e1, e2)

let e_call call = E_Call { call with call_type = ST_Function }
let s_call call = S_Call { call with call_type = ST_Procedure }

let le_var x = LE_Var x.desc |> add_pos_from x
let lhs_access ~base ~index ~fields = { base; index; fields; slices=add_dummy_annotation ~version [] }


# 82 "asllib/Parser.ml"
  
  type ('s, 'r) _menhir_state = 
    | MenhirState000 : ('s, _menhir_box_opn) _menhir_state
      (** State 000.
        Stack shape : .
        Start symbol: opn. *)

    | MenhirState001 : (('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_state
      (** State 001.
        Stack shape : WHILE.
        Start symbol: <undetermined>. *)

    | MenhirState004 : (('s, 'r) _menhir_cell1_NOT, 'r) _menhir_state
      (** State 004.
        Stack shape : NOT.
        Start symbol: <undetermined>. *)

    | MenhirState005 : (('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 005.
        Stack shape : MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState006 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 006.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState008 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 008.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState009 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 009.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState010 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 010.
        Stack shape : IDENTIFIER LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState012 : (('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 012.
        Stack shape : BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState015 : (('s, 'r) _menhir_cell1_ARBITRARY, 'r) _menhir_state
      (** State 015.
        Stack shape : ARBITRARY.
        Start symbol: <undetermined>. *)

    | MenhirState018 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 018.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState019 : (('s, 'r) _menhir_cell1_INTEGER, 'r) _menhir_state
      (** State 019.
        Stack shape : INTEGER.
        Start symbol: <undetermined>. *)

    | MenhirState020 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 020.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState021 : ((('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 021.
        Stack shape : LBRACE MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState024 : ((('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 024.
        Stack shape : MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState025 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 025.
        Stack shape : expr LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState026 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 026.
        Stack shape : expr LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState027 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_XOR, 'r) _menhir_state
      (** State 027.
        Stack shape : expr XOR.
        Start symbol: <undetermined>. *)

    | MenhirState028 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_XOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 028.
        Stack shape : expr XOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState029 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHR, 'r) _menhir_state
      (** State 029.
        Stack shape : expr SHR.
        Start symbol: <undetermined>. *)

    | MenhirState030 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 030.
        Stack shape : expr SHR expr.
        Start symbol: <undetermined>. *)

    | MenhirState031 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_POW, 'r) _menhir_state
      (** State 031.
        Stack shape : expr POW.
        Start symbol: <undetermined>. *)

    | MenhirState032 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_POW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 032.
        Stack shape : expr POW expr.
        Start symbol: <undetermined>. *)

    | MenhirState033 : (('s, 'r) _menhir_cell1_LBRACKET, 'r) _menhir_state
      (** State 033.
        Stack shape : LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState034 : (('s, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 034.
        Stack shape : COLON.
        Start symbol: <undetermined>. *)

    | MenhirState035 : ((('s, 'r) _menhir_cell1_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 035.
        Stack shape : COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState036 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHL, 'r) _menhir_state
      (** State 036.
        Stack shape : expr SHL.
        Start symbol: <undetermined>. *)

    | MenhirState037 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SHL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 037.
        Stack shape : expr SHL expr.
        Start symbol: <undetermined>. *)

    | MenhirState038 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IN, 'r) _menhir_state
      (** State 038.
        Stack shape : expr IN.
        Start symbol: <undetermined>. *)

    | MenhirState039 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 039.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState040 : (('s, 'r) _menhir_cell1_NOT, 'r) _menhir_state
      (** State 040.
        Stack shape : NOT.
        Start symbol: <undetermined>. *)

    | MenhirState041 : ((('s, 'r) _menhir_cell1_NOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 041.
        Stack shape : NOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState043 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 043.
        Stack shape : expr DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState045 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 045.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState052 : (('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 052.
        Stack shape : MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState053 : ((('s, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 053.
        Stack shape : MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState055 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 055.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState056 : (('s, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 056.
        Stack shape : LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState057 : ((('s, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 057.
        Stack shape : LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState058 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RDIV, 'r) _menhir_state
      (** State 058.
        Stack shape : expr RDIV.
        Start symbol: <undetermined>. *)

    | MenhirState059 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RDIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 059.
        Stack shape : expr RDIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState060 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS, 'r) _menhir_state
      (** State 060.
        Stack shape : expr PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState061 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 061.
        Stack shape : expr PLUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState062 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MUL, 'r) _menhir_state
      (** State 062.
        Stack shape : expr MUL.
        Start symbol: <undetermined>. *)

    | MenhirState063 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MUL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 063.
        Stack shape : expr MUL expr.
        Start symbol: <undetermined>. *)

    | MenhirState064 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MOD, 'r) _menhir_state
      (** State 064.
        Stack shape : expr MOD.
        Start symbol: <undetermined>. *)

    | MenhirState065 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MOD, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 065.
        Stack shape : expr MOD expr.
        Start symbol: <undetermined>. *)

    | MenhirState066 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIVRM, 'r) _menhir_state
      (** State 066.
        Stack shape : expr DIVRM.
        Start symbol: <undetermined>. *)

    | MenhirState067 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIVRM, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 067.
        Stack shape : expr DIVRM expr.
        Start symbol: <undetermined>. *)

    | MenhirState068 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIV, 'r) _menhir_state
      (** State 068.
        Stack shape : expr DIV.
        Start symbol: <undetermined>. *)

    | MenhirState069 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_DIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 069.
        Stack shape : expr DIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState070 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OR, 'r) _menhir_state
      (** State 070.
        Stack shape : expr OR.
        Start symbol: <undetermined>. *)

    | MenhirState071 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 071.
        Stack shape : expr OR expr.
        Start symbol: <undetermined>. *)

    | MenhirState072 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_NEQ, 'r) _menhir_state
      (** State 072.
        Stack shape : expr NEQ.
        Start symbol: <undetermined>. *)

    | MenhirState074 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_NEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 074.
        Stack shape : expr NEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState075 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 075.
        Stack shape : expr MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState076 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 076.
        Stack shape : expr MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState077 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LT, 'r) _menhir_state
      (** State 077.
        Stack shape : expr LT.
        Start symbol: <undetermined>. *)

    | MenhirState078 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 078.
        Stack shape : expr LT expr.
        Start symbol: <undetermined>. *)

    | MenhirState079 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_state
      (** State 079.
        Stack shape : expr COLON_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState080 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 080.
        Stack shape : expr COLON_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState081 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AND, 'r) _menhir_state
      (** State 081.
        Stack shape : expr AND.
        Start symbol: <undetermined>. *)

    | MenhirState082 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 082.
        Stack shape : expr AND expr.
        Start symbol: <undetermined>. *)

    | MenhirState083 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 083.
        Stack shape : expr LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState084 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 084.
        Stack shape : expr LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState085 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GT, 'r) _menhir_state
      (** State 085.
        Stack shape : expr GT.
        Start symbol: <undetermined>. *)

    | MenhirState086 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 086.
        Stack shape : expr GT expr.
        Start symbol: <undetermined>. *)

    | MenhirState087 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 087.
        Stack shape : expr GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState088 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 088.
        Stack shape : expr GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState089 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IMPL, 'r) _menhir_state
      (** State 089.
        Stack shape : expr IMPL.
        Start symbol: <undetermined>. *)

    | MenhirState090 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_IMPL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 090.
        Stack shape : expr IMPL expr.
        Start symbol: <undetermined>. *)

    | MenhirState091 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_state
      (** State 091.
        Stack shape : expr EQ_OP.
        Start symbol: <undetermined>. *)

    | MenhirState093 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 093.
        Stack shape : expr EQ_OP expr.
        Start symbol: <undetermined>. *)

    | MenhirState094 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BOR, 'r) _menhir_state
      (** State 094.
        Stack shape : expr BOR.
        Start symbol: <undetermined>. *)

    | MenhirState095 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 095.
        Stack shape : expr BOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState096 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BEQ, 'r) _menhir_state
      (** State 096.
        Stack shape : expr BEQ.
        Start symbol: <undetermined>. *)

    | MenhirState097 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 097.
        Stack shape : expr BEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState098 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BAND, 'r) _menhir_state
      (** State 098.
        Stack shape : expr BAND.
        Start symbol: <undetermined>. *)

    | MenhirState099 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_BAND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 099.
        Stack shape : expr BAND expr.
        Start symbol: <undetermined>. *)

    | MenhirState100 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_AS, 'r) _menhir_state
      (** State 100.
        Stack shape : expr AS.
        Start symbol: <undetermined>. *)

    | MenhirState104 : (('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 104.
        Stack shape : BITS LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState105 : ((('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 105.
        Stack shape : BITS LPAR expr.
        Start symbol: <undetermined>. *)

    | MenhirState106 : (((('s, 'r) _menhir_cell1_BITS _menhir_cell0_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RPAR, 'r) _menhir_state
      (** State 106.
        Stack shape : BITS LPAR expr RPAR.
        Start symbol: <undetermined>. *)

    | MenhirState107 : (('s, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 107.
        Stack shape : LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState112 : (('s, 'r) _menhir_cell1_slices _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 112.
        Stack shape : slices IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState113 : (('s, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 113.
        Stack shape : COLON.
        Start symbol: <undetermined>. *)

    | MenhirState116 : (('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_state
      (** State 116.
        Stack shape : ARRAY.
        Start symbol: <undetermined>. *)

    | MenhirState117 : ((('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 117.
        Stack shape : ARRAY expr.
        Start symbol: <undetermined>. *)

    | MenhirState119 : (((('s, 'r) _menhir_cell1_ARRAY, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RRBRACKET, 'r) _menhir_state
      (** State 119.
        Stack shape : ARRAY expr RRBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState125 : (('s, 'r) _menhir_cell1_bitfield, 'r) _menhir_state
      (** State 125.
        Stack shape : bitfield.
        Start symbol: <undetermined>. *)

    | MenhirState131 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 131.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState132 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 132.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState133 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 133.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState134 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 134.
        Stack shape : IF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState135 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_state
      (** State 135.
        Stack shape : expr THEN expr ELSIF.
        Start symbol: <undetermined>. *)

    | MenhirState136 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 136.
        Stack shape : expr THEN expr ELSIF expr.
        Start symbol: <undetermined>. *)

    | MenhirState137 : ((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 137.
        Stack shape : expr THEN expr ELSIF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState138 : (((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 138.
        Stack shape : expr THEN expr ELSIF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState139 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSE, 'r) _menhir_state
      (** State 139.
        Stack shape : expr THEN expr ELSE.
        Start symbol: <undetermined>. *)

    | MenhirState140 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ELSE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 140.
        Stack shape : expr THEN expr ELSE expr.
        Start symbol: <undetermined>. *)

    | MenhirState143 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 143.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState144 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 144.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState146 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 146.
        Stack shape : IDENTIFIER LBRACE IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState147 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 147.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState150 : (('s, 'r) _menhir_cell1_field_assign, 'r) _menhir_state
      (** State 150.
        Stack shape : field_assign.
        Start symbol: <undetermined>. *)

    | MenhirState151 : ((('s, 'r) _menhir_cell1_field_assign, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 151.
        Stack shape : field_assign IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState152 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 152.
        Stack shape : IDENTIFIER EQ.
        Start symbol: <undetermined>. *)

    | MenhirState153 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 153.
        Stack shape : IDENTIFIER EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState155 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 155.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState156 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 156.
        Stack shape : expr COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState162 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_clist1_expr_ _menhir_cell0_RBRACE _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 162.
        Stack shape : IDENTIFIER LBRACE clist1(expr) RBRACE LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState168 : (('s, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 168.
        Stack shape : GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState169 : ((('s, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 169.
        Stack shape : GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState170 : (('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 170.
        Stack shape : BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState171 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 171.
        Stack shape : BNOT LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState173 : (('s, 'r) _menhir_cell1_ARBITRARY, 'r) _menhir_state
      (** State 173.
        Stack shape : ARBITRARY.
        Start symbol: <undetermined>. *)

    | MenhirState180 : (('s, 'r) _menhir_cell1_pattern, 'r) _menhir_state
      (** State 180.
        Stack shape : pattern.
        Start symbol: <undetermined>. *)

    | MenhirState181 : (('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_state
      (** State 181.
        Stack shape : expr_pattern.
        Start symbol: <undetermined>. *)

    | MenhirState182 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_XOR, 'r) _menhir_state
      (** State 182.
        Stack shape : expr_pattern XOR.
        Start symbol: <undetermined>. *)

    | MenhirState183 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_XOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 183.
        Stack shape : expr_pattern XOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState184 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SLICING, 'r) _menhir_state
      (** State 184.
        Stack shape : expr_pattern SLICING.
        Start symbol: <undetermined>. *)

    | MenhirState185 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SLICING, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 185.
        Stack shape : expr_pattern SLICING expr.
        Start symbol: <undetermined>. *)

    | MenhirState186 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHR, 'r) _menhir_state
      (** State 186.
        Stack shape : expr_pattern SHR.
        Start symbol: <undetermined>. *)

    | MenhirState187 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 187.
        Stack shape : expr_pattern SHR expr.
        Start symbol: <undetermined>. *)

    | MenhirState188 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHL, 'r) _menhir_state
      (** State 188.
        Stack shape : expr_pattern SHL.
        Start symbol: <undetermined>. *)

    | MenhirState189 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_SHL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 189.
        Stack shape : expr_pattern SHL expr.
        Start symbol: <undetermined>. *)

    | MenhirState190 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_RDIV, 'r) _menhir_state
      (** State 190.
        Stack shape : expr_pattern RDIV.
        Start symbol: <undetermined>. *)

    | MenhirState191 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_RDIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 191.
        Stack shape : expr_pattern RDIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState192 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_POW, 'r) _menhir_state
      (** State 192.
        Stack shape : expr_pattern POW.
        Start symbol: <undetermined>. *)

    | MenhirState193 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_POW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 193.
        Stack shape : expr_pattern POW expr.
        Start symbol: <undetermined>. *)

    | MenhirState194 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_PLUS, 'r) _menhir_state
      (** State 194.
        Stack shape : expr_pattern PLUS.
        Start symbol: <undetermined>. *)

    | MenhirState195 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_PLUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 195.
        Stack shape : expr_pattern PLUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState196 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_OR, 'r) _menhir_state
      (** State 196.
        Stack shape : expr_pattern OR.
        Start symbol: <undetermined>. *)

    | MenhirState197 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_OR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 197.
        Stack shape : expr_pattern OR expr.
        Start symbol: <undetermined>. *)

    | MenhirState198 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_NEQ, 'r) _menhir_state
      (** State 198.
        Stack shape : expr_pattern NEQ.
        Start symbol: <undetermined>. *)

    | MenhirState200 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_NEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 200.
        Stack shape : expr_pattern NEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState201 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MUL, 'r) _menhir_state
      (** State 201.
        Stack shape : expr_pattern MUL.
        Start symbol: <undetermined>. *)

    | MenhirState202 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MUL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 202.
        Stack shape : expr_pattern MUL expr.
        Start symbol: <undetermined>. *)

    | MenhirState203 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MOD, 'r) _menhir_state
      (** State 203.
        Stack shape : expr_pattern MOD.
        Start symbol: <undetermined>. *)

    | MenhirState204 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MOD, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 204.
        Stack shape : expr_pattern MOD expr.
        Start symbol: <undetermined>. *)

    | MenhirState205 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MINUS, 'r) _menhir_state
      (** State 205.
        Stack shape : expr_pattern MINUS.
        Start symbol: <undetermined>. *)

    | MenhirState206 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_MINUS, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 206.
        Stack shape : expr_pattern MINUS expr.
        Start symbol: <undetermined>. *)

    | MenhirState207 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LT, 'r) _menhir_state
      (** State 207.
        Stack shape : expr_pattern LT.
        Start symbol: <undetermined>. *)

    | MenhirState208 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 208.
        Stack shape : expr_pattern LT expr.
        Start symbol: <undetermined>. *)

    | MenhirState209 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 209.
        Stack shape : expr_pattern LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState210 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 210.
        Stack shape : expr_pattern LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState212 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LEQ, 'r) _menhir_state
      (** State 212.
        Stack shape : expr_pattern LEQ.
        Start symbol: <undetermined>. *)

    | MenhirState213 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_LEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 213.
        Stack shape : expr_pattern LEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState214 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IN, 'r) _menhir_state
      (** State 214.
        Stack shape : expr_pattern IN.
        Start symbol: <undetermined>. *)

    | MenhirState215 : ((('s, 'r) _menhir_cell1_IN, 'r) _menhir_cell1_BNOT, 'r) _menhir_state
      (** State 215.
        Stack shape : IN BNOT.
        Start symbol: <undetermined>. *)

    | MenhirState217 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IMPL, 'r) _menhir_state
      (** State 217.
        Stack shape : expr_pattern IMPL.
        Start symbol: <undetermined>. *)

    | MenhirState218 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_IMPL, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 218.
        Stack shape : expr_pattern IMPL expr.
        Start symbol: <undetermined>. *)

    | MenhirState219 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GT, 'r) _menhir_state
      (** State 219.
        Stack shape : expr_pattern GT.
        Start symbol: <undetermined>. *)

    | MenhirState220 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 220.
        Stack shape : expr_pattern GT expr.
        Start symbol: <undetermined>. *)

    | MenhirState221 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GEQ, 'r) _menhir_state
      (** State 221.
        Stack shape : expr_pattern GEQ.
        Start symbol: <undetermined>. *)

    | MenhirState222 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_GEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 222.
        Stack shape : expr_pattern GEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState223 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_state
      (** State 223.
        Stack shape : expr_pattern EQ_OP.
        Start symbol: <undetermined>. *)

    | MenhirState225 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_EQ_OP, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 225.
        Stack shape : expr_pattern EQ_OP expr.
        Start symbol: <undetermined>. *)

    | MenhirState227 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 227.
        Stack shape : expr_pattern DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState231 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIVRM, 'r) _menhir_state
      (** State 231.
        Stack shape : expr_pattern DIVRM.
        Start symbol: <undetermined>. *)

    | MenhirState232 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIVRM, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 232.
        Stack shape : expr_pattern DIVRM expr.
        Start symbol: <undetermined>. *)

    | MenhirState233 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIV, 'r) _menhir_state
      (** State 233.
        Stack shape : expr_pattern DIV.
        Start symbol: <undetermined>. *)

    | MenhirState234 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_DIV, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 234.
        Stack shape : expr_pattern DIV expr.
        Start symbol: <undetermined>. *)

    | MenhirState235 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_state
      (** State 235.
        Stack shape : expr_pattern COLON_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState236 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_COLON_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 236.
        Stack shape : expr_pattern COLON_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState237 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BOR, 'r) _menhir_state
      (** State 237.
        Stack shape : expr_pattern BOR.
        Start symbol: <undetermined>. *)

    | MenhirState238 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BOR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 238.
        Stack shape : expr_pattern BOR expr.
        Start symbol: <undetermined>. *)

    | MenhirState239 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BEQ, 'r) _menhir_state
      (** State 239.
        Stack shape : expr_pattern BEQ.
        Start symbol: <undetermined>. *)

    | MenhirState240 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BEQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 240.
        Stack shape : expr_pattern BEQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState241 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BAND, 'r) _menhir_state
      (** State 241.
        Stack shape : expr_pattern BAND.
        Start symbol: <undetermined>. *)

    | MenhirState242 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_BAND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 242.
        Stack shape : expr_pattern BAND expr.
        Start symbol: <undetermined>. *)

    | MenhirState243 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AS, 'r) _menhir_state
      (** State 243.
        Stack shape : expr_pattern AS.
        Start symbol: <undetermined>. *)

    | MenhirState246 : ((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AND, 'r) _menhir_state
      (** State 246.
        Stack shape : expr_pattern AND.
        Start symbol: <undetermined>. *)

    | MenhirState247 : (((('s, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_cell1_AND, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 247.
        Stack shape : expr_pattern AND expr.
        Start symbol: <undetermined>. *)

    | MenhirState252 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 252.
        Stack shape : BNOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState254 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_pattern, 'r) _menhir_state
      (** State 254.
        Stack shape : LPAR pattern.
        Start symbol: <undetermined>. *)

    | MenhirState256 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr_pattern, 'r) _menhir_state
      (** State 256.
        Stack shape : LPAR expr_pattern.
        Start symbol: <undetermined>. *)

    | MenhirState264 : (('s, 'r) _menhir_cell1_slice, 'r) _menhir_state
      (** State 264.
        Stack shape : slice.
        Start symbol: <undetermined>. *)

    | MenhirState265 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 265.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState266 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_STAR_COLON, 'r) _menhir_state
      (** State 266.
        Stack shape : expr STAR_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState267 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_STAR_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 267.
        Stack shape : expr STAR_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState268 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS_COLON, 'r) _menhir_state
      (** State 268.
        Stack shape : expr PLUS_COLON.
        Start symbol: <undetermined>. *)

    | MenhirState269 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_PLUS_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 269.
        Stack shape : expr PLUS_COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState270 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON, 'r) _menhir_state
      (** State 270.
        Stack shape : expr COLON.
        Start symbol: <undetermined>. *)

    | MenhirState271 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COLON, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 271.
        Stack shape : expr COLON expr.
        Start symbol: <undetermined>. *)

    | MenhirState277 : (('s, 'r) _menhir_cell1_int_constraint, 'r) _menhir_state
      (** State 277.
        Stack shape : int_constraint.
        Start symbol: <undetermined>. *)

    | MenhirState278 : (('s, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 278.
        Stack shape : expr.
        Start symbol: <undetermined>. *)

    | MenhirState279 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SLICING, 'r) _menhir_state
      (** State 279.
        Stack shape : expr SLICING.
        Start symbol: <undetermined>. *)

    | MenhirState280 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_SLICING, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 280.
        Stack shape : expr SLICING expr.
        Start symbol: <undetermined>. *)

    | MenhirState287 : (('s, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 287.
        Stack shape : ty.
        Start symbol: <undetermined>. *)

    | MenhirState293 : ((('s, 'r) _menhir_cell1_BNOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 293.
        Stack shape : BNOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState296 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 296.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState297 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 297.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState298 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 298.
        Stack shape : IF expr THEN expr.
        Start symbol: <undetermined>. *)

    | MenhirState300 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 300.
        Stack shape : LPAR expr.
        Start symbol: <undetermined>. *)

    | MenhirState302 : (((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 302.
        Stack shape : LPAR expr COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState306 : ((('s, 'r) _menhir_cell1_NOT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 306.
        Stack shape : NOT expr.
        Start symbol: <undetermined>. *)

    | MenhirState307 : ((('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 307.
        Stack shape : WHILE expr.
        Start symbol: <undetermined>. *)

    | MenhirState308 : ((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LOOPLIMIT, 'r) _menhir_state
      (** State 308.
        Stack shape : expr LOOPLIMIT.
        Start symbol: <undetermined>. *)

    | MenhirState309 : (((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_LOOPLIMIT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 309.
        Stack shape : expr LOOPLIMIT expr.
        Start symbol: <undetermined>. *)

    | MenhirState311 : (((('s, 'r) _menhir_cell1_WHILE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_loop_limit, 'r) _menhir_state
      (** State 311.
        Stack shape : WHILE expr loop_limit.
        Start symbol: <undetermined>. *)

    | MenhirState312 : (('s, 'r) _menhir_cell1_VAR, 'r) _menhir_state
      (** State 312.
        Stack shape : VAR.
        Start symbol: <undetermined>. *)

    | MenhirState314 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 314.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState317 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_state
      (** State 317.
        Stack shape : LPAR discard_or_identifier.
        Start symbol: <undetermined>. *)

    | MenhirState319 : ((('s, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_cell1_discard_or_identifier, 'r) _menhir_state
      (** State 319.
        Stack shape : discard_or_identifier discard_or_identifier.
        Start symbol: <undetermined>. *)

    | MenhirState325 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 325.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState328 : ((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_state
      (** State 328.
        Stack shape : VAR decl_item.
        Start symbol: <undetermined>. *)

    | MenhirState329 : (((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 329.
        Stack shape : VAR decl_item EQ.
        Start symbol: <undetermined>. *)

    | MenhirState330 : ((('s, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 330.
        Stack shape : EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState333 : (((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_state
      (** State 333.
        Stack shape : VAR decl_item as_ty.
        Start symbol: <undetermined>. *)

    | MenhirState334 : ((((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 334.
        Stack shape : VAR decl_item as_ty EQ.
        Start symbol: <undetermined>. *)

    | MenhirState335 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 335.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState336 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 336.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState338 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_RBRACE _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 338.
        Stack shape : IDENTIFIER LBRACE RBRACE LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState341 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_COMMA, 'r) _menhir_state
      (** State 341.
        Stack shape : IDENTIFIER LBRACE COMMA.
        Start symbol: <undetermined>. *)

    | MenhirState344 : ((((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_cell1_COMMA, 'r) _menhir_cell1_clist1_expr_ _menhir_cell0_RBRACE _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 344.
        Stack shape : IDENTIFIER LBRACE COMMA clist1(expr) RBRACE LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState351 : ((('s, 'r) _menhir_cell1_VAR, 'r) _menhir_cell1_clist2_IDENTIFIER_, 'r) _menhir_state
      (** State 351.
        Stack shape : VAR clist2(IDENTIFIER).
        Start symbol: <undetermined>. *)

    | MenhirState358 : (('s, 'r) _menhir_cell1_TRY, 'r) _menhir_state
      (** State 358.
        Stack shape : TRY.
        Start symbol: <undetermined>. *)

    | MenhirState359 : (('s, 'r) _menhir_cell1_THROW, 'r) _menhir_state
      (** State 359.
        Stack shape : THROW.
        Start symbol: <undetermined>. *)

    | MenhirState361 : ((('s, 'r) _menhir_cell1_THROW, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 361.
        Stack shape : THROW expr.
        Start symbol: <undetermined>. *)

    | MenhirState363 : (('s, 'r) _menhir_cell1_RETURN, 'r) _menhir_state
      (** State 363.
        Stack shape : RETURN.
        Start symbol: <undetermined>. *)

    | MenhirState366 : ((('s, 'r) _menhir_cell1_RETURN, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 366.
        Stack shape : RETURN expr.
        Start symbol: <undetermined>. *)

    | MenhirState367 : (('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_state
      (** State 367.
        Stack shape : REPEAT.
        Start symbol: <undetermined>. *)

    | MenhirState369 : (('s, 'r) _menhir_cell1_PRINTLN _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 369.
        Stack shape : PRINTLN LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState374 : (('s, 'r) _menhir_cell1_PRINT _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 374.
        Stack shape : PRINT LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState379 : (('s, 'r) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 379.
        Stack shape : PRAGMA IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState385 : (('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_state
      (** State 385.
        Stack shape : LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState387 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 387.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState388 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_state
      (** State 388.
        Stack shape : IDENTIFIER LLBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState389 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 389.
        Stack shape : IDENTIFIER LLBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState390 : ((((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LLBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_RRBRACKET, 'r) _menhir_state
      (** State 390.
        Stack shape : IDENTIFIER LLBRACKET expr RRBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState392 : (('s, 'r) _menhir_cell1_DOT _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 392.
        Stack shape : DOT IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState398 : ((('s, 'r) _menhir_cell1_LPAR, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_state
      (** State 398.
        Stack shape : LPAR discard_or_sliced_basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState400 : ((('s, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr, 'r) _menhir_state
      (** State 400.
        Stack shape : discard_or_sliced_basic_lexpr discard_or_sliced_basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState402 : (('s, 'r) _menhir_cell1_basic_lexpr, 'r) _menhir_state
      (** State 402.
        Stack shape : basic_lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState408 : (('s, 'r) _menhir_cell1_IF, 'r) _menhir_state
      (** State 408.
        Stack shape : IF.
        Start symbol: <undetermined>. *)

    | MenhirState409 : ((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 409.
        Stack shape : IF expr.
        Start symbol: <undetermined>. *)

    | MenhirState410 : (((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 410.
        Stack shape : IF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState411 : (('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 411.
        Stack shape : IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState412 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_LBRACE, 'r) _menhir_state
      (** State 412.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: <undetermined>. *)

    | MenhirState414 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 414.
        Stack shape : IDENTIFIER DOT LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState418 : (((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LPAR, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_state
      (** State 418.
        Stack shape : IDENTIFIER DOT LPAR discard_or_field.
        Start symbol: <undetermined>. *)

    | MenhirState420 : ((('s, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_cell1_discard_or_field, 'r) _menhir_state
      (** State 420.
        Stack shape : discard_or_field discard_or_field.
        Start symbol: <undetermined>. *)

    | MenhirState425 : ((('s, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_DOT _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 425.
        Stack shape : IDENTIFIER DOT LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState431 : (('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 431.
        Stack shape : FOR IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState432 : ((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 432.
        Stack shape : FOR IDENTIFIER expr.
        Start symbol: <undetermined>. *)

    | MenhirState435 : (((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_state
      (** State 435.
        Stack shape : FOR IDENTIFIER expr direction.
        Start symbol: <undetermined>. *)

    | MenhirState436 : ((((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 436.
        Stack shape : FOR IDENTIFIER expr direction expr.
        Start symbol: <undetermined>. *)

    | MenhirState438 : (((((('s, 'r) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_direction, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_loop_limit, 'r) _menhir_state
      (** State 438.
        Stack shape : FOR IDENTIFIER expr direction expr loop_limit.
        Start symbol: <undetermined>. *)

    | MenhirState440 : (('s, 'r) _menhir_cell1_DEBUG _menhir_cell0_LPAR, 'r) _menhir_state
      (** State 440.
        Stack shape : DEBUG LPAR.
        Start symbol: <undetermined>. *)

    | MenhirState445 : (('s, 'r) _menhir_cell1_CASE, 'r) _menhir_state
      (** State 445.
        Stack shape : CASE.
        Start symbol: <undetermined>. *)

    | MenhirState446 : ((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 446.
        Stack shape : CASE expr.
        Start symbol: <undetermined>. *)

    | MenhirState447 : (((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OF, 'r) _menhir_state
      (** State 447.
        Stack shape : CASE expr OF.
        Start symbol: <undetermined>. *)

    | MenhirState448 : (('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_state
      (** State 448.
        Stack shape : WHEN.
        Start symbol: <undetermined>. *)

    | MenhirState450 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_state
      (** State 450.
        Stack shape : WHEN pattern_list.
        Start symbol: <undetermined>. *)

    | MenhirState451 : (((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 451.
        Stack shape : WHEN pattern_list expr.
        Start symbol: <undetermined>. *)

    | MenhirState452 : ((((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_ARROW, 'r) _menhir_state
      (** State 452.
        Stack shape : WHEN pattern_list expr ARROW.
        Start symbol: <undetermined>. *)

    | MenhirState453 : (('s, 'r) _menhir_cell1_ASSERT, 'r) _menhir_state
      (** State 453.
        Stack shape : ASSERT.
        Start symbol: <undetermined>. *)

    | MenhirState454 : ((('s, 'r) _menhir_cell1_ASSERT, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 454.
        Stack shape : ASSERT expr.
        Start symbol: <undetermined>. *)

    | MenhirState457 : (('s, 'r) _menhir_cell1_stmt, 'r) _menhir_state
      (** State 457.
        Stack shape : stmt.
        Start symbol: <undetermined>. *)

    | MenhirState459 : (('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_state
      (** State 459.
        Stack shape : local_decl_keyword_non_var.
        Start symbol: <undetermined>. *)

    | MenhirState460 : ((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_state
      (** State 460.
        Stack shape : local_decl_keyword_non_var decl_item.
        Start symbol: <undetermined>. *)

    | MenhirState461 : (((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_state
      (** State 461.
        Stack shape : local_decl_keyword_non_var decl_item EQ.
        Start symbol: <undetermined>. *)

    | MenhirState462 : ((((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_EQ, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 462.
        Stack shape : local_decl_keyword_non_var decl_item EQ expr.
        Start symbol: <undetermined>. *)

    | MenhirState465 : (((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_state
      (** State 465.
        Stack shape : local_decl_keyword_non_var decl_item as_ty.
        Start symbol: <undetermined>. *)

    | MenhirState466 : ((((('s, 'r) _menhir_cell1_local_decl_keyword_non_var, 'r) _menhir_cell1_decl_item, 'r) _menhir_cell1_as_ty, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 466.
        Stack shape : local_decl_keyword_non_var decl_item as_ty expr.
        Start symbol: <undetermined>. *)

    | MenhirState472 : (('s, 'r) _menhir_cell1_lexpr, 'r) _menhir_state
      (** State 472.
        Stack shape : lexpr.
        Start symbol: <undetermined>. *)

    | MenhirState473 : ((('s, 'r) _menhir_cell1_lexpr, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 473.
        Stack shape : lexpr expr.
        Start symbol: <undetermined>. *)

    | MenhirState477 : (('s, 'r) _menhir_cell1_call, 'r) _menhir_state
      (** State 477.
        Stack shape : call.
        Start symbol: <undetermined>. *)

    | MenhirState478 : ((('s, 'r) _menhir_cell1_call, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 478.
        Stack shape : call expr.
        Start symbol: <undetermined>. *)

    | MenhirState481 : (('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_state
      (** State 481.
        Stack shape : call LBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState484 : ((('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET, 'r) _menhir_state
      (** State 484.
        Stack shape : call LBRACKET clist2(IDENTIFIER) RBRACKET.
        Start symbol: <undetermined>. *)

    | MenhirState485 : (((('s, 'r) _menhir_cell1_call _menhir_cell0_LBRACKET, 'r) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 485.
        Stack shape : call LBRACKET clist2(IDENTIFIER) RBRACKET expr.
        Start symbol: <undetermined>. *)

    | MenhirState488 : (('s, 'r) _menhir_cell1_call _menhir_cell0_IDENTIFIER, 'r) _menhir_state
      (** State 488.
        Stack shape : call IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState489 : ((('s, 'r) _menhir_cell1_call _menhir_cell0_IDENTIFIER, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 489.
        Stack shape : call IDENTIFIER expr.
        Start symbol: <undetermined>. *)

    | MenhirState492 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_pattern_list, 'r) _menhir_state
      (** State 492.
        Stack shape : WHEN pattern_list.
        Start symbol: <undetermined>. *)

    | MenhirState496 : ((((('s, 'r) _menhir_cell1_CASE, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_OF, 'r) _menhir_cell1_list1_case_alt_, 'r) _menhir_state
      (** State 496.
        Stack shape : CASE expr OF list1(case_alt).
        Start symbol: <undetermined>. *)

    | MenhirState502 : (('s, 'r) _menhir_cell1_case_alt, 'r) _menhir_state
      (** State 502.
        Stack shape : case_alt.
        Start symbol: <undetermined>. *)

    | MenhirState507 : ((((('s, 'r) _menhir_cell1_IF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 507.
        Stack shape : IF expr THEN stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState508 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_state
      (** State 508.
        Stack shape : expr THEN stmt_list ELSIF.
        Start symbol: <undetermined>. *)

    | MenhirState509 : (((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 509.
        Stack shape : expr THEN stmt_list ELSIF expr.
        Start symbol: <undetermined>. *)

    | MenhirState510 : ((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_state
      (** State 510.
        Stack shape : expr THEN stmt_list ELSIF expr THEN.
        Start symbol: <undetermined>. *)

    | MenhirState511 : (((((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSIF, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 511.
        Stack shape : expr THEN stmt_list ELSIF expr THEN stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState512 : ((((('s, 'r) _menhir_cell1_expr, 'r) _menhir_cell1_THEN, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_ELSE, 'r) _menhir_state
      (** State 512.
        Stack shape : expr THEN stmt_list ELSE.
        Start symbol: <undetermined>. *)

    | MenhirState519 : ((('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 519.
        Stack shape : REPEAT stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState520 : (((('s, 'r) _menhir_cell1_REPEAT, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_expr, 'r) _menhir_state
      (** State 520.
        Stack shape : REPEAT stmt_list expr.
        Start symbol: <undetermined>. *)

    | MenhirState524 : ((('s, 'r) _menhir_cell1_TRY, 'r) _menhir_cell1_stmt_list, 'r) _menhir_state
      (** State 524.
        Stack shape : TRY stmt_list.
        Start symbol: <undetermined>. *)

    | MenhirState525 : (('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_state
      (** State 525.
        Stack shape : WHEN.
        Start symbol: <undetermined>. *)

    | MenhirState527 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_state
      (** State 527.
        Stack shape : WHEN IDENTIFIER.
        Start symbol: <undetermined>. *)

    | MenhirState529 : (((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_IDENTIFIER, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 529.
        Stack shape : WHEN IDENTIFIER ty.
        Start symbol: <undetermined>. *)

    | MenhirState532 : ((('s, 'r) _menhir_cell1_WHEN, 'r) _menhir_cell1_ty, 'r) _menhir_state
      (** State 532.
        Stack shape : WHEN ty.
        Start symbol: <undetermined>. *)

    | MenhirState536 : (((('s, 'r) _menhir_cell1_TRY, 'r) _menhir_cell1_stmt_list, 'r) _menhir_cell1_list1_catcher_, 'r) _menhir_state
      (** State 536.
        Stack shape : TRY stmt_list list1(catcher).
        Start symbol: <undetermined>. *)

    | MenhirState541 : (('s, 'r) _menhir_cell1_catcher, 'r) _menhir_state
      (** State 541.
        Stack shape : catcher.
        Start symbol: <undetermined>. *)

    | MenhirState549 : ('s, _menhir_box_spec) _menhir_state
      (** State 549.
        Stack shape : .
        Start symbol: spec. *)

    | MenhirState550 : (('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_state
      (** State 550.
        Stack shape : VAR.
        Start symbol: spec. *)

    | MenhirState553 : ((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_state
      (** State 553.
        Stack shape : VAR ignored_or_identifier.
        Start symbol: spec. *)

    | MenhirState555 : (((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_state
      (** State 555.
        Stack shape : VAR ignored_or_identifier option(as_ty).
        Start symbol: spec. *)

    | MenhirState556 : ((((('s, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 556.
        Stack shape : VAR ignored_or_identifier option(as_ty) expr.
        Start symbol: spec. *)

    | MenhirState561 : (('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 561.
        Stack shape : TYPE IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState564 : (('s, _menhir_box_spec) _menhir_cell1_SUBTYPES _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 564.
        Stack shape : SUBTYPES IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState565 : (('s, _menhir_box_spec) _menhir_cell1_LBRACE, _menhir_box_spec) _menhir_state
      (** State 565.
        Stack shape : LBRACE.
        Start symbol: spec. *)

    | MenhirState566 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 566.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState569 : (('s, _menhir_box_spec) _menhir_cell1_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 569.
        Stack shape : typed_identifier.
        Start symbol: spec. *)

    | MenhirState575 : ((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_state
      (** State 575.
        Stack shape : TYPE IDENTIFIER OF.
        Start symbol: spec. *)

    | MenhirState576 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_RECORD, _menhir_box_spec) _menhir_state
      (** State 576.
        Stack shape : TYPE IDENTIFIER OF RECORD.
        Start symbol: spec. *)

    | MenhirState579 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_EXCEPTION, _menhir_box_spec) _menhir_state
      (** State 579.
        Stack shape : TYPE IDENTIFIER OF EXCEPTION.
        Start symbol: spec. *)

    | MenhirState582 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ENUMERATION _menhir_cell0_LBRACE, _menhir_box_spec) _menhir_state
      (** State 582.
        Stack shape : TYPE IDENTIFIER OF ENUMERATION LBRACE.
        Start symbol: spec. *)

    | MenhirState584 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 584.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState588 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_COLLECTION, _menhir_box_spec) _menhir_state
      (** State 588.
        Stack shape : TYPE IDENTIFIER OF COLLECTION.
        Start symbol: spec. *)

    | MenhirState590 : (((('s, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl, _menhir_box_spec) _menhir_state
      (** State 590.
        Stack shape : TYPE IDENTIFIER OF ty_decl.
        Start symbol: spec. *)

    | MenhirState598 : (('s, _menhir_box_spec) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 598.
        Stack shape : PRAGMA IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState604 : (('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 604.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState605 : (('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_LBRACE, _menhir_box_spec) _menhir_state
      (** State 605.
        Stack shape : IDENTIFIER LBRACE.
        Start symbol: spec. *)

    | MenhirState606 : (('s, _menhir_box_spec) _menhir_cell1_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 606.
        Stack shape : IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState609 : (('s, _menhir_box_spec) _menhir_cell1_opt_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 609.
        Stack shape : opt_typed_identifier.
        Start symbol: spec. *)

    | MenhirState613 : ((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 613.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState614 : ((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR, _menhir_box_spec) _menhir_state
      (** State 614.
        Stack shape : IDENTIFIER params_opt LPAR.
        Start symbol: spec. *)

    | MenhirState616 : (('s, _menhir_box_spec) _menhir_cell1_typed_identifier, _menhir_box_spec) _menhir_state
      (** State 616.
        Stack shape : typed_identifier.
        Start symbol: spec. *)

    | MenhirState621 : (((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 621.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState622 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN, _menhir_box_spec) _menhir_state
      (** State 622.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER params_opt func_args BEGIN.
        Start symbol: spec. *)

    | MenhirState627 : (((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ARROW, _menhir_box_spec) _menhir_state
      (** State 627.
        Stack shape : IDENTIFIER params_opt func_args ARROW.
        Start symbol: spec. *)

    | MenhirState629 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_state
      (** State 629.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER params_opt func_args return_type.
        Start symbol: spec. *)

    | MenhirState630 : ((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_RECURSELIMIT, _menhir_box_spec) _menhir_state
      (** State 630.
        Stack shape : IDENTIFIER params_opt func_args return_type RECURSELIMIT.
        Start symbol: spec. *)

    | MenhirState631 : (((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_RECURSELIMIT, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 631.
        Stack shape : IDENTIFIER params_opt func_args return_type RECURSELIMIT expr.
        Start symbol: spec. *)

    | MenhirState633 : (((((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit, _menhir_box_spec) _menhir_state
      (** State 633.
        Stack shape : IMPLEMENTATION FUNC IDENTIFIER params_opt func_args return_type recurse_limit.
        Start symbol: spec. *)

    | MenhirState638 : (('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 638.
        Stack shape : IMPLEMENTATION IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState639 : ((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 639.
        Stack shape : IMPLEMENTATION IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState641 : (((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 641.
        Stack shape : IMPLEMENTATION IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState643 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_state
      (** State 643.
        Stack shape : IMPLEMENTATION IDENTIFIER params_opt func_args ty.
        Start symbol: spec. *)

    | MenhirState647 : ((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 647.
        Stack shape : IDENTIFIER params_opt func_args ty SETTER IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState652 : (((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_SEMI_COLON, _menhir_box_spec) _menhir_state
      (** State 652.
        Stack shape : IDENTIFIER params_opt func_args ty SETTER IDENTIFIER maybe_empty_stmt_list END SEMI_COLON.
        Start symbol: spec. *)

    | MenhirState657 : (((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END, _menhir_box_spec) _menhir_state
      (** State 657.
        Stack shape : IDENTIFIER params_opt func_args ty SETTER IDENTIFIER maybe_empty_stmt_list END.
        Start symbol: spec. *)

    | MenhirState662 : ((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER, _menhir_box_spec) _menhir_state
      (** State 662.
        Stack shape : IDENTIFIER params_opt func_args ty GETTER.
        Start symbol: spec. *)

    | MenhirState668 : (((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 668.
        Stack shape : IDENTIFIER params_opt func_args ty GETTER maybe_empty_stmt_list END IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState676 : (((((('s _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_SEMI_COLON _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 676.
        Stack shape : IDENTIFIER params_opt func_args ty GETTER maybe_empty_stmt_list END SEMI_COLON IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState685 : (('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 685.
        Stack shape : IMPDEF FUNC IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState686 : ((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 686.
        Stack shape : IMPDEF FUNC IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState687 : (((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 687.
        Stack shape : IMPDEF FUNC IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState688 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN, _menhir_box_spec) _menhir_state
      (** State 688.
        Stack shape : IMPDEF FUNC IDENTIFIER params_opt func_args BEGIN.
        Start symbol: spec. *)

    | MenhirState692 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_state
      (** State 692.
        Stack shape : IMPDEF FUNC IDENTIFIER params_opt func_args return_type.
        Start symbol: spec. *)

    | MenhirState694 : (((((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit, _menhir_box_spec) _menhir_state
      (** State 694.
        Stack shape : IMPDEF FUNC IDENTIFIER params_opt func_args return_type recurse_limit.
        Start symbol: spec. *)

    | MenhirState699 : (('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 699.
        Stack shape : IMPDEF IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState700 : ((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 700.
        Stack shape : IMPDEF IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState702 : (((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 702.
        Stack shape : IMPDEF IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState704 : ((((('s, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_state
      (** State 704.
        Stack shape : IMPDEF IDENTIFIER params_opt func_args ty.
        Start symbol: spec. *)

    | MenhirState709 : (('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 709.
        Stack shape : FUNC IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState710 : ((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 710.
        Stack shape : FUNC IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState711 : (((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 711.
        Stack shape : FUNC IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState712 : ((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN, _menhir_box_spec) _menhir_state
      (** State 712.
        Stack shape : FUNC IDENTIFIER params_opt func_args BEGIN.
        Start symbol: spec. *)

    | MenhirState716 : ((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_state
      (** State 716.
        Stack shape : FUNC IDENTIFIER params_opt func_args return_type.
        Start symbol: spec. *)

    | MenhirState718 : (((((('s, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit, _menhir_box_spec) _menhir_state
      (** State 718.
        Stack shape : FUNC IDENTIFIER params_opt func_args return_type recurse_limit.
        Start symbol: spec. *)

    | MenhirState723 : (('s, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_state
      (** State 723.
        Stack shape : CONFIG.
        Start symbol: spec. *)

    | MenhirState724 : ((('s, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_state
      (** State 724.
        Stack shape : CONFIG ignored_or_identifier.
        Start symbol: spec. *)

    | MenhirState726 : (((('s, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_as_ty, _menhir_box_spec) _menhir_state
      (** State 726.
        Stack shape : CONFIG ignored_or_identifier as_ty.
        Start symbol: spec. *)

    | MenhirState727 : ((((('s, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_as_ty, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 727.
        Stack shape : CONFIG ignored_or_identifier as_ty expr.
        Start symbol: spec. *)

    | MenhirState730 : (('s, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_state
      (** State 730.
        Stack shape : ACCESSOR IDENTIFIER.
        Start symbol: spec. *)

    | MenhirState731 : ((('s, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_state
      (** State 731.
        Stack shape : ACCESSOR IDENTIFIER params_opt.
        Start symbol: spec. *)

    | MenhirState733 : (((('s, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_state
      (** State 733.
        Stack shape : ACCESSOR IDENTIFIER params_opt func_args.
        Start symbol: spec. *)

    | MenhirState735 : ((((('s, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_state
      (** State 735.
        Stack shape : ACCESSOR IDENTIFIER params_opt func_args ty.
        Start symbol: spec. *)

    | MenhirState742 : (('s, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_state
      (** State 742.
        Stack shape : global_let_or_constant.
        Start symbol: spec. *)

    | MenhirState743 : ((('s, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_state
      (** State 743.
        Stack shape : global_let_or_constant ignored_or_identifier.
        Start symbol: spec. *)

    | MenhirState745 : (((('s, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_state
      (** State 745.
        Stack shape : global_let_or_constant ignored_or_identifier option(as_ty).
        Start symbol: spec. *)

    | MenhirState746 : ((((('s, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_, _menhir_box_spec) _menhir_cell1_expr, _menhir_box_spec) _menhir_state
      (** State 746.
        Stack shape : global_let_or_constant ignored_or_identifier option(as_ty) expr.
        Start symbol: spec. *)

    | MenhirState749 : (('s, _menhir_box_spec) _menhir_cell1_decl, _menhir_box_spec) _menhir_state
      (** State 749.
        Stack shape : decl.
        Start symbol: spec. *)

    | MenhirState751 : ('s, _menhir_box_stmts) _menhir_state
      (** State 751.
        Stack shape : .
        Start symbol: stmts. *)

  
  and ('s, 'r) _menhir_cell1_as_ty = 
    | MenhirCell1_as_ty of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_basic_lexpr = 
    | MenhirCell1_basic_lexpr of 's * ('s, 'r) _menhir_state * (Desugar.lhs_access) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_bitfield = 
    | MenhirCell1_bitfield of 's * ('s, 'r) _menhir_state * (AST.bitfield)
  
  and ('s, 'r) _menhir_cell1_call = 
    | MenhirCell1_call of 's * ('s, 'r) _menhir_state * (AST.call) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_case_alt = 
    | MenhirCell1_case_alt of 's * ('s, 'r) _menhir_state * (AST.case_alt_desc AST.annotated)
  
  and ('s, 'r) _menhir_cell1_catcher = 
    | MenhirCell1_catcher of 's * ('s, 'r) _menhir_state * (AST.catcher)
  
  and ('s, 'r) _menhir_cell1_clist1_expr_ = 
    | MenhirCell1_clist1_expr_ of 's * ('s, 'r) _menhir_state * (AST.expr list)
  
  and ('s, 'r) _menhir_cell1_clist2_IDENTIFIER_ = 
    | MenhirCell1_clist2_IDENTIFIER_ of 's * ('s, 'r) _menhir_state * (string list)
  
  and ('s, 'r) _menhir_cell1_decl = 
    | MenhirCell1_decl of 's * ('s, 'r) _menhir_state * (AST.decl list)
  
  and ('s, 'r) _menhir_cell1_decl_item = 
    | MenhirCell1_decl_item of 's * ('s, 'r) _menhir_state * (AST.local_decl_item)
  
  and ('s, 'r) _menhir_cell1_direction = 
    | MenhirCell1_direction of 's * ('s, 'r) _menhir_state * (AST.for_direction)
  
  and ('s, 'r) _menhir_cell1_discard_or_field = 
    | MenhirCell1_discard_or_field of 's * ('s, 'r) _menhir_state * (Desugar.lhs_field option)
  
  and ('s, 'r) _menhir_cell1_discard_or_identifier = 
    | MenhirCell1_discard_or_identifier of 's * ('s, 'r) _menhir_state * (string)
  
  and ('s, 'r) _menhir_cell1_discard_or_sliced_basic_lexpr = 
    | MenhirCell1_discard_or_sliced_basic_lexpr of 's * ('s, 'r) _menhir_state * (Desugar.lhs_access option)
  
  and ('s, 'r) _menhir_cell1_expr = 
    | MenhirCell1_expr of 's * ('s, 'r) _menhir_state * (AST.expr) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_expr_pattern = 
    | MenhirCell1_expr_pattern of 's * ('s, 'r) _menhir_state * (AST.expr) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_field_assign = 
    | MenhirCell1_field_assign of 's * ('s, 'r) _menhir_state * (string * AST.expr)
  
  and ('s, 'r) _menhir_cell1_func_args = 
    | MenhirCell1_func_args of 's * ('s, 'r) _menhir_state * (AST.typed_identifier list)
  
  and ('s, 'r) _menhir_cell1_global_let_or_constant = 
    | MenhirCell1_global_let_or_constant of 's * ('s, 'r) _menhir_state * (AST.global_decl_keyword) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ignored_or_identifier = 
    | MenhirCell1_ignored_or_identifier of 's * ('s, 'r) _menhir_state * (string)
  
  and ('s, 'r) _menhir_cell1_int_constraint = 
    | MenhirCell1_int_constraint of 's * ('s, 'r) _menhir_state * (AST.int_constraint)
  
  and ('s, 'r) _menhir_cell1_lexpr = 
    | MenhirCell1_lexpr of 's * ('s, 'r) _menhir_state * (AST.lexpr) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_list1_case_alt_ = 
    | MenhirCell1_list1_case_alt_ of 's * ('s, 'r) _menhir_state * (AST.case_alt_desc AST.annotated list)
  
  and ('s, 'r) _menhir_cell1_list1_catcher_ = 
    | MenhirCell1_list1_catcher_ of 's * ('s, 'r) _menhir_state * (AST.catcher list)
  
  and ('s, 'r) _menhir_cell1_local_decl_keyword_non_var = 
    | MenhirCell1_local_decl_keyword_non_var of 's * ('s, 'r) _menhir_state * (AST.local_decl_keyword) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_loop_limit = 
    | MenhirCell1_loop_limit of 's * ('s, 'r) _menhir_state * (AST.expr option)
  
  and ('s, 'r) _menhir_cell1_maybe_empty_stmt_list = 
    | MenhirCell1_maybe_empty_stmt_list of 's * ('s, 'r) _menhir_state * (AST.stmt)
  
  and ('s, 'r) _menhir_cell1_opt_typed_identifier = 
    | MenhirCell1_opt_typed_identifier of 's * ('s, 'r) _menhir_state * (string * AST.ty option)
  
  and ('s, 'r) _menhir_cell1_option_as_ty_ = 
    | MenhirCell1_option_as_ty_ of 's * ('s, 'r) _menhir_state * (AST.ty option)
  
  and ('s, 'r) _menhir_cell1_params_opt = 
    | MenhirCell1_params_opt of 's * ('s, 'r) _menhir_state * ((string * AST.ty option) list)
  
  and ('s, 'r) _menhir_cell1_pattern = 
    | MenhirCell1_pattern of 's * ('s, 'r) _menhir_state * (AST.pattern) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_pattern_list = 
    | MenhirCell1_pattern_list of 's * ('s, 'r) _menhir_state * (AST.pattern)
  
  and ('s, 'r) _menhir_cell1_recurse_limit = 
    | MenhirCell1_recurse_limit of 's * ('s, 'r) _menhir_state * (AST.expr option)
  
  and ('s, 'r) _menhir_cell1_return_type = 
    | MenhirCell1_return_type of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_slice = 
    | MenhirCell1_slice of 's * ('s, 'r) _menhir_state * (AST.slice)
  
  and ('s, 'r) _menhir_cell1_slices = 
    | MenhirCell1_slices of 's * ('s, 'r) _menhir_state * (AST.slice list) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_stmt = 
    | MenhirCell1_stmt of 's * ('s, 'r) _menhir_state * (AST.stmt_desc AST.annotated) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_stmt_list = 
    | MenhirCell1_stmt_list of 's * ('s, 'r) _menhir_state * (AST.stmt) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ty = 
    | MenhirCell1_ty of 's * ('s, 'r) _menhir_state * (AST.ty) * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ty_decl = 
    | MenhirCell1_ty_decl of 's * ('s, 'r) _menhir_state * (AST.ty)
  
  and ('s, 'r) _menhir_cell1_typed_identifier = 
    | MenhirCell1_typed_identifier of 's * ('s, 'r) _menhir_state * (AST.field)
  
  and ('s, 'r) _menhir_cell1_ACCESSOR = 
    | MenhirCell1_ACCESSOR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_AND = 
    | MenhirCell1_AND of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ARBITRARY = 
    | MenhirCell1_ARBITRARY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ARRAY = 
    | MenhirCell1_ARRAY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ARROW = 
    | MenhirCell1_ARROW of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_AS = 
    | MenhirCell1_AS of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ASSERT = 
    | MenhirCell1_ASSERT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BAND = 
    | MenhirCell1_BAND of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BEGIN = 
    | MenhirCell1_BEGIN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BEQ = 
    | MenhirCell1_BEQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_BITS = 
    | MenhirCell1_BITS of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BNOT = 
    | MenhirCell1_BNOT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_BOR = 
    | MenhirCell1_BOR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_CASE = 
    | MenhirCell1_CASE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_COLLECTION = 
    | MenhirCell1_COLLECTION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_COLON = 
    | MenhirCell1_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_COLON_COLON = 
    | MenhirCell1_COLON_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_COMMA = 
    | MenhirCell1_COMMA of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_CONFIG = 
    | MenhirCell1_CONFIG of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_DEBUG = 
    | MenhirCell1_DEBUG of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_DIV = 
    | MenhirCell1_DIV of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DIVRM = 
    | MenhirCell1_DIVRM of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_DOT = 
    | MenhirCell1_DOT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ELSE = 
    | MenhirCell1_ELSE of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_ELSIF = 
    | MenhirCell1_ELSIF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_END = 
    | MenhirCell0_END of 's * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_ENUMERATION = 
    | MenhirCell1_ENUMERATION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_EQ = 
    | MenhirCell1_EQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_EQ_OP = 
    | MenhirCell1_EQ_OP of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_EXCEPTION = 
    | MenhirCell1_EXCEPTION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FOR = 
    | MenhirCell1_FOR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_FUNC = 
    | MenhirCell1_FUNC of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_FUNC = 
    | MenhirCell0_FUNC of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_GEQ = 
    | MenhirCell1_GEQ of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_GETTER = 
    | MenhirCell1_GETTER of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_GT = 
    | MenhirCell1_GT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_IDENTIFIER = 
    | MenhirCell1_IDENTIFIER of 's * ('s, 'r) _menhir_state * (
# 134 "asllib/Tokens.mly"
       (string)
# 2146 "asllib/Parser.ml"
  ) * Lexing.position * Lexing.position
  
  and 's _menhir_cell0_IDENTIFIER = 
    | MenhirCell0_IDENTIFIER of 's * (
# 134 "asllib/Tokens.mly"
       (string)
# 2153 "asllib/Parser.ml"
  ) * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IF = 
    | MenhirCell1_IF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IMPDEF = 
    | MenhirCell1_IMPDEF of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IMPL = 
    | MenhirCell1_IMPL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_IMPLEMENTATION = 
    | MenhirCell1_IMPLEMENTATION of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_IN = 
    | MenhirCell1_IN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_INTEGER = 
    | MenhirCell1_INTEGER of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LBRACE = 
    | MenhirCell1_LBRACE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LBRACE = 
    | MenhirCell0_LBRACE of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LBRACKET = 
    | MenhirCell1_LBRACKET of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LBRACKET = 
    | MenhirCell0_LBRACKET of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LEQ = 
    | MenhirCell1_LEQ of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LLBRACKET = 
    | MenhirCell1_LLBRACKET of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_LOOPLIMIT = 
    | MenhirCell1_LOOPLIMIT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_LPAR = 
    | MenhirCell1_LPAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_LPAR = 
    | MenhirCell0_LPAR of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_LT = 
    | MenhirCell1_LT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_MINUS = 
    | MenhirCell1_MINUS of 's * ('s, 'r) _menhir_state * Lexing.position * Lexing.position
  
  and ('s, 'r) _menhir_cell1_MOD = 
    | MenhirCell1_MOD of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_MUL = 
    | MenhirCell1_MUL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_NEQ = 
    | MenhirCell1_NEQ of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_NOT = 
    | MenhirCell1_NOT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_OF = 
    | MenhirCell1_OF of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_OR = 
    | MenhirCell1_OR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PLUS = 
    | MenhirCell1_PLUS of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PLUS_COLON = 
    | MenhirCell1_PLUS_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_POW = 
    | MenhirCell1_POW of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_PRAGMA = 
    | MenhirCell1_PRAGMA of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PRINT = 
    | MenhirCell1_PRINT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_PRINTLN = 
    | MenhirCell1_PRINTLN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RBRACE = 
    | MenhirCell1_RBRACE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_RBRACE = 
    | MenhirCell0_RBRACE of 's * Lexing.position
  
  and 's _menhir_cell0_RBRACKET = 
    | MenhirCell0_RBRACKET of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RDIV = 
    | MenhirCell1_RDIV of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_RECORD = 
    | MenhirCell1_RECORD of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RECURSELIMIT = 
    | MenhirCell1_RECURSELIMIT of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_REPEAT = 
    | MenhirCell1_REPEAT of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RETURN = 
    | MenhirCell1_RETURN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RPAR = 
    | MenhirCell1_RPAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_RRBRACKET = 
    | MenhirCell1_RRBRACKET of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and 's _menhir_cell0_SEMI_COLON = 
    | MenhirCell0_SEMI_COLON of 's * Lexing.position
  
  and ('s, 'r) _menhir_cell1_SETTER = 
    | MenhirCell1_SETTER of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SHL = 
    | MenhirCell1_SHL of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SHR = 
    | MenhirCell1_SHR of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SLICING = 
    | MenhirCell1_SLICING of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_STAR_COLON = 
    | MenhirCell1_STAR_COLON of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_SUBTYPES = 
    | MenhirCell1_SUBTYPES of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_THEN = 
    | MenhirCell1_THEN of 's * ('s, 'r) _menhir_state
  
  and ('s, 'r) _menhir_cell1_THROW = 
    | MenhirCell1_THROW of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TRY = 
    | MenhirCell1_TRY of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_TYPE = 
    | MenhirCell1_TYPE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_VAR = 
    | MenhirCell1_VAR of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_WHEN = 
    | MenhirCell1_WHEN of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_WHILE = 
    | MenhirCell1_WHILE of 's * ('s, 'r) _menhir_state * Lexing.position
  
  and ('s, 'r) _menhir_cell1_XOR = 
    | MenhirCell1_XOR of 's * ('s, 'r) _menhir_state
  
  and _menhir_box_stmts = 
    | MenhirBox_stmts of (AST.stmt) [@@unboxed]
  
  and _menhir_box_spec = 
    | MenhirBox_spec of (AST.t) [@@unboxed]
  
  and _menhir_box_opn = 
    | MenhirBox_opn of (AST.t) [@@unboxed]
  
  let _menhir_action_003 =
    fun _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _3 _4 setter_arg x x_inlined1 ->
      let setter =
        let (_2, _1_inlined1, x) = (_2_inlined1, _1_inlined4, x_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2338 "asllib/Parser.ml"
        
      in
      let getter =
        let _1_inlined1 = _1_inlined2 in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2350 "asllib/Parser.ml"
        
      in
      (
# 597 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2356 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_004 =
    fun _1 _1_inlined2 _1_inlined4 _2 _3 _4 _endpos__1_inlined4_ _startpos__1_inlined4_ setter_arg x x_inlined1 ->
      let setter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, x) = (_endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4, x_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2373 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2379 "asllib/Parser.ml"
        
      in
      let getter =
        let _1_inlined1 = _1_inlined2 in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2391 "asllib/Parser.ml"
        
      in
      (
# 597 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2397 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_005 =
    fun _1 _1_inlined2 _1_inlined4 _2 _3 _4 _endpos__1_inlined2_ _startpos__1_inlined2_ setter_arg x x_inlined1 ->
      let setter =
        let (_1_inlined1, x) = (_1_inlined4, x_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2411 "asllib/Parser.ml"
        
      in
      let getter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2426 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2432 "asllib/Parser.ml"
        
      in
      (
# 597 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2438 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_006 =
    fun _1 _1_inlined2 _1_inlined4 _3 _4 _endpos__1_inlined2_ _endpos__1_inlined4_ _startpos__1_inlined2_ _startpos__1_inlined4_ setter_arg x x_inlined1 ->
      let setter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, x) = (_endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4, x_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2455 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2461 "asllib/Parser.ml"
        
      in
      let getter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2476 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2482 "asllib/Parser.ml"
        
      in
      (
# 597 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2488 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_007 =
    fun _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _2_inlined2 _5 setter_arg x x_inlined1 ->
      let getter =
        let (_2, _1_inlined1, x) = (_2_inlined2, _1_inlined4, x_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2502 "asllib/Parser.ml"
        
      in
      let setter =
        let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2514 "asllib/Parser.ml"
        
      in
      (
# 600 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2520 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_008 =
    fun _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _5 _endpos__1_inlined4_ _startpos__1_inlined4_ setter_arg x x_inlined1 ->
      let getter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, x) = (_endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4, x_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2537 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2543 "asllib/Parser.ml"
        
      in
      let setter =
        let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2555 "asllib/Parser.ml"
        
      in
      (
# 600 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2561 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_009 =
    fun _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _5 _endpos__1_inlined2_ _startpos__1_inlined2_ setter_arg x x_inlined1 ->
      let getter =
        let (_2, _1_inlined1, x) = (_2_inlined1, _1_inlined4, x_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 205 "<standard.mly>"
    ( x )
# 2575 "asllib/Parser.ml"
        
      in
      let setter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2590 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2596 "asllib/Parser.ml"
        
      in
      (
# 600 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2602 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_010 =
    fun _1 _1_inlined2 _1_inlined4 _2 _5 _endpos__1_inlined2_ _endpos__1_inlined4_ _startpos__1_inlined2_ _startpos__1_inlined4_ setter_arg x x_inlined1 ->
      let getter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, x) = (_endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4, x_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2619 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2625 "asllib/Parser.ml"
        
      in
      let setter =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 2640 "asllib/Parser.ml"
          
        in
        
# 205 "<standard.mly>"
    ( x )
# 2646 "asllib/Parser.ml"
        
      in
      (
# 600 "asllib/Parser.mly"
    ( { getter; setter; setter_arg } )
# 2652 "asllib/Parser.ml"
       : (Desugar.accessor_pair))
  
  let _menhir_action_011 =
    fun _1 _2 ->
      (_2 : (AST.ty))
  
  let _menhir_action_012 =
    fun _endpos_desc_ _startpos_desc_ desc fields ->
      let base =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2667 "asllib/Parser.ml"
        
      in
      (
# 423 "asllib/Parser.mly"
    ( lhs_access ~base ~index:None ~fields )
# 2673 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_013 =
    fun _2 _4 _endpos_desc_ _startpos_desc_ desc fields idx ->
      let base =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2684 "asllib/Parser.ml"
        
      in
      (
# 425 "asllib/Parser.mly"
    ( lhs_access ~base ~index:(Some idx) ~fields )
# 2690 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_014 =
    fun s x ->
      (
# 375 "asllib/Parser.mly"
                                              ( BitField_Simple (x, s)     )
# 2698 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_015 =
    fun bf s x ->
      (
# 376 "asllib/Parser.mly"
                                              ( BitField_Nested (x, s, bf) )
# 2706 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_016 =
    fun s ty x ->
      (
# 377 "asllib/Parser.mly"
                                              ( BitField_Type   (x, s, ty) )
# 2714 "asllib/Parser.ml"
       : (AST.bitfield))
  
  let _menhir_action_017 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 2722 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_018 =
    fun () ->
      (
# 372 "asllib/Parser.mly"
                     ( [] )
# 2730 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_019 =
    fun _1 ->
      (_1 : (AST.bitfield list))
  
  let _menhir_action_020 =
    fun name x ->
      let args = 
# 205 "<standard.mly>"
    ( x )
# 2742 "asllib/Parser.ml"
       in
      (
# 574 "asllib/Parser.mly"
    ( { name; params=[]; args; call_type = ST_Function } )
# 2747 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_021 =
    fun name x ->
      let args = 
# 571 "asllib/Parser.mly"
                     ( [] )
# 2755 "asllib/Parser.ml"
       in
      let params = 
# 205 "<standard.mly>"
    ( x )
# 2760 "asllib/Parser.ml"
       in
      (
# 576 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2765 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_022 =
    fun name x x_inlined1 ->
      let args =
        let x = x_inlined1 in
        
# 205 "<standard.mly>"
    ( x )
# 2775 "asllib/Parser.ml"
        
      in
      let params = 
# 205 "<standard.mly>"
    ( x )
# 2781 "asllib/Parser.ml"
       in
      (
# 576 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 2786 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_023 =
    fun _1 _4 _endpos_stmt_ _startpos__1_ pattern stmt ->
      let desc =
        let where = 
# 123 "<standard.mly>"
    ( None )
# 2795 "asllib/Parser.ml"
         in
        
# 487 "asllib/Parser.mly"
      ( { pattern; where; stmt } )
# 2800 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_stmt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2809 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated))
  
  let _menhir_action_024 =
    fun _1 _1_inlined1 _2 _4 _endpos_stmt_ _startpos__1_ pattern stmt ->
      let desc =
        let where =
          let _1 = _1_inlined1 in
          let x = _2 in
          
# 126 "<standard.mly>"
    ( Some x )
# 2821 "asllib/Parser.ml"
          
        in
        
# 487 "asllib/Parser.mly"
      ( { pattern; where; stmt } )
# 2827 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_stmt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 2836 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated))
  
  let _menhir_action_025 =
    fun _1 _4 stmt_list ty ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 2844 "asllib/Parser.ml"
       in
      let _2_1 = _2 in
      ((_2_1, ty, stmt_list) : (AST.catcher))
  
  let _menhir_action_026 =
    fun _1 _4 stmt_list ty x ->
      let _2 =
        let x = 
# 196 "<standard.mly>"
    ( x )
# 2855 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 2860 "asllib/Parser.ml"
        
      in
      let _2_1 = _2 in
      ((_2_1, ty, stmt_list) : (AST.catcher))
  
  let _menhir_action_027 =
    fun () ->
      (
# 167 "asllib/Parser.mly"
                 ( [] )
# 2871 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_028 =
    fun _1 ->
      (_1 : (AST.expr list))
  
  let _menhir_action_029 =
    fun () ->
      (
# 167 "asllib/Parser.mly"
                 ( [] )
# 2883 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_030 =
    fun _1 ->
      (_1 : (AST.ty list))
  
  let _menhir_action_031 =
    fun () ->
      (
# 167 "asllib/Parser.mly"
                 ( [] )
# 2895 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_032 =
    fun _1 ->
      (_1 : (AST.typed_identifier list))
  
  let _menhir_action_033 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2907 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_034 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2915 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_035 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2923 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_036 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2931 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_037 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2939 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_038 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2947 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_039 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2955 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_040 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2963 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_041 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2971 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_042 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2979 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_043 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 2987 "asllib/Parser.ml"
       : ((string * AST.expr) list))
  
  let _menhir_action_044 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 2995 "asllib/Parser.ml"
       : ((string * AST.expr) list))
  
  let _menhir_action_045 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3003 "asllib/Parser.ml"
       : (AST.int_constraint list))
  
  let _menhir_action_046 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3011 "asllib/Parser.ml"
       : (AST.int_constraint list))
  
  let _menhir_action_047 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3019 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_048 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3027 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_049 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3035 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_050 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3043 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_051 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3051 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_052 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3059 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_053 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3067 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_054 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3075 "asllib/Parser.ml"
       : (AST.ty list))
  
  let _menhir_action_055 =
    fun x ->
      (
# 163 "asllib/Parser.mly"
         ( [ x ] )
# 3083 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_056 =
    fun _2 h t ->
      (
# 164 "asllib/Parser.mly"
                             ( h :: t )
# 3091 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_057 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3099 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_058 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3107 "asllib/Parser.ml"
       : (Desugar.lhs_field option list))
  
  let _menhir_action_059 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3115 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_060 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3123 "asllib/Parser.ml"
       : (Desugar.lhs_access option list))
  
  let _menhir_action_061 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3131 "asllib/Parser.ml"
       : (AST.expr list))
  
  let _menhir_action_062 =
    fun _2 li x ->
      (
# 170 "asllib/Parser.mly"
                                           ( x :: li )
# 3139 "asllib/Parser.ml"
       : (AST.pattern list))
  
  let _menhir_action_063 =
    fun x ->
      let cs = 
# 205 "<standard.mly>"
    ( x )
# 3147 "asllib/Parser.ml"
       in
      (
# 298 "asllib/Parser.mly"
                                       ( WellConstrained (cs, Precision_Full) )
# 3152 "asllib/Parser.ml"
       : (AST.constraint_kind))
  
  let _menhir_action_064 =
    fun x ->
      let _1 = 
# 205 "<standard.mly>"
    ( x )
# 3160 "asllib/Parser.ml"
       in
      (
# 299 "asllib/Parser.mly"
                   ( PendingConstrained )
# 3165 "asllib/Parser.ml"
       : (AST.constraint_kind))
  
  let _menhir_action_065 =
    fun _1 ->
      (_1 : (AST.constraint_kind))
  
  let _menhir_action_066 =
    fun () ->
      (
# 296 "asllib/Parser.mly"
                                             ( UnConstrained )
# 3177 "asllib/Parser.ml"
       : (AST.constraint_kind))
  
  let _menhir_action_067 =
    fun _1_inlined1 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__2_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let _2 = _2_inlined1 in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3193 "asllib/Parser.ml"
            
          in
          let override = 
# 123 "<standard.mly>"
    ( None )
# 3199 "asllib/Parser.ml"
           in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3216 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__2_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3225 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3231 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_068 =
    fun _1_inlined1 _2 _endpos__1_inlined1_ _startpos__1_inlined1_ _startpos__2_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3249 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3255 "asllib/Parser.ml"
            
          in
          let override = 
# 123 "<standard.mly>"
    ( None )
# 3261 "asllib/Parser.ml"
           in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3278 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__2_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3287 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3293 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_069 =
    fun _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3309 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 3316 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3321 "asllib/Parser.ml"
            
          in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3339 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3348 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3354 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_070 =
    fun _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3373 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3379 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 3386 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3391 "asllib/Parser.ml"
            
          in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3409 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3418 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3424 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_071 =
    fun _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3440 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 3447 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3452 "asllib/Parser.ml"
            
          in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3470 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3479 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3485 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_072 =
    fun _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt recurse_limit return_type x ->
      let d =
        let desc =
          let body =
            let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3504 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3510 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 3517 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3522 "asllib/Parser.ml"
            
          in
          
# 606 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = Some return_type;
            subprogram_type = ST_Function;
            recurse_limit;
            override;
            builtin = false;
          }
        )
# 3540 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3549 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3555 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_073 =
    fun _1_inlined1 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__2_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let _2 = _2_inlined1 in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3571 "asllib/Parser.ml"
            
          in
          let override = 
# 123 "<standard.mly>"
    ( None )
# 3577 "asllib/Parser.ml"
           in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3594 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__2_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3603 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3609 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_074 =
    fun _1_inlined1 _2 _endpos__1_inlined1_ _startpos__1_inlined1_ _startpos__2_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3627 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3633 "asllib/Parser.ml"
            
          in
          let override = 
# 123 "<standard.mly>"
    ( None )
# 3639 "asllib/Parser.ml"
           in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3656 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__2_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3665 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3671 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_075 =
    fun _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3687 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 3694 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3699 "asllib/Parser.ml"
            
          in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3717 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3726 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3732 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_076 =
    fun _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3751 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3757 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 3764 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3769 "asllib/Parser.ml"
            
          in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3787 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3796 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3802 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_077 =
    fun _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let (_2, _1_inlined1) = (_2_inlined1, _1_inlined2) in
            let _3 =
              let _1 = _1_inlined1 in
              ()
            in
            
# 205 "<standard.mly>"
    ( x )
# 3818 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 3825 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3830 "asllib/Parser.ml"
            
          in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3848 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_inlined1_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3857 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3863 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_078 =
    fun _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt x ->
      let d =
        let desc =
          let body =
            let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
            let _3 =
              let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
              let _endpos = _endpos__1_ in
              let _startpos = _startpos__1_ in
              
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 3882 "asllib/Parser.ml"
              
            in
            
# 205 "<standard.mly>"
    ( x )
# 3888 "asllib/Parser.ml"
            
          in
          let override =
            let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 3895 "asllib/Parser.ml"
             in
            
# 126 "<standard.mly>"
    ( Some x )
# 3900 "asllib/Parser.ml"
            
          in
          
# 622 "asllib/Parser.mly"
        (
          D_Func {
            name;
            parameters = params_opt;
            args = func_args;
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override;
            builtin = false;
          }
        )
# 3918 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3927 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3933 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_079 =
    fun _1 _3 _endpos__2_ _startpos__1_ subtype_opt t x ->
      let d =
        let desc =
          let x = 
# 638 "asllib/Parser.mly"
                                                            D_TypeDecl            
# 3943 "asllib/Parser.ml"
           (x, t, subtype_opt) in
          
# 196 "<standard.mly>"
    ( x )
# 3948 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3957 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 3963 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_080 =
    fun _1 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc x ->
      let d =
        let desc =
          let x =
            let s =
              let _endpos = _endpos_desc_ in
              let _symbolstartpos = _startpos_desc_ in
              
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3977 "asllib/Parser.ml"
              
            in
            
# 641 "asllib/Parser.mly"
                                                            make_ty_decl_subtype  
# 3983 "asllib/Parser.ml"
             (x, s)
          in
          
# 196 "<standard.mly>"
    ( x )
# 3989 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 3998 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4004 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_081 =
    fun _4 _endpos__2_ _startpos_keyword_ keyword name ty x ->
      let d =
        let desc =
          let x =
            let initial_value = 
# 149 "asllib/Parser.mly"
                        Some 
# 4015 "asllib/Parser.ml"
             x in
            
# 646 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword; name; ty; initial_value } )
# 4020 "asllib/Parser.ml"
            
          in
          
# 196 "<standard.mly>"
    ( x )
# 4026 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_keyword_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4035 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4041 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_082 =
    fun _1 _4 _endpos__2_ _startpos__1_ name ty x ->
      let d =
        let desc =
          let x =
            let initial_value = 
# 149 "asllib/Parser.mly"
                        Some 
# 4052 "asllib/Parser.ml"
             x in
            
# 649 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword=GDK_Config; name; ty=Some ty; initial_value } )
# 4057 "asllib/Parser.ml"
            
          in
          
# 196 "<standard.mly>"
    ( x )
# 4063 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4072 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4078 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_083 =
    fun _1 _4 _endpos__2_ _startpos__1_ name ty x ->
      let d =
        let desc =
          let x =
            let initial_value = 
# 149 "asllib/Parser.mly"
                        Some 
# 4089 "asllib/Parser.ml"
             x in
            
# 652 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword=GDK_Var; name; ty; initial_value } )
# 4094 "asllib/Parser.ml"
            
          in
          
# 196 "<standard.mly>"
    ( x )
# 4100 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4109 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4115 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_084 =
    fun _1 _endpos__2_ _startpos__1_ name x ->
      let d =
        let desc =
          let x =
            let ty = 
# 149 "asllib/Parser.mly"
                        Some 
# 4126 "asllib/Parser.ml"
             x in
            
# 656 "asllib/Parser.mly"
        ( D_GlobalStorage { keyword=GDK_Var; name; ty; initial_value=None})
# 4131 "asllib/Parser.ml"
            
          in
          
# 196 "<standard.mly>"
    ( x )
# 4137 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4146 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4152 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_085 =
    fun _1 _endpos__2_ _startpos__1_ e x ->
      let d =
        let desc =
          let x = 
# 659 "asllib/Parser.mly"
                                                D_Pragma  
# 4162 "asllib/Parser.ml"
           (x, e) in
          
# 196 "<standard.mly>"
    ( x )
# 4167 "asllib/Parser.ml"
          
        in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4176 "asllib/Parser.ml"
        
      in
      (
# 662 "asllib/Parser.mly"
     ( [d] )
# 4182 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_086 =
    fun _1 _1_inlined1 _2 _2_inlined1 _6 accessors func_args name params_opt ty ->
      let accessor_body =
        let _2 = _2_inlined1 in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4196 "asllib/Parser.ml"
        
      in
      let override = 
# 123 "<standard.mly>"
    ( None )
# 4202 "asllib/Parser.ml"
       in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4207 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_087 =
    fun _1 _1_inlined1 _2 _6 _endpos__1_inlined1_ _startpos__1_inlined1_ accessors func_args name params_opt ty ->
      let accessor_body =
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 4223 "asllib/Parser.ml"
          
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4229 "asllib/Parser.ml"
        
      in
      let override = 
# 123 "<standard.mly>"
    ( None )
# 4235 "asllib/Parser.ml"
       in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4240 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_088 =
    fun _1 _1_inlined1 _1_inlined2 _2 _2_inlined1 _6 accessors func_args name params_opt ty ->
      let accessor_body =
        let (_2, _1_inlined1, _1) = (_2_inlined1, _1_inlined2, _1_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4254 "asllib/Parser.ml"
        
      in
      let override =
        let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 4261 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 4266 "asllib/Parser.ml"
        
      in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4272 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_089 =
    fun _1 _1_inlined1 _1_inlined2 _2 _6 _endpos__1_inlined2_ _startpos__1_inlined2_ accessors func_args name params_opt ty ->
      let accessor_body =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, _1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2, _1_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 4289 "asllib/Parser.ml"
          
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4295 "asllib/Parser.ml"
        
      in
      let override =
        let x = 
# 591 "asllib/Parser.mly"
              ( Impdef )
# 4302 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 4307 "asllib/Parser.ml"
        
      in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4313 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_090 =
    fun _1 _1_inlined1 _1_inlined2 _2 _2_inlined1 _6 accessors func_args name params_opt ty ->
      let accessor_body =
        let (_2, _1_inlined1, _1) = (_2_inlined1, _1_inlined2, _1_inlined1) in
        let _3 =
          let _1 = _1_inlined1 in
          ()
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4327 "asllib/Parser.ml"
        
      in
      let override =
        let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 4334 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 4339 "asllib/Parser.ml"
        
      in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4345 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_091 =
    fun _1 _1_inlined1 _1_inlined2 _2 _6 _endpos__1_inlined2_ _startpos__1_inlined2_ accessors func_args name params_opt ty ->
      let accessor_body =
        let (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, _1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2, _1_inlined1) in
        let _3 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 4362 "asllib/Parser.ml"
          
        in
        
# 668 "asllib/Parser.mly"
  ( accessors )
# 4368 "asllib/Parser.ml"
        
      in
      let override =
        let x = 
# 592 "asllib/Parser.mly"
                      ( Implementation )
# 4375 "asllib/Parser.ml"
         in
        
# 126 "<standard.mly>"
    ( Some x )
# 4380 "asllib/Parser.ml"
        
      in
      (
# 665 "asllib/Parser.mly"
    ( desugar_accessor_pair override name params_opt func_args ty accessor_body )
# 4386 "asllib/Parser.ml"
       : (AST.decl list))
  
  let _menhir_action_092 =
    fun discard_or_identifier ->
      (
# 463 "asllib/Parser.mly"
                                         LDI_Var  
# 4394 "asllib/Parser.ml"
       discard_or_identifier : (AST.local_decl_item))
  
  let _menhir_action_093 =
    fun x ->
      let _1 = 
# 205 "<standard.mly>"
    ( x )
# 4402 "asllib/Parser.ml"
       in
      let _1_1 = _1 in
      (
# 464 "asllib/Parser.mly"
                                         LDI_Tuple  
# 4408 "asllib/Parser.ml"
       _1_1 : (AST.local_decl_item))
  
  let _menhir_action_094 =
    fun _1 ->
      (
# 482 "asllib/Parser.mly"
                       ( AST.Up )
# 4416 "asllib/Parser.ml"
       : (AST.for_direction))
  
  let _menhir_action_095 =
    fun _1 ->
      (
# 482 "asllib/Parser.mly"
                                            ( AST.Down )
# 4424 "asllib/Parser.ml"
       : (AST.for_direction))
  
  let _menhir_action_096 =
    fun _1 ->
      (
# 440 "asllib/Parser.mly"
                             ( None )
# 4432 "asllib/Parser.ml"
       : (Desugar.lhs_field option))
  
  let _menhir_action_097 =
    fun _endpos_desc_ _startpos_desc_ desc ->
      let _1 =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4443 "asllib/Parser.ml"
        
      in
      let _1_1 = _1 in
      (
# 441 "asllib/Parser.mly"
                               Some  
# 4450 "asllib/Parser.ml"
       _1_1 : (Desugar.lhs_field option))
  
  let _menhir_action_098 =
    fun _1 ->
      (
# 459 "asllib/Parser.mly"
                   ( fresh_var "__ldi_discard" )
# 4458 "asllib/Parser.ml"
       : (string))
  
  let _menhir_action_099 =
    fun _1 ->
      (_1 : (string))
  
  let _menhir_action_100 =
    fun _1 ->
      (
# 436 "asllib/Parser.mly"
                          ( None )
# 4470 "asllib/Parser.ml"
       : (Desugar.lhs_access option))
  
  let _menhir_action_101 =
    fun sliced_basic_lexpr ->
      (
# 437 "asllib/Parser.mly"
                            Some  
# 4478 "asllib/Parser.ml"
       sliced_basic_lexpr : (Desugar.lhs_access option))
  
  let _menhir_action_102 =
    fun _1 _2 ->
      (_2 : (AST.expr))
  
  let _menhir_action_103 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ c e e_else ->
      let desc = 
# 253 "asllib/Parser.mly"
                                                        E_Cond 
# 4490 "asllib/Parser.ml"
       (c, e, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4498 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_104 =
    fun _2 _3 name x ->
      let args = 
# 205 "<standard.mly>"
    ( x )
# 4506 "asllib/Parser.ml"
       in
      (
# 579 "asllib/Parser.mly"
    ( { name; params=[]; args; call_type = ST_Function } )
# 4511 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_105 =
    fun _2 _3 _5 name params ->
      let args = 
# 571 "asllib/Parser.mly"
                     ( [] )
# 4519 "asllib/Parser.ml"
       in
      (
# 581 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 4524 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_106 =
    fun _2 _3 _5 name params x ->
      let args = 
# 205 "<standard.mly>"
    ( x )
# 4532 "asllib/Parser.ml"
       in
      (
# 581 "asllib/Parser.mly"
    ( { name; params; args; call_type = ST_Function } )
# 4537 "asllib/Parser.ml"
       : (AST.call))
  
  let _menhir_action_107 =
    fun _endpos_value_ _startpos_value_ value ->
      let desc = 
# 259 "asllib/Parser.mly"
                                                                    E_Literal             
# 4545 "asllib/Parser.ml"
       value in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_value_, _startpos_value_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4553 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_108 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 260 "asllib/Parser.mly"
                                                                    E_Var                 
# 4563 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4572 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_109 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 218 "asllib/Parser.mly"
                  ( `AND    )
# 4581 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4586 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4595 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_110 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 219 "asllib/Parser.mly"
                  ( `BAND   )
# 4604 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4609 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4618 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_111 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 220 "asllib/Parser.mly"
                  ( `BOR    )
# 4627 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4632 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4641 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_112 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 221 "asllib/Parser.mly"
                  ( `BEQ    )
# 4650 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4655 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4664 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_113 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 222 "asllib/Parser.mly"
                  ( `DIV    )
# 4673 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4678 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4687 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_114 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 223 "asllib/Parser.mly"
                  ( `DIVRM  )
# 4696 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4701 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4710 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_115 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 224 "asllib/Parser.mly"
                  ( `XOR    )
# 4719 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4724 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4733 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_116 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 225 "asllib/Parser.mly"
                  ( `EQ_OP  )
# 4742 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4747 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4756 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_117 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 226 "asllib/Parser.mly"
                  ( `NEQ    )
# 4765 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4770 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4779 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_118 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 227 "asllib/Parser.mly"
                  ( `GT     )
# 4788 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4793 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4802 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_119 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 228 "asllib/Parser.mly"
                  ( `GEQ    )
# 4811 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4816 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4825 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_120 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 229 "asllib/Parser.mly"
                  ( `IMPL   )
# 4834 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4839 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4848 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_121 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 230 "asllib/Parser.mly"
                  ( `LT     )
# 4857 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4862 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4871 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_122 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 231 "asllib/Parser.mly"
                  ( `LEQ    )
# 4880 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4885 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4894 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_123 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 232 "asllib/Parser.mly"
                  ( `PLUS   )
# 4903 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4908 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4917 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_124 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 233 "asllib/Parser.mly"
                  ( `MINUS  )
# 4926 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4931 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4940 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_125 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 234 "asllib/Parser.mly"
                  ( `MOD    )
# 4949 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4954 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4963 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_126 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 235 "asllib/Parser.mly"
                  ( `MUL    )
# 4972 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 4977 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 4986 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_127 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 236 "asllib/Parser.mly"
                  ( `OR     )
# 4995 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5000 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5009 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_128 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 237 "asllib/Parser.mly"
                  ( `RDIV   )
# 5018 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5023 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5032 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_129 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 238 "asllib/Parser.mly"
                  ( `SHL    )
# 5041 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5046 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5055 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_130 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 239 "asllib/Parser.mly"
                  ( `SHR    )
# 5064 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5069 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5078 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_131 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 240 "asllib/Parser.mly"
                  ( `POW    )
# 5087 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5092 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5101 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_132 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 241 "asllib/Parser.mly"
                  ( `BV_CONCAT )
# 5110 "asllib/Parser.ml"
         in
        
# 261 "asllib/Parser.mly"
                                                                e_binop               
# 5115 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5124 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_133 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 213 "asllib/Parser.mly"
            ( BNOT )
# 5133 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                    E_Unop                
# 5138 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5147 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_134 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 214 "asllib/Parser.mly"
            ( NEG )
# 5156 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                    E_Unop                
# 5161 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5170 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_135 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 215 "asllib/Parser.mly"
            ( NOT )
# 5179 "asllib/Parser.ml"
         in
        
# 262 "asllib/Parser.mly"
                                                                    E_Unop                
# 5184 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5193 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_136 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else ->
      let desc = 
# 263 "asllib/Parser.mly"
                                                                    E_Cond                
# 5201 "asllib/Parser.ml"
       (e1, e2, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5209 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_137 =
    fun _endpos_call_ _startpos_call_ call ->
      let desc = 
# 264 "asllib/Parser.mly"
                                                                    e_call                
# 5217 "asllib/Parser.ml"
       call in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_call_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5225 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_138 =
    fun _endpos_slices_ _startpos_e_ e slices ->
      let desc = 
# 265 "asllib/Parser.mly"
                                                                E_Slice               
# 5233 "asllib/Parser.ml"
       (e, slices) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_slices_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5241 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_139 =
    fun _2 _4 _endpos__4_ _startpos_e1_ e1 e2 ->
      let desc = 
# 266 "asllib/Parser.mly"
                                                                E_GetArray            
# 5249 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__4_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5257 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_140 =
    fun _2 _endpos_x_ _startpos_e_ e x ->
      let desc = 
# 267 "asllib/Parser.mly"
                                                                E_GetField            
# 5265 "asllib/Parser.ml"
       (e, x) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_x_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5273 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_141 =
    fun _2 _endpos__3_ _startpos_e_ e x ->
      let desc =
        let fs = 
# 205 "<standard.mly>"
    ( x )
# 5282 "asllib/Parser.ml"
         in
        
# 268 "asllib/Parser.mly"
                                                                E_GetFields           
# 5287 "asllib/Parser.ml"
         (e, fs)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5296 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_142 =
    fun _2 _endpos_ty_ _startpos_expr_ expr ty ->
      let desc = 
# 270 "asllib/Parser.mly"
                                                                E_ATC                 
# 5304 "asllib/Parser.ml"
       (expr, ty) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5312 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_143 =
    fun _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_ constraint_kind expr ->
      let desc =
        let implicit_t_int =
          let desc = 
# 411 "asllib/Parser.mly"
                                                       T_Int 
# 5322 "asllib/Parser.ml"
           constraint_kind in
          let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_constraint_kind_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5330 "asllib/Parser.ml"
          
        in
        
# 271 "asllib/Parser.mly"
                                                                E_ATC                 
# 5336 "asllib/Parser.ml"
         (expr, implicit_t_int)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5345 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_144 =
    fun _2 _endpos_pattern_set_ _startpos_expr_ expr pattern_set ->
      let desc = 
# 273 "asllib/Parser.mly"
                                                                E_Pattern             
# 5353 "asllib/Parser.ml"
       (expr, pattern_set) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_pattern_set_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5361 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_145 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_ expr ->
      let desc =
        let pattern_mask =
          let desc =
            let _1_1 = _1 in
            
# 338 "asllib/Parser.mly"
                                            Pattern_Mask  
# 5373 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5382 "asllib/Parser.ml"
          
        in
        
# 274 "asllib/Parser.mly"
                                                                E_Pattern             
# 5388 "asllib/Parser.ml"
         (expr, pattern_mask)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_expr_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5397 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_146 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e ->
      let desc =
        let p =
          let desc =
            let _1_1 = _1 in
            
# 338 "asllib/Parser.mly"
                                            Pattern_Mask  
# 5409 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5418 "asllib/Parser.ml"
          
        in
        
# 275 "asllib/Parser.mly"
                                                              ( E_Pattern (e, Pattern_Not (p) |> add_pos_from p) )
# 5424 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5433 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_147 =
    fun _1 _2 _endpos_ty_ _startpos__1_ ty ->
      let desc = 
# 276 "asllib/Parser.mly"
                                                                E_Arbitrary         
# 5441 "asllib/Parser.ml"
       ty in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5449 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_148 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let e = 
# 205 "<standard.mly>"
    ( x )
# 5458 "asllib/Parser.ml"
         in
        
# 277 "asllib/Parser.mly"
                                                              ( E_Tuple [ e ]        )
# 5463 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5472 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_149 =
    fun _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc ->
      let desc =
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5484 "asllib/Parser.ml"
          
        in
        
# 281 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), []) )
# 5490 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5499 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_150 =
    fun _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let fields = 
# 205 "<standard.mly>"
    ( x )
# 5508 "asllib/Parser.ml"
         in
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5516 "asllib/Parser.ml"
          
        in
        
# 283 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), fields) )
# 5522 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5531 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_151 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let _1 = 
# 205 "<standard.mly>"
    ( x )
# 5540 "asllib/Parser.ml"
         in
        let _1_1 = _1 in
        
# 285 "asllib/Parser.mly"
                                                                    E_Tuple               
# 5546 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5555 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_152 =
    fun _endpos_value_ _startpos_value_ value ->
      let desc = 
# 310 "asllib/Parser.mly"
                                                                    E_Literal             
# 5563 "asllib/Parser.ml"
       value in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_value_, _startpos_value_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5571 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_153 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 311 "asllib/Parser.mly"
                                                                    E_Var                 
# 5581 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5590 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_154 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 218 "asllib/Parser.mly"
                  ( `AND    )
# 5599 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5604 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5613 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_155 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 219 "asllib/Parser.mly"
                  ( `BAND   )
# 5622 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5627 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5636 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_156 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 220 "asllib/Parser.mly"
                  ( `BOR    )
# 5645 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5650 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5659 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_157 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 221 "asllib/Parser.mly"
                  ( `BEQ    )
# 5668 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5673 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5682 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_158 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 222 "asllib/Parser.mly"
                  ( `DIV    )
# 5691 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5696 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5705 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_159 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 223 "asllib/Parser.mly"
                  ( `DIVRM  )
# 5714 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5719 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5728 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_160 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 224 "asllib/Parser.mly"
                  ( `XOR    )
# 5737 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5742 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5751 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_161 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 225 "asllib/Parser.mly"
                  ( `EQ_OP  )
# 5760 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5765 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5774 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_162 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 226 "asllib/Parser.mly"
                  ( `NEQ    )
# 5783 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5788 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5797 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_163 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 227 "asllib/Parser.mly"
                  ( `GT     )
# 5806 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5811 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5820 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_164 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 228 "asllib/Parser.mly"
                  ( `GEQ    )
# 5829 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5834 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5843 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_165 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 229 "asllib/Parser.mly"
                  ( `IMPL   )
# 5852 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5857 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5866 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_166 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 230 "asllib/Parser.mly"
                  ( `LT     )
# 5875 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5880 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5889 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_167 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 231 "asllib/Parser.mly"
                  ( `LEQ    )
# 5898 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5903 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5912 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_168 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 232 "asllib/Parser.mly"
                  ( `PLUS   )
# 5921 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5926 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5935 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_169 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 233 "asllib/Parser.mly"
                  ( `MINUS  )
# 5944 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5949 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5958 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_170 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 234 "asllib/Parser.mly"
                  ( `MOD    )
# 5967 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5972 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 5981 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_171 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 235 "asllib/Parser.mly"
                  ( `MUL    )
# 5990 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 5995 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6004 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_172 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 236 "asllib/Parser.mly"
                  ( `OR     )
# 6013 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6018 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6027 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_173 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 237 "asllib/Parser.mly"
                  ( `RDIV   )
# 6036 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6041 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6050 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_174 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 238 "asllib/Parser.mly"
                  ( `SHL    )
# 6059 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6064 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6073 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_175 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 239 "asllib/Parser.mly"
                  ( `SHR    )
# 6082 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6087 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6096 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_176 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 240 "asllib/Parser.mly"
                  ( `POW    )
# 6105 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6110 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6119 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_177 =
    fun _1 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc =
        let op = 
# 241 "asllib/Parser.mly"
                  ( `BV_CONCAT )
# 6128 "asllib/Parser.ml"
         in
        
# 312 "asllib/Parser.mly"
                                                                        e_binop               
# 6133 "asllib/Parser.ml"
         (e1, op, e2)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6142 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_178 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 213 "asllib/Parser.mly"
            ( BNOT )
# 6151 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                    E_Unop                
# 6156 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6165 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_179 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 214 "asllib/Parser.mly"
            ( NEG )
# 6174 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                    E_Unop                
# 6179 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6188 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_180 =
    fun _1 _endpos_e_ _startpos__1_ e ->
      let desc =
        let op = 
# 215 "asllib/Parser.mly"
            ( NOT )
# 6197 "asllib/Parser.ml"
         in
        
# 313 "asllib/Parser.mly"
                                                                    E_Unop                
# 6202 "asllib/Parser.ml"
         (op, e)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6211 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_181 =
    fun _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else ->
      let desc = 
# 314 "asllib/Parser.mly"
                                                                    E_Cond                
# 6219 "asllib/Parser.ml"
       (e1, e2, e_else) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e_else_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6227 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_182 =
    fun _endpos_call_ _startpos_call_ call ->
      let desc = 
# 315 "asllib/Parser.mly"
                                                                    e_call                
# 6235 "asllib/Parser.ml"
       call in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_call_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6243 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_183 =
    fun _endpos_slices_ _startpos_e_ e slices ->
      let desc = 
# 316 "asllib/Parser.mly"
                                                                        E_Slice               
# 6251 "asllib/Parser.ml"
       (e, slices) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_slices_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6259 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_184 =
    fun _2 _4 _endpos__4_ _startpos_e1_ e1 e2 ->
      let desc = 
# 317 "asllib/Parser.mly"
                                                                        E_GetArray            
# 6267 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__4_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6275 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_185 =
    fun _2 _endpos_x_ _startpos_e_ e x ->
      let desc = 
# 318 "asllib/Parser.mly"
                                                                        E_GetField            
# 6283 "asllib/Parser.ml"
       (e, x) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_x_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6291 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_186 =
    fun _2 _endpos__3_ _startpos_e_ e x ->
      let desc =
        let fs = 
# 205 "<standard.mly>"
    ( x )
# 6300 "asllib/Parser.ml"
         in
        
# 319 "asllib/Parser.mly"
                                                                        E_GetFields           
# 6305 "asllib/Parser.ml"
         (e, fs)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6314 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_187 =
    fun _2 _endpos_ty_ _startpos_expr_pattern_ expr_pattern ty ->
      let desc = 
# 321 "asllib/Parser.mly"
                                                                        E_ATC                 
# 6322 "asllib/Parser.ml"
       (expr_pattern, ty) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6330 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_188 =
    fun _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_pattern_ constraint_kind expr_pattern ->
      let desc =
        let implicit_t_int =
          let desc = 
# 411 "asllib/Parser.mly"
                                                       T_Int 
# 6340 "asllib/Parser.ml"
           constraint_kind in
          let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_constraint_kind_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6348 "asllib/Parser.ml"
          
        in
        
# 322 "asllib/Parser.mly"
                                                                        E_ATC                 
# 6354 "asllib/Parser.ml"
         (expr_pattern, implicit_t_int)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_constraint_kind_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6363 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_189 =
    fun _2 _endpos_pattern_set_ _startpos_expr_pattern_ expr_pattern pattern_set ->
      let desc = 
# 324 "asllib/Parser.mly"
                                                                        E_Pattern             
# 6371 "asllib/Parser.ml"
       (expr_pattern, pattern_set) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_pattern_set_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6379 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_190 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_pattern_ expr_pattern ->
      let desc =
        let pattern_mask =
          let desc =
            let _1_1 = _1 in
            
# 338 "asllib/Parser.mly"
                                            Pattern_Mask  
# 6391 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6400 "asllib/Parser.ml"
          
        in
        
# 325 "asllib/Parser.mly"
                                                                        E_Pattern             
# 6406 "asllib/Parser.ml"
         (expr_pattern, pattern_mask)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6415 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_191 =
    fun _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e ->
      let desc =
        let p =
          let desc =
            let _1_1 = _1 in
            
# 338 "asllib/Parser.mly"
                                            Pattern_Mask  
# 6427 "asllib/Parser.ml"
             _1_1
          in
          let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6436 "asllib/Parser.ml"
          
        in
        
# 326 "asllib/Parser.mly"
                                                                      ( E_Pattern (e, Pattern_Not (p) |> add_pos_from p) )
# 6442 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos_e_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6451 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_192 =
    fun _1 _2 _endpos_ty_ _startpos__1_ ty ->
      let desc = 
# 328 "asllib/Parser.mly"
                                                                        E_Arbitrary         
# 6459 "asllib/Parser.ml"
       ty in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_ty_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6467 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_193 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let e = 
# 205 "<standard.mly>"
    ( x )
# 6476 "asllib/Parser.ml"
         in
        
# 329 "asllib/Parser.mly"
                                                                      ( E_Tuple [ e ]        )
# 6481 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6490 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_194 =
    fun _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc ->
      let desc =
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6502 "asllib/Parser.ml"
          
        in
        
# 333 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), []) )
# 6508 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6517 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_195 =
    fun _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let fields = 
# 205 "<standard.mly>"
    ( x )
# 6526 "asllib/Parser.ml"
         in
        let t =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6534 "asllib/Parser.ml"
          
        in
        
# 335 "asllib/Parser.mly"
        ( E_Record (add_pos_from t (T_Named t.desc), fields) )
# 6540 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6549 "asllib/Parser.ml"
       : (AST.expr))
  
  let _menhir_action_196 =
    fun x y ->
      (
# 180 "<standard.mly>"
    ( (x, y) )
# 6557 "asllib/Parser.ml"
       : (string * AST.expr))
  
  let _menhir_action_197 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6565 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_198 =
    fun () ->
      (
# 360 "asllib/Parser.mly"
                  ( [] )
# 6573 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_199 =
    fun _1 ->
      (_1 : (AST.field list))
  
  let _menhir_action_200 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 6585 "asllib/Parser.ml"
       : (AST.typed_identifier list))
  
  let _menhir_action_201 =
    fun _1 ->
      (
# 477 "asllib/Parser.mly"
                ( GDK_Let      )
# 6593 "asllib/Parser.ml"
       : (AST.global_decl_keyword))
  
  let _menhir_action_202 =
    fun _1 ->
      (
# 478 "asllib/Parser.mly"
                ( GDK_Constant )
# 6601 "asllib/Parser.ml"
       : (AST.global_decl_keyword))
  
  let _menhir_action_203 =
    fun _1 ->
      (
# 587 "asllib/Parser.mly"
           ( global_ignored () )
# 6609 "asllib/Parser.ml"
       : (string))
  
  let _menhir_action_204 =
    fun _1 ->
      (_1 : (string))
  
  let _menhir_action_205 =
    fun expr ->
      (
# 302 "asllib/Parser.mly"
                                  Constraint_Exact  
# 6621 "asllib/Parser.ml"
       expr : (AST.int_constraint))
  
  let _menhir_action_206 =
    fun _2 e1 e2 ->
      (
# 303 "asllib/Parser.mly"
                                  Constraint_Range  
# 6629 "asllib/Parser.ml"
       (e1, e2) : (AST.int_constraint))
  
  let _menhir_action_207 =
    fun sliced_basic_lexpr ->
      (
# 444 "asllib/Parser.mly"
                            desugar_lhs_access  
# 6637 "asllib/Parser.ml"
       sliced_basic_lexpr : (AST.lexpr))
  
  let _menhir_action_208 =
    fun _endpos__3_ _startpos__1_ x ->
      let _1 =
        let desc = 
# 205 "<standard.mly>"
    ( x )
# 6646 "asllib/Parser.ml"
         in
        let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6654 "asllib/Parser.ml"
        
      in
      let _1_1 = _1 in
      (
# 445 "asllib/Parser.mly"
                                                          desugar_lhs_tuple  
# 6661 "asllib/Parser.ml"
       _1_1 : (AST.lexpr))
  
  let _menhir_action_209 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 447 "asllib/Parser.mly"
             ( LE_Discard )
# 6669 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6677 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_210 =
    fun _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let flds = 
# 205 "<standard.mly>"
    ( x )
# 6686 "asllib/Parser.ml"
         in
        let x =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6694 "asllib/Parser.ml"
          
        in
        
# 449 "asllib/Parser.mly"
      ( LE_SetFields (le_var x, flds, []) )
# 6700 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6709 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_211 =
    fun _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x ->
      let desc =
        let flds = 
# 205 "<standard.mly>"
    ( x )
# 6718 "asllib/Parser.ml"
         in
        let x =
          let _endpos = _endpos_desc_ in
          let _symbolstartpos = _startpos_desc_ in
          
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6726 "asllib/Parser.ml"
          
        in
        
# 451 "asllib/Parser.mly"
      ( desugar_lhs_fields_tuple x flds )
# 6732 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__3_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6741 "asllib/Parser.ml"
       : (AST.lexpr))
  
  let _menhir_action_212 =
    fun () ->
      (
# 216 "<standard.mly>"
    ( [] )
# 6749 "asllib/Parser.ml"
       : (AST.decl list list))
  
  let _menhir_action_213 =
    fun x xs ->
      (
# 219 "<standard.mly>"
    ( x :: xs )
# 6757 "asllib/Parser.ml"
       : (AST.decl list list))
  
  let _menhir_action_214 =
    fun x ->
      (
# 188 "asllib/Parser.mly"
         ( [ x ] )
# 6765 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated list))
  
  let _menhir_action_215 =
    fun l x ->
      (
# 189 "asllib/Parser.mly"
                     ( x :: l )
# 6773 "asllib/Parser.ml"
       : (AST.case_alt_desc AST.annotated list))
  
  let _menhir_action_216 =
    fun x ->
      (
# 188 "asllib/Parser.mly"
         ( [ x ] )
# 6781 "asllib/Parser.ml"
       : (AST.catcher list))
  
  let _menhir_action_217 =
    fun l x ->
      (
# 189 "asllib/Parser.mly"
                     ( x :: l )
# 6789 "asllib/Parser.ml"
       : (AST.catcher list))
  
  let _menhir_action_218 =
    fun x ->
      (
# 188 "asllib/Parser.mly"
         ( [ x ] )
# 6797 "asllib/Parser.ml"
       : (AST.stmt list))
  
  let _menhir_action_219 =
    fun l x ->
      (
# 189 "asllib/Parser.mly"
                     ( x :: l )
# 6805 "asllib/Parser.ml"
       : (AST.stmt list))
  
  let _menhir_action_220 =
    fun _1 ->
      (
# 470 "asllib/Parser.mly"
                ( LDK_Let       )
# 6813 "asllib/Parser.ml"
       : (AST.local_decl_keyword))
  
  let _menhir_action_221 =
    fun _1 ->
      (
# 471 "asllib/Parser.mly"
                ( LDK_Constant  )
# 6821 "asllib/Parser.ml"
       : (AST.local_decl_keyword))
  
  let _menhir_action_222 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 6829 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_223 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 6838 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_224 =
    fun _1 ->
      (_1 : (AST.stmt))
  
  let _menhir_action_225 =
    fun _endpos__0_ ->
      let desc = 
# 583 "asllib/Parser.mly"
                                                   ( S_Pass )
# 6850 "asllib/Parser.ml"
       in
      let _endpos_desc_ = _endpos__0_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _endpos in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6858 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_226 =
    fun () ->
      (
# 428 "asllib/Parser.mly"
    ( [] )
# 6866 "asllib/Parser.ml"
       : (Desugar.lhs_field list))
  
  let _menhir_action_227 =
    fun _1 _endpos_desc_ _startpos_desc_ desc t ->
      let h =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 6877 "asllib/Parser.ml"
        
      in
      (
# 429 "asllib/Parser.mly"
                                                   ( h :: t )
# 6883 "asllib/Parser.ml"
       : (Desugar.lhs_field list))
  
  let _menhir_action_228 =
    fun _2 body ->
      (
# 675 "asllib/Parser.mly"
    (
      [
        D_Func
          {
            name = "main";
            args = [];
            parameters = [];
            body = SB_ASL body;
            return_type = None;
            subprogram_type = ST_Procedure;
            recurse_limit = None;
            override = None;
            builtin = false;
          }
        |> ASTUtils.add_pos_from body
      ]
    )
# 6907 "asllib/Parser.ml"
       : (AST.t))
  
  let _menhir_action_229 =
    fun x ->
      let y = 
# 123 "<standard.mly>"
    ( None )
# 6915 "asllib/Parser.ml"
       in
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 6920 "asllib/Parser.ml"
       : (string * AST.ty option))
  
  let _menhir_action_230 =
    fun x x_inlined1 ->
      let y =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 6930 "asllib/Parser.ml"
        
      in
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 6936 "asllib/Parser.ml"
       : (string * AST.ty option))
  
  let _menhir_action_231 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 6944 "asllib/Parser.ml"
       : (AST.ty option))
  
  let _menhir_action_232 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 6952 "asllib/Parser.ml"
       : (AST.ty option))
  
  let _menhir_action_233 =
    fun () ->
      (
# 111 "<standard.mly>"
    ( None )
# 6960 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_234 =
    fun x ->
      (
# 114 "<standard.mly>"
    ( Some x )
# 6968 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_235 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 6976 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_236 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 6985 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_237 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 6993 "asllib/Parser.ml"
       : (AST.stmt option))
  
  let _menhir_action_238 =
    fun _1 _2 _3 ->
      let x = _3 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 7002 "asllib/Parser.ml"
       : (AST.stmt option))
  
  let _menhir_action_239 =
    fun () ->
      (
# 569 "asllib/Parser.mly"
                  ( [] )
# 7010 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_240 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 7018 "asllib/Parser.ml"
       : ((string * AST.ty option) list))
  
  let _menhir_action_241 =
    fun _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern ->
      let desc = 
# 343 "asllib/Parser.mly"
                        Pattern_Single  
# 7026 "asllib/Parser.ml"
       expr_pattern in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_pattern_, _startpos_expr_pattern_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7034 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_242 =
    fun _2 _endpos_e2_ _startpos_e1_ e1 e2 ->
      let desc = 
# 344 "asllib/Parser.mly"
                                           Pattern_Range  
# 7042 "asllib/Parser.ml"
       (e1, e2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_e2_, _startpos_e1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7050 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_243 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 345 "asllib/Parser.mly"
             ( Pattern_All )
# 7058 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7066 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_244 =
    fun _1 _endpos_expr_ _startpos__1_ expr ->
      let desc = 
# 346 "asllib/Parser.mly"
                     Pattern_Leq  
# 7074 "asllib/Parser.ml"
       expr in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7082 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_245 =
    fun _1 _endpos_expr_ _startpos__1_ expr ->
      let desc = 
# 347 "asllib/Parser.mly"
                     Pattern_Geq  
# 7090 "asllib/Parser.ml"
       expr in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_expr_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7098 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_246 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let _1 = 
# 205 "<standard.mly>"
    ( x )
# 7107 "asllib/Parser.ml"
         in
        let _1_1 = _1 in
        
# 348 "asllib/Parser.mly"
                           Pattern_Tuple  
# 7113 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7122 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_247 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 338 "asllib/Parser.mly"
                                            Pattern_Mask  
# 7132 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7141 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_248 =
    fun _1 ->
      (_1 : (AST.pattern))
  
  let _menhir_action_249 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc =
        let _1_1 = _1 in
        
# 339 "asllib/Parser.mly"
                                                   Pattern_Any  
# 7155 "asllib/Parser.ml"
         _1_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7164 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_250 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 7172 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_251 =
    fun _1 _endpos__3_ _startpos__1_ x ->
      let desc =
        let _2 = 
# 205 "<standard.mly>"
    ( x )
# 7181 "asllib/Parser.ml"
         in
        let _2_1 = _2 in
        
# 356 "asllib/Parser.mly"
                                      Pattern_Not  
# 7187 "asllib/Parser.ml"
         _2_1
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7196 "asllib/Parser.ml"
       : (AST.pattern))
  
  let _menhir_action_252 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 7204 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_253 =
    fun _1 _2 ->
      let x = _2 in
      (
# 126 "<standard.mly>"
    ( Some x )
# 7213 "asllib/Parser.ml"
       : (AST.expr option))
  
  let _menhir_action_254 =
    fun _1 _2 ->
      (_2 : (AST.ty))
  
  let _menhir_action_255 =
    fun _1 _3 _endpos_s2_ _startpos__1_ e s1 s2 ->
      let desc = 
# 551 "asllib/Parser.mly"
                                                     S_Cond 
# 7225 "asllib/Parser.ml"
       (e, s1, s2) in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_s2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = if _startpos_desc_ != _endpos_desc_ then
        _startpos_desc_
      else
        _endpos in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7236 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_256 =
    fun _endpos__0_ ->
      let desc = 
# 480 "asllib/Parser.mly"
            ( S_Pass )
# 7244 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__0_, _endpos__0_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = if _startpos_desc_ != _endpos_desc_ then
        _startpos_desc_
      else
        _endpos in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7255 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_257 =
    fun _1 _2 ->
      (_2 : (AST.stmt))
  
  let _menhir_action_258 =
    fun expr ->
      (
# 365 "asllib/Parser.mly"
                                    Slice_Single   
# 7267 "asllib/Parser.ml"
       expr : (AST.slice))
  
  let _menhir_action_259 =
    fun _2 e1 e2 ->
      (
# 366 "asllib/Parser.mly"
                                    Slice_Range    
# 7275 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_260 =
    fun _2 e1 e2 ->
      (
# 367 "asllib/Parser.mly"
                                    Slice_Length   
# 7283 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_261 =
    fun _1 e ->
      (
# 368 "asllib/Parser.mly"
                                  ( Slice_Length(zero, e) )
# 7291 "asllib/Parser.ml"
       : (AST.slice))
  
  let _menhir_action_262 =
    fun _2 e1 e2 ->
      (
# 369 "asllib/Parser.mly"
                                    Slice_Star     
# 7299 "asllib/Parser.ml"
       (e1, e2) : (AST.slice))
  
  let _menhir_action_263 =
    fun basic_lexpr ->
      (basic_lexpr : (Desugar.lhs_access))
  
  let _menhir_action_264 =
    fun _endpos_desc_ _startpos_desc_ basic desc ->
      let slices =
        let _endpos = _endpos_desc_ in
        let _symbolstartpos = _startpos_desc_ in
        
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7314 "asllib/Parser.ml"
        
      in
      (
# 433 "asllib/Parser.mly"
                                                 ( {basic with slices} )
# 7320 "asllib/Parser.ml"
       : (Desugar.lhs_access))
  
  let _menhir_action_265 =
    fun x ->
      (
# 205 "<standard.mly>"
    ( x )
# 7328 "asllib/Parser.ml"
       : (AST.slice list))
  
  let _menhir_action_266 =
    fun x ->
      let _1 = 
# 196 "<standard.mly>"
    ( x )
# 7336 "asllib/Parser.ml"
       in
      let _1_1 = _1 in
      (
# 671 "asllib/Parser.mly"
                                             List.concat  
# 7342 "asllib/Parser.ml"
       _1_1 : (AST.t))
  
  let _menhir_action_267 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ e s1 s2 ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 506 "asllib/Parser.mly"
                                                       S_Cond 
# 7355 "asllib/Parser.ml"
         (e, s1, s2) in
        
# 196 "<standard.mly>"
    ( x )
# 7360 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7369 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_268 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ e s1 s2 ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7385 "asllib/Parser.ml"
          
        in
        let x = 
# 506 "asllib/Parser.mly"
                                                       S_Cond 
# 7391 "asllib/Parser.ml"
         (e, s1, s2) in
        
# 196 "<standard.mly>"
    ( x )
# 7396 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7405 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_269 =
    fun _1 _1_inlined1 _1_inlined2 _2 _3 _endpos__2_ _startpos__1_ expr ->
      let desc =
        let _2 =
          let _1 = _1_inlined2 in
          ()
        in
        let x =
          let alt = _1_inlined1 in
          
# 508 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt (S_Unreachable |> ASTUtils.add_pos_from expr))
# 7420 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7426 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7435 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_270 =
    fun _1 _1_inlined1 _1_inlined2 _3 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ expr ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined2_, _startpos__1_inlined2_, _1_inlined2) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7451 "asllib/Parser.ml"
          
        in
        let x =
          let alt = _1_inlined1 in
          
# 508 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt (S_Unreachable |> ASTUtils.add_pos_from expr))
# 7459 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7465 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7474 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_271 =
    fun _1 _1_inlined1 _1_inlined2 _1_inlined3 _2 _2_inlined1 _3 _endpos__2_ _startpos__1_ expr otherwise_stmt ->
      let desc =
        let _2 =
          let _1 = _1_inlined3 in
          ()
        in
        let x =
          let _2 = _2_inlined1 in
          let case_otherwise =
            let _1 = _1_inlined2 in
            
# 492 "asllib/Parser.mly"
      ( otherwise_stmt )
# 7491 "asllib/Parser.ml"
            
          in
          let alt = _1_inlined1 in
          
# 510 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt case_otherwise )
# 7498 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7504 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7513 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_272 =
    fun _1 _1_inlined1 _1_inlined2 _1_inlined3 _2_inlined1 _3 _endpos__1_inlined3_ _startpos__1_ _startpos__1_inlined3_ expr otherwise_stmt ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined3_, _startpos__1_inlined3_, _1_inlined3) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7529 "asllib/Parser.ml"
          
        in
        let x =
          let _2 = _2_inlined1 in
          let case_otherwise =
            let _1 = _1_inlined2 in
            
# 492 "asllib/Parser.mly"
      ( otherwise_stmt )
# 7539 "asllib/Parser.ml"
            
          in
          let alt = _1_inlined1 in
          
# 510 "asllib/Parser.mly"
          ( desugar_case_stmt expr alt case_otherwise )
# 7546 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7552 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7561 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_273 =
    fun _1 _1_inlined1 _2 _4 _endpos__2_ _startpos__1_ expr loop_limit stmt_list ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 511 "asllib/Parser.mly"
                                                       S_While 
# 7574 "asllib/Parser.ml"
         (expr, loop_limit, stmt_list) in
        
# 196 "<standard.mly>"
    ( x )
# 7579 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7588 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_274 =
    fun _1 _1_inlined1 _4 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ expr loop_limit stmt_list ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7604 "asllib/Parser.ml"
          
        in
        let x = 
# 511 "asllib/Parser.mly"
                                                       S_While 
# 7610 "asllib/Parser.ml"
         (expr, loop_limit, stmt_list) in
        
# 196 "<standard.mly>"
    ( x )
# 7615 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7624 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_275 =
    fun _1 _1_inlined1 _2 _3 _8 _endpos__2_ _startpos__1_ body dir end_e index_name limit start_e ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 514 "asllib/Parser.mly"
          ( S_For { index_name; start_e; end_e; dir; body; limit } )
# 7637 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7642 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7651 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_276 =
    fun _1 _1_inlined1 _3 _8 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ body dir end_e index_name limit start_e ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7667 "asllib/Parser.ml"
          
        in
        let x = 
# 514 "asllib/Parser.mly"
          ( S_For { index_name; start_e; end_e; dir; body; limit } )
# 7673 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7678 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7687 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_277 =
    fun _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ c o s ->
      let desc =
        let _2 =
          let _1 = _1_inlined1 in
          ()
        in
        let x = 
# 515 "asllib/Parser.mly"
                                                                      S_Try  
# 7700 "asllib/Parser.ml"
         (s, c, o) in
        
# 196 "<standard.mly>"
    ( x )
# 7705 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7714 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_278 =
    fun _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ c o s ->
      let desc =
        let _2 =
          let (_endpos__1_, _startpos__1_, _1) = (_endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) in
          let _endpos = _endpos__1_ in
          let _startpos = _startpos__1_ in
          
# 193 "asllib/Parser.mly"
                          (
      if not Config.allow_no_end_semicolon then
          Error.fatal_here _startpos _endpos @@ Error.ObsoleteSyntax "Missing ';' after 'end' keyword.";
  )
# 7730 "asllib/Parser.ml"
          
        in
        let x = 
# 515 "asllib/Parser.mly"
                                                                      S_Try  
# 7736 "asllib/Parser.ml"
         (s, c, o) in
        
# 196 "<standard.mly>"
    ( x )
# 7741 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_inlined1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7750 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_279 =
    fun _1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x = 
# 480 "asllib/Parser.mly"
            ( S_Pass )
# 7759 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 7764 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7773 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_280 =
    fun _1 _2_inlined1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let _2_1 = _2 in
          
# 519 "asllib/Parser.mly"
                                                              S_Return  
# 7785 "asllib/Parser.ml"
           _2_1
        in
        
# 196 "<standard.mly>"
    ( x )
# 7791 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7800 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_281 =
    fun _endpos__2_ _startpos_call_ call ->
      let desc =
        let x = 
# 520 "asllib/Parser.mly"
                                                               s_call  
# 7809 "asllib/Parser.ml"
         call in
        
# 196 "<standard.mly>"
    ( x )
# 7814 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_call_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7823 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_282 =
    fun _1 _endpos__2_ _startpos__1_ e ->
      let desc =
        let x = 
# 521 "asllib/Parser.mly"
                                                               S_Assert  
# 7832 "asllib/Parser.ml"
         e in
        
# 196 "<standard.mly>"
    ( x )
# 7837 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7846 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_283 =
    fun _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x ->
      let desc =
        let x =
          let _5 = 
# 149 "asllib/Parser.mly"
                        Some 
# 7856 "asllib/Parser.ml"
           x in
          let ty_opt = 
# 123 "<standard.mly>"
    ( None )
# 7861 "asllib/Parser.ml"
           in
          let _5_1 = _5 in
          
# 522 "asllib/Parser.mly"
                                                                                 S_Decl    
# 7867 "asllib/Parser.ml"
           (local_decl_keyword_non_var, decl_item, ty_opt, _5_1)
        in
        
# 196 "<standard.mly>"
    ( x )
# 7873 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_local_decl_keyword_non_var_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7882 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_284 =
    fun _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x x_inlined1 ->
      let desc =
        let x =
          let _5 =
            let x = x_inlined1 in
            
# 149 "asllib/Parser.mly"
                        Some 
# 7894 "asllib/Parser.ml"
             x
          in
          let ty_opt = 
# 126 "<standard.mly>"
    ( Some x )
# 7900 "asllib/Parser.ml"
           in
          let _5_1 = _5 in
          
# 522 "asllib/Parser.mly"
                                                                                 S_Decl    
# 7906 "asllib/Parser.ml"
           (local_decl_keyword_non_var, decl_item, ty_opt, _5_1)
        in
        
# 196 "<standard.mly>"
    ( x )
# 7912 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_local_decl_keyword_non_var_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7921 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_285 =
    fun _2_inlined1 _endpos__2_ _startpos_le_ e le ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          
# 523 "asllib/Parser.mly"
                                                               S_Assign  
# 7932 "asllib/Parser.ml"
           (le, e)
        in
        
# 196 "<standard.mly>"
    ( x )
# 7938 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_le_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7947 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_286 =
    fun _2_inlined1 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7961 "asllib/Parser.ml"
            
          in
          
# 525 "asllib/Parser.mly"
        ( desugar_setter call [] rhs )
# 7967 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 7973 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7982 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_287 =
    fun _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc fld rhs ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 7996 "asllib/Parser.ml"
            
          in
          
# 527 "asllib/Parser.mly"
        ( desugar_setter call [fld] rhs )
# 8002 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8008 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8017 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_288 =
    fun _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs x ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let flds = 
# 205 "<standard.mly>"
    ( x )
# 8028 "asllib/Parser.ml"
           in
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8036 "asllib/Parser.ml"
            
          in
          
# 529 "asllib/Parser.mly"
        ( desugar_setter call flds rhs )
# 8042 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8048 "asllib/Parser.ml"
        
      in
      let _endpos_desc_ = _endpos__2_ in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8057 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_289 =
    fun _4 _endpos__2_ _endpos_desc_ _startpos_desc_ _startpos_ldk_ desc ldk lhs ty ->
      let desc =
        let x =
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8070 "asllib/Parser.ml"
            
          in
          
# 531 "asllib/Parser.mly"
        ( desugar_elided_parameter ldk lhs ty call)
# 8076 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8082 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos_ldk_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8091 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_290 =
    fun _1 _endpos__2_ _startpos__1_ e ldi ->
      let desc =
        let x =
          let ty = 
# 123 "<standard.mly>"
    ( None )
# 8101 "asllib/Parser.ml"
           in
          
# 532 "asllib/Parser.mly"
                                                              ( S_Decl (LDK_Var, ldi, ty, e) )
# 8106 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8112 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8121 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_291 =
    fun _1 _endpos__2_ _startpos__1_ e ldi x ->
      let desc =
        let x =
          let ty = 
# 126 "<standard.mly>"
    ( Some x )
# 8131 "asllib/Parser.ml"
           in
          
# 532 "asllib/Parser.mly"
                                                              ( S_Decl (LDK_Var, ldi, ty, e) )
# 8136 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8142 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8151 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_292 =
    fun _1 _2_inlined1 _endpos__2_ _startpos__1_ as_ty ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          let _2_1 = _2 in
          
# 533 "asllib/Parser.mly"
                                                               make_ldi_vars  
# 8163 "asllib/Parser.ml"
           (_2_1, as_ty)
        in
        
# 196 "<standard.mly>"
    ( x )
# 8169 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8178 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_293 =
    fun _1 _4 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc lhs ty ->
      let desc =
        let x =
          let call =
            let _endpos = _endpos_desc_ in
            let _symbolstartpos = _startpos_desc_ in
            
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8191 "asllib/Parser.ml"
            
          in
          
# 535 "asllib/Parser.mly"
        ( desugar_elided_parameter LDK_Var lhs ty call)
# 8197 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8203 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8212 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_294 =
    fun _1 _endpos__2_ _startpos__1_ x ->
      let desc =
        let x =
          let args = 
# 205 "<standard.mly>"
    ( x )
# 8222 "asllib/Parser.ml"
           in
          
# 536 "asllib/Parser.mly"
                                                              ( S_Print { args; newline = true; debug = false } )
# 8227 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8233 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8242 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_295 =
    fun _1 _endpos__2_ _startpos__1_ x ->
      let desc =
        let x =
          let args = 
# 205 "<standard.mly>"
    ( x )
# 8252 "asllib/Parser.ml"
           in
          
# 537 "asllib/Parser.mly"
                                                              ( S_Print { args; newline = false; debug = false } )
# 8257 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8263 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8272 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_296 =
    fun _1 _endpos__2_ _startpos__1_ x ->
      let desc =
        let x =
          let args = 
# 205 "<standard.mly>"
    ( x )
# 8282 "asllib/Parser.ml"
           in
          
# 538 "asllib/Parser.mly"
                                             ( S_Print { args; newline = true; debug = true } )
# 8287 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8293 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8302 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_297 =
    fun _1 _2_inlined1 _3 _endpos__2_ _startpos__1_ ->
      let desc =
        let x =
          let _2 = _2_inlined1 in
          
# 539 "asllib/Parser.mly"
                                                             ( S_Unreachable )
# 8313 "asllib/Parser.ml"
          
        in
        
# 196 "<standard.mly>"
    ( x )
# 8319 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8328 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_298 =
    fun _1 _3 _endpos__2_ _startpos__1_ expr loop_limit stmt_list ->
      let desc =
        let x = 
# 540 "asllib/Parser.mly"
                                                               S_Repeat  
# 8337 "asllib/Parser.ml"
         (stmt_list, expr, loop_limit) in
        
# 196 "<standard.mly>"
    ( x )
# 8342 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8351 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_299 =
    fun _1 _endpos__2_ _startpos__1_ e ->
      let desc =
        let x = 
# 541 "asllib/Parser.mly"
                                                             ( S_Throw (Some (e, None)) )
# 8360 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 8365 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8374 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_300 =
    fun _1 _endpos__2_ _startpos__1_ ->
      let desc =
        let x = 
# 542 "asllib/Parser.mly"
                                                             ( S_Throw None             )
# 8383 "asllib/Parser.ml"
         in
        
# 196 "<standard.mly>"
    ( x )
# 8388 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8397 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_301 =
    fun _1 _endpos__2_ _startpos__1_ e x ->
      let desc =
        let x = 
# 543 "asllib/Parser.mly"
                                                                S_Pragma  
# 8406 "asllib/Parser.ml"
         (x, e) in
        
# 196 "<standard.mly>"
    ( x )
# 8411 "asllib/Parser.ml"
        
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__2_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8420 "asllib/Parser.ml"
       : (AST.stmt_desc AST.annotated))
  
  let _menhir_action_302 =
    fun _1 ->
      let _1_1 = _1 in
      (
# 547 "asllib/Parser.mly"
                                    stmt_from_list 
# 8429 "asllib/Parser.ml"
       _1_1 : (AST.stmt))
  
  let _menhir_action_303 =
    fun x ->
      (
# 196 "<standard.mly>"
    ( x )
# 8437 "asllib/Parser.ml"
       : (AST.stmt))
  
  let _menhir_action_304 =
    fun _1 _2 ->
      let with_opt = 
# 562 "asllib/Parser.mly"
                ( [] )
# 8445 "asllib/Parser.ml"
       in
      let _2_1 = _2 in
      ((_2_1, with_opt) : (string * AST.field list))
  
  let _menhir_action_305 =
    fun _1 _1_inlined1 _2 fields ->
      let with_opt =
        let _1 = _1_inlined1 in
        fields
      in
      let _2_1 = _2 in
      ((_2_1, with_opt) : (string * AST.field list))
  
  let _menhir_action_306 =
    fun () ->
      (
# 123 "<standard.mly>"
    ( None )
# 8464 "asllib/Parser.ml"
       : ((string * AST.field list) option))
  
  let _menhir_action_307 =
    fun x ->
      (
# 126 "<standard.mly>"
    ( Some x )
# 8472 "asllib/Parser.ml"
       : ((string * AST.field list) option))
  
  let _menhir_action_308 =
    fun () ->
      (
# 173 "asllib/Parser.mly"
                  ( [] )
# 8480 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_309 =
    fun _1 ->
      (_1 : (AST.bitfield list))
  
  let _menhir_action_310 =
    fun () ->
      (
# 173 "asllib/Parser.mly"
                  ( [] )
# 8492 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_311 =
    fun _1 ->
      (_1 : (AST.field list))
  
  let _menhir_action_312 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 8504 "asllib/Parser.ml"
       in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8509 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_313 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 8519 "asllib/Parser.ml"
        
      in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8525 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_314 =
    fun _2 h t ->
      (
# 178 "asllib/Parser.mly"
                              ( h :: t )
# 8533 "asllib/Parser.ml"
       : (string list))
  
  let _menhir_action_315 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 8541 "asllib/Parser.ml"
       in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8546 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_316 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 8556 "asllib/Parser.ml"
        
      in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8562 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_317 =
    fun _2 h t ->
      (
# 178 "asllib/Parser.mly"
                              ( h :: t )
# 8570 "asllib/Parser.ml"
       : (AST.bitfield list))
  
  let _menhir_action_318 =
    fun x ->
      let _2 = 
# 123 "<standard.mly>"
    ( None )
# 8578 "asllib/Parser.ml"
       in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8583 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_319 =
    fun x x_inlined1 ->
      let _2 =
        let x = x_inlined1 in
        
# 126 "<standard.mly>"
    ( Some x )
# 8593 "asllib/Parser.ml"
        
      in
      (
# 177 "asllib/Parser.mly"
                         ( [ x ] )
# 8599 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_320 =
    fun _2 h t ->
      (
# 178 "asllib/Parser.mly"
                              ( h :: t )
# 8607 "asllib/Parser.ml"
       : (AST.field list))
  
  let _menhir_action_321 =
    fun _1 _endpos_c_ _startpos__1_ c ->
      let desc = 
# 382 "asllib/Parser.mly"
                                                          T_Int         
# 8615 "asllib/Parser.ml"
       c in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_c_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8623 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_322 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 383 "asllib/Parser.mly"
                                                        ( T_Real       )
# 8631 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8639 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_323 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 384 "asllib/Parser.mly"
                                                        ( T_Bool       )
# 8647 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8655 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_324 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 385 "asllib/Parser.mly"
                                                        ( T_String     )
# 8663 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8671 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_325 =
    fun _1 _endpos__1_ _startpos__1_ ->
      let desc = 
# 386 "asllib/Parser.mly"
                                                        ( t_bit        )
# 8679 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__1_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8687 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_326 =
    fun _1 _endpos_bitfields_opt_ _startpos__1_ bitfields_opt x ->
      let desc =
        let _2 = 
# 205 "<standard.mly>"
    ( x )
# 8696 "asllib/Parser.ml"
         in
        let _2_1 = _2 in
        
# 387 "asllib/Parser.mly"
                                                          T_Bits        
# 8702 "asllib/Parser.ml"
         (_2_1, bitfields_opt)
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_bitfields_opt_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8711 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_327 =
    fun _endpos__3_ _startpos__1_ x ->
      let desc =
        let l = 
# 205 "<standard.mly>"
    ( x )
# 8720 "asllib/Parser.ml"
         in
        
# 388 "asllib/Parser.mly"
                                                          T_Tuple       
# 8725 "asllib/Parser.ml"
         l
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8734 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_328 =
    fun _endpos_name_ _startpos_name_ name ->
      let desc = 
# 389 "asllib/Parser.mly"
                                                          T_Named       
# 8742 "asllib/Parser.ml"
       name in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_name_, _startpos_name_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8750 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_329 =
    fun _1 _2 _4 _5 _endpos_t_ _startpos__1_ e t ->
      let desc = 
# 390 "asllib/Parser.mly"
                                                        ( T_Array (ArrayLength_Expr e, t) )
# 8758 "asllib/Parser.ml"
       in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_t_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8766 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_330 =
    fun _1 ->
      (_1 : (AST.ty))
  
  let _menhir_action_331 =
    fun _1 _endpos__3_ _startpos__1_ x ->
      let desc =
        let l = 
# 205 "<standard.mly>"
    ( x )
# 8779 "asllib/Parser.ml"
         in
        
# 395 "asllib/Parser.mly"
                                                          T_Enum        
# 8784 "asllib/Parser.ml"
         l
      in
      let (_endpos_desc_, _startpos_desc_) = (_endpos__3_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8793 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_332 =
    fun _1 _endpos_l_ _startpos__1_ l ->
      let desc = 
# 396 "asllib/Parser.mly"
                                                          T_Record      
# 8801 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8809 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_333 =
    fun _1 _endpos_l_ _startpos__1_ l ->
      let desc = 
# 397 "asllib/Parser.mly"
                                                          T_Exception   
# 8817 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8825 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_334 =
    fun _1 _endpos_l_ _startpos__1_ l ->
      let desc = 
# 398 "asllib/Parser.mly"
                                                          T_Collection  
# 8833 "asllib/Parser.ml"
       l in
      let (_endpos_desc_, _startpos_desc_) = (_endpos_l_, _startpos__1_) in
      let _endpos = _endpos_desc_ in
      let _symbolstartpos = _startpos_desc_ in
      (
# 156 "asllib/Parser.mly"
                              ( { desc; pos_start=_symbolstartpos; pos_end=_endpos; version } )
# 8841 "asllib/Parser.ml"
       : (AST.ty))
  
  let _menhir_action_335 =
    fun x y ->
      (
# 172 "<standard.mly>"
    ( (x, y) )
# 8849 "asllib/Parser.ml"
       : (AST.field))
  
  let _menhir_action_336 =
    fun i ->
      (
# 206 "asllib/Parser.mly"
                        L_Int          
# 8857 "asllib/Parser.ml"
       i : (AST.literal))
  
  let _menhir_action_337 =
    fun b ->
      (
# 207 "asllib/Parser.mly"
                        L_Bool         
# 8865 "asllib/Parser.ml"
       b : (AST.literal))
  
  let _menhir_action_338 =
    fun r ->
      (
# 208 "asllib/Parser.mly"
                        L_Real         
# 8873 "asllib/Parser.ml"
       r : (AST.literal))
  
  let _menhir_action_339 =
    fun b ->
      (
# 209 "asllib/Parser.mly"
                        L_BitVector    
# 8881 "asllib/Parser.ml"
       b : (AST.literal))
  
  let _menhir_action_340 =
    fun s ->
      (
# 210 "asllib/Parser.mly"
                        L_String       
# 8889 "asllib/Parser.ml"
       s : (AST.literal))
  
  let _menhir_print_token : token -> string =
    fun _tok ->
      match _tok with
      | Tokens.ACCESSOR ->
          "ACCESSOR"
      | Tokens.AND ->
          "AND"
      | Tokens.ARBITRARY ->
          "ARBITRARY"
      | Tokens.ARRAY ->
          "ARRAY"
      | Tokens.ARROW ->
          "ARROW"
      | Tokens.AS ->
          "AS"
      | Tokens.ASSERT ->
          "ASSERT"
      | Tokens.BAND ->
          "BAND"
      | Tokens.BEGIN ->
          "BEGIN"
      | Tokens.BEQ ->
          "BEQ"
      | Tokens.BIARROW ->
          "BIARROW"
      | Tokens.BIT ->
          "BIT"
      | Tokens.BITS ->
          "BITS"
      | Tokens.BITVECTOR_LIT _ ->
          "BITVECTOR_LIT"
      | Tokens.BNOT ->
          "BNOT"
      | Tokens.BOOLEAN ->
          "BOOLEAN"
      | Tokens.BOOL_LIT _ ->
          "BOOL_LIT"
      | Tokens.BOR ->
          "BOR"
      | Tokens.CASE ->
          "CASE"
      | Tokens.CATCH ->
          "CATCH"
      | Tokens.COLLECTION ->
          "COLLECTION"
      | Tokens.COLON ->
          "COLON"
      | Tokens.COLON_COLON ->
          "COLON_COLON"
      | Tokens.COMMA ->
          "COMMA"
      | Tokens.CONFIG ->
          "CONFIG"
      | Tokens.CONSTANT ->
          "CONSTANT"
      | Tokens.DEBUG ->
          "DEBUG"
      | Tokens.DIV ->
          "DIV"
      | Tokens.DIVRM ->
          "DIVRM"
      | Tokens.DO ->
          "DO"
      | Tokens.DOT ->
          "DOT"
      | Tokens.DOWNTO ->
          "DOWNTO"
      | Tokens.ELSE ->
          "ELSE"
      | Tokens.ELSIF ->
          "ELSIF"
      | Tokens.END ->
          "END"
      | Tokens.ENUMERATION ->
          "ENUMERATION"
      | Tokens.EOF ->
          "EOF"
      | Tokens.EQ ->
          "EQ"
      | Tokens.EQ_OP ->
          "EQ_OP"
      | Tokens.EXCEPTION ->
          "EXCEPTION"
      | Tokens.FOR ->
          "FOR"
      | Tokens.FUNC ->
          "FUNC"
      | Tokens.GEQ ->
          "GEQ"
      | Tokens.GETTER ->
          "GETTER"
      | Tokens.GT ->
          "GT"
      | Tokens.IDENTIFIER _ ->
          "IDENTIFIER"
      | Tokens.IF ->
          "IF"
      | Tokens.IMPDEF ->
          "IMPDEF"
      | Tokens.IMPL ->
          "IMPL"
      | Tokens.IMPLEMENTATION ->
          "IMPLEMENTATION"
      | Tokens.IN ->
          "IN"
      | Tokens.INTEGER ->
          "INTEGER"
      | Tokens.INT_LIT _ ->
          "INT_LIT"
      | Tokens.LBRACE ->
          "LBRACE"
      | Tokens.LBRACKET ->
          "LBRACKET"
      | Tokens.LEQ ->
          "LEQ"
      | Tokens.LET ->
          "LET"
      | Tokens.LLBRACKET ->
          "LLBRACKET"
      | Tokens.LOOPLIMIT ->
          "LOOPLIMIT"
      | Tokens.LPAR ->
          "LPAR"
      | Tokens.LT ->
          "LT"
      | Tokens.MASK_LIT _ ->
          "MASK_LIT"
      | Tokens.MINUS ->
          "MINUS"
      | Tokens.MOD ->
          "MOD"
      | Tokens.MUL ->
          "MUL"
      | Tokens.NEQ ->
          "NEQ"
      | Tokens.NOT ->
          "NOT"
      | Tokens.OF ->
          "OF"
      | Tokens.OR ->
          "OR"
      | Tokens.OTHERWISE ->
          "OTHERWISE"
      | Tokens.PASS ->
          "PASS"
      | Tokens.PLUS ->
          "PLUS"
      | Tokens.PLUS_COLON ->
          "PLUS_COLON"
      | Tokens.POW ->
          "POW"
      | Tokens.PRAGMA ->
          "PRAGMA"
      | Tokens.PRINT ->
          "PRINT"
      | Tokens.PRINTLN ->
          "PRINTLN"
      | Tokens.RBRACE ->
          "RBRACE"
      | Tokens.RBRACKET ->
          "RBRACKET"
      | Tokens.RDIV ->
          "RDIV"
      | Tokens.REAL ->
          "REAL"
      | Tokens.REAL_LIT _ ->
          "REAL_LIT"
      | Tokens.RECORD ->
          "RECORD"
      | Tokens.RECURSELIMIT ->
          "RECURSELIMIT"
      | Tokens.REPEAT ->
          "REPEAT"
      | Tokens.RETURN ->
          "RETURN"
      | Tokens.RPAR ->
          "RPAR"
      | Tokens.RRBRACKET ->
          "RRBRACKET"
      | Tokens.SEMI_COLON ->
          "SEMI_COLON"
      | Tokens.SETTER ->
          "SETTER"
      | Tokens.SHL ->
          "SHL"
      | Tokens.SHR ->
          "SHR"
      | Tokens.SLICING ->
          "SLICING"
      | Tokens.STAR_COLON ->
          "STAR_COLON"
      | Tokens.STRING ->
          "STRING"
      | Tokens.STRING_LIT _ ->
          "STRING_LIT"
      | Tokens.SUBTYPES ->
          "SUBTYPES"
      | Tokens.THEN ->
          "THEN"
      | Tokens.THROW ->
          "THROW"
      | Tokens.TO ->
          "TO"
      | Tokens.TRY ->
          "TRY"
      | Tokens.TYPE ->
          "TYPE"
      | Tokens.UNREACHABLE ->
          "UNREACHABLE"
      | Tokens.UNTIL ->
          "UNTIL"
      | Tokens.VAR ->
          "VAR"
      | Tokens.WHEN ->
          "WHEN"
      | Tokens.WHERE ->
          "WHERE"
      | Tokens.WHILE ->
          "WHILE"
      | Tokens.WITH ->
          "WITH"
      | Tokens.XOR ->
          "XOR"
  
  let _menhir_fail : unit -> 'a =
    fun () ->
      Printf.eprintf "Internal failure -- please contact the parser generator's developers.\n%!";
      assert false
  
  include struct
    
    [@@@ocaml.warning "-4-37"]
    
    let _menhir_run_546 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_opn =
      fun _menhir_stack _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.EOF ->
            let (_2, body) = ((), _v) in
            let _v = _menhir_action_228 _2 body in
            MenhirBox_opn _v
        | _ ->
            _eRR ()
    
    let _menhir_run_753 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_stmts =
      fun _menhir_stack _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.EOF ->
            let x = _v in
            let _v = _menhir_action_303 x in
            MenhirBox_stmts _v
        | _ ->
            _eRR ()
    
    let _menhir_run_740 : type  ttv_stack. ttv_stack -> _ -> _menhir_box_spec =
      fun _menhir_stack _v ->
        let x = _v in
        let _v = _menhir_action_266 x in
        MenhirBox_spec _v
    
    let rec _menhir_run_750 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_decl -> _ -> _menhir_box_spec =
      fun _menhir_stack _v ->
        let MenhirCell1_decl (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let xs = _v in
        let _v = _menhir_action_213 x xs in
        _menhir_goto_list_decl_ _menhir_stack _v _menhir_s
    
    and _menhir_goto_list_decl_ : type  ttv_stack. ttv_stack -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _v _menhir_s ->
        match _menhir_s with
        | MenhirState749 ->
            _menhir_run_750 _menhir_stack _v
        | MenhirState549 ->
            _menhir_run_740 _menhir_stack _v
        | _ ->
            _menhir_fail ()
    
    let rec _menhir_run_001 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState001 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_002 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_s_, _startpos_s_, s) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_340 s in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s_ _startpos_s_ _v _menhir_s _tok
    
    and _menhir_goto_value : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState448 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_175 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState745 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState726 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState630 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState598 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState555 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState508 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState488 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState484 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState477 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState472 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState461 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState453 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState450 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState445 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState440 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState435 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState412 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState388 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState379 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState374 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState369 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState334 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState336 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState344 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState341 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState329 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState308 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState297 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState277 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState233 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState231 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState217 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState212 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState201 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState147 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState133 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState131 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState116 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState104 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState098 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState096 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState091 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState089 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState087 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState085 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState083 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState081 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState079 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState077 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState075 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState072 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState070 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState066 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState064 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState062 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState060 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState056 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState052 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState040 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState036 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState034 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState031 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState029 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState027 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState025 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_023 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_175 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_value_, _startpos_value_, value) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_152 _endpos_value_ _startpos_value_ value in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_value_ _startpos_value_ _v _menhir_s _tok
    
    and _menhir_goto_expr_pattern : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState055 ->
            _menhir_run_256 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState448 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_181 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_256 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.RPAR ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos_0) in
            let _v = _menhir_action_193 _endpos__3_ _startpos__1_ x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_201 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_203 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_205 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_207 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_212 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_214 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_219 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_221 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_223 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_231 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_233 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_235 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_237 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_239 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_241 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_243 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState256
        | Tokens.COMMA ->
            let (_endpos_expr_pattern_, _startpos_expr_pattern_, expr_pattern) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_241 _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_pattern_ _startpos_expr_pattern_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_182 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_XOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState182 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_003 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_r_, _startpos_r_, r) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_338 r in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_r_ _startpos_r_ _v _menhir_s _tok
    
    and _menhir_run_004 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState004 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_005 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState005 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_006 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState006 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_007 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_i_, _startpos_i_, i) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_336 i in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_i_ _startpos_i_ _v _menhir_s _tok
    
    and _menhir_run_008 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState008 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_009 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState009
        | Tokens.LBRACE ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState009
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_108 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_010 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState010
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState010
        | Tokens.RPAR ->
            let _v = _menhir_action_027 () in
            _menhir_run_294 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_011 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_b_, _startpos_b_, b) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_337 b in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_b_ _v _menhir_s _tok
    
    and _menhir_run_012 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState012 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_013 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_b_, _startpos_b_, b) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_339 b in
        _menhir_goto_value _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_b_ _startpos_b_ _v _menhir_s _tok
    
    and _menhir_run_014 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_s = MenhirState015 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_016 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_324 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_ty : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState733 ->
            _menhir_run_734 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState702 ->
            _menhir_run_703 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState641 ->
            _menhir_run_642 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState627 ->
            _menhir_run_628 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState575 ->
            _menhir_run_594 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState525 ->
            _menhir_run_531 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState527 ->
            _menhir_run_528 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState015 ->
            _menhir_run_292 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState287 ->
            _menhir_run_286 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState018 ->
            _menhir_run_286 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState243 ->
            _menhir_run_244 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState173 ->
            _menhir_run_174 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState100 ->
            _menhir_run_129 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState113 ->
            _menhir_run_121 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState119 ->
            _menhir_run_120 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_734 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _menhir_s = MenhirState735 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SETTER ->
                _menhir_run_644 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.GETTER ->
                _menhir_run_661 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_644 : type  ttv_stack. ((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SETTER (_menhir_stack, _menhir_s) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.BEGIN ->
                    let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.WHILE ->
                        _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.VAR ->
                        _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.UNREACHABLE ->
                        _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.TRY ->
                        _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.THROW ->
                        _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.RETURN ->
                        _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.REPEAT ->
                        _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.PRINTLN ->
                        _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.PRINT ->
                        _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.PRAGMA ->
                        _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.PASS ->
                        _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.MINUS ->
                        _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.LPAR ->
                        _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.LET ->
                        _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.IF ->
                        _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.IDENTIFIER _v_1 ->
                        _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState647
                    | Tokens.FOR ->
                        _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.DEBUG ->
                        _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.CONSTANT ->
                        _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.CASE ->
                        _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.ASSERT ->
                        _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState647
                    | Tokens.END ->
                        let _v_2 =
                          let _endpos__0_ = _endpos_0 in
                          _menhir_action_225 _endpos__0_
                        in
                        _menhir_run_648 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState647 _tok
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_312 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState312
        | Tokens.LPAR ->
            _menhir_run_314 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState312
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, MenhirState312, _v, _startpos_0, _endpos) in
                _menhir_run_325 _menhir_stack _menhir_lexbuf _menhir_lexer
            | Tokens.COLON | Tokens.EQ | Tokens.SEMI_COLON ->
                let _v =
                  let _1 = _v in
                  _menhir_action_099 _1
                in
                _menhir_run_327 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState312 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_313 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_098 _1 in
        _menhir_goto_discard_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_identifier : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState459 ->
            _menhir_run_327 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState312 ->
            _menhir_run_327 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState319 ->
            _menhir_run_318 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState317 ->
            _menhir_run_318 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState314 ->
            _menhir_run_316 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_327 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let discard_or_identifier = _v in
        let _v = _menhir_action_092 discard_or_identifier in
        _menhir_goto_decl_item _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_decl_item : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState459 ->
            _menhir_run_460 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState312 ->
            _menhir_run_328 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_460 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl_item (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState460) in
            let _menhir_s = MenhirState461 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState460
        | _ ->
            _eRR ()
    
    and _menhir_run_113 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState113 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_017 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_322 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_018 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState018 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RPAR ->
            let _v = _menhir_action_029 () in
            _menhir_goto_clist0_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_019 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_INTEGER (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState019 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RECURSELIMIT | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.SUBTYPES | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let _v = _menhir_action_066 () in
            _menhir_goto_constraint_kind_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_020 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.MINUS ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.REAL_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.RBRACE ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let (_startpos__1_, x, _endpos__3_) = (_startpos, (), _endpos_1) in
                let _v = _menhir_action_064 x in
                _menhir_goto_constraint_kind _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.INT_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.IDENTIFIER _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.BOOL_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | Tokens.BITVECTOR_LIT _v ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState021
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, MenhirState020, _startpos_0, _endpos) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState021
            | _ ->
                _eRR ())
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState020
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState020
        | _ ->
            _eRR ()
    
    and _menhir_goto_constraint_kind : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState019 ->
            _menhir_run_285 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState243 ->
            _menhir_run_245 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState100 ->
            _menhir_run_130 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_285 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_INTEGER -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_065 _1 in
        _menhir_goto_constraint_kind_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _tok
    
    and _menhir_goto_constraint_kind_opt : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_INTEGER -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_INTEGER (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_c_, c) = ((), _endpos, _v) in
        let _v = _menhir_action_321 _1 _endpos_c_ _startpos__1_ c in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_c_ _v _menhir_s _tok
    
    and _menhir_run_245 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_constraint_kind_, _startpos_constraint_kind_, constraint_kind) = ((), _endpos, _startpos, _v) in
        let _v = _menhir_action_188 _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_pattern_ constraint_kind expr_pattern in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_constraint_kind_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_130 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_constraint_kind_, _startpos_constraint_kind_, constraint_kind) = ((), _endpos, _startpos, _v) in
        let _v = _menhir_action_143 _2 _endpos_constraint_kind_ _startpos_constraint_kind_ _startpos_expr_ constraint_kind expr in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_constraint_kind_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_goto_expr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState745 ->
            _menhir_run_746 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState726 ->
            _menhir_run_727 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState630 ->
            _menhir_run_631 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState555 ->
            _menhir_run_556 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_520 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState508 ->
            _menhir_run_509 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState488 ->
            _menhir_run_489 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState484 ->
            _menhir_run_485 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState477 ->
            _menhir_run_478 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState472 ->
            _menhir_run_473 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_466 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState461 ->
            _menhir_run_462 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState453 ->
            _menhir_run_454 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState450 ->
            _menhir_run_451 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState445 ->
            _menhir_run_446 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState435 ->
            _menhir_run_436 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_432 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_409 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState388 ->
            _menhir_run_389 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_366 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_361 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState334 ->
            _menhir_run_330 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState329 ->
            _menhir_run_330 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState308 ->
            _menhir_run_309 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_307 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_306 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_300 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState297 ->
            _menhir_run_298 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_296 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_293 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_280 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_278 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState277 ->
            _menhir_run_278 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_271 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_269 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_267 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_265 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_265 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_252 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_247 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_242 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_240 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_238 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_236 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState233 ->
            _menhir_run_234 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState231 ->
            _menhir_run_232 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_225 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_222 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_220 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState217 ->
            _menhir_run_218 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState212 ->
            _menhir_run_213 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_210 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_208 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_206 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_204 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState201 ->
            _menhir_run_202 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_200 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_197 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_195 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_193 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_191 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_189 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_187 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_185 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_183 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_169 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState598 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState440 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState412 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState379 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState374 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState369 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState336 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState344 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState341 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState147 ->
            _menhir_run_155 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_153 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_140 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_138 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_136 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState133 ->
            _menhir_run_134 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState131 ->
            _menhir_run_132 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState116 ->
            _menhir_run_117 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState104 ->
            _menhir_run_105 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState098 ->
            _menhir_run_099 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState096 ->
            _menhir_run_097 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_095 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState091 ->
            _menhir_run_093 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState089 ->
            _menhir_run_090 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState087 ->
            _menhir_run_088 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState085 ->
            _menhir_run_086 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState083 ->
            _menhir_run_084 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState081 ->
            _menhir_run_082 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState079 ->
            _menhir_run_080 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState077 ->
            _menhir_run_078 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState075 ->
            _menhir_run_076 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState072 ->
            _menhir_run_074 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState070 ->
            _menhir_run_071 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_069 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState066 ->
            _menhir_run_067 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState064 ->
            _menhir_run_065 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState062 ->
            _menhir_run_063 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState060 ->
            _menhir_run_061 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_059 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState056 ->
            _menhir_run_057 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState052 ->
            _menhir_run_053 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState040 ->
            _menhir_run_041 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState036 ->
            _menhir_run_037 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState034 ->
            _menhir_run_035 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState031 ->
            _menhir_run_032 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState029 ->
            _menhir_run_030 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState027 ->
            _menhir_run_028 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState025 ->
            _menhir_run_026 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_024 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_746 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_option_as_ty_ (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_global_let_or_constant (_menhir_stack, _menhir_s, keyword, _startpos_keyword_) = _menhir_stack in
            let (_endpos__2_, x, _4) = (_endpos_0, _v, ()) in
            let _v = _menhir_action_081 _4 _endpos__2_ _startpos_keyword_ keyword name ty x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState746
        | _ ->
            _eRR ()
    
    and _menhir_run_027 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_XOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState027 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_029 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState029 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_036 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState036 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_decl : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.VAR ->
            _menhir_run_550 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.TYPE ->
            _menhir_run_560 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.PRAGMA ->
            _menhir_run_597 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.LET ->
            _menhir_run_601 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.IMPLEMENTATION ->
            _menhir_run_602 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.IMPDEF ->
            _menhir_run_683 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.FUNC ->
            _menhir_run_708 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.CONSTANT ->
            _menhir_run_722 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.CONFIG ->
            _menhir_run_723 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.ACCESSOR ->
            _menhir_run_729 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState749
        | Tokens.EOF ->
            let _v_0 = _menhir_action_212 () in
            _menhir_run_750 _menhir_stack _v_0
        | _ ->
            _eRR ()
    
    and _menhir_run_550 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState550 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_551 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_552 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_551 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_203 _1 in
        _menhir_goto_ignored_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_ignored_or_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState742 ->
            _menhir_run_743 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState723 ->
            _menhir_run_724 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState550 ->
            _menhir_run_553 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_743 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_global_let_or_constant as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ignored_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState743
        | Tokens.EQ ->
            let _v_0 = _menhir_action_231 () in
            _menhir_run_744 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState743 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_744 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_option_as_ty_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState745 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_724 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_CONFIG as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ignored_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState724
        | _ ->
            _eRR ()
    
    and _menhir_run_553 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ignored_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState553
        | Tokens.EQ ->
            let _v_0 = _menhir_action_231 () in
            _menhir_run_554 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState553 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_554 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_option_as_ty_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState555 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_552 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = _v in
        let _v = _menhir_action_204 _1 in
        _menhir_goto_ignored_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_560 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SUBTYPES ->
                _menhir_run_562 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState561
            | Tokens.OF ->
                let _menhir_stack = MenhirCell1_OF (_menhir_stack, MenhirState561) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.RECORD ->
                    let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos_2 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_RECORD (_menhir_stack, MenhirState575, _startpos_1) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState576
                    | Tokens.SEMI_COLON | Tokens.SUBTYPES ->
                        let _v_3 = _menhir_action_198 () in
                        _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_2 _v_3 _tok
                    | _ ->
                        _eRR ())
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.IDENTIFIER _v_4 ->
                    _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState575
                | Tokens.EXCEPTION ->
                    let _startpos_5 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos_6 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_EXCEPTION (_menhir_stack, MenhirState575, _startpos_5) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState579
                    | Tokens.SEMI_COLON | Tokens.SUBTYPES ->
                        let _v_7 = _menhir_action_198 () in
                        _menhir_run_580 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_6 _v_7 _tok
                    | _ ->
                        _eRR ())
                | Tokens.ENUMERATION ->
                    let _menhir_s = MenhirState575 in
                    let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                    let _menhir_stack = MenhirCell1_ENUMERATION (_menhir_stack, _menhir_s, _startpos) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                        let _menhir_stack = MenhirCell0_LBRACE (_menhir_stack, _startpos) in
                        let _menhir_s = MenhirState582 in
                        let _tok = _menhir_lexer _menhir_lexbuf in
                        (match (_tok : MenhirBasics.token) with
                        | Tokens.IDENTIFIER _v ->
                            _menhir_run_583 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                        | _ ->
                            _eRR ())
                    | _ ->
                        _eRR ())
                | Tokens.COLLECTION ->
                    let _startpos_11 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos_12 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_COLLECTION (_menhir_stack, MenhirState575, _startpos_11) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.LBRACE ->
                        _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState588
                    | Tokens.SEMI_COLON | Tokens.SUBTYPES ->
                        let _v_13 = _menhir_action_198 () in
                        _menhir_run_589 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_12 _v_13 _tok
                    | _ ->
                        _eRR ())
                | Tokens.BOOLEAN ->
                    _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.BITS ->
                    _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.BIT ->
                    _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | Tokens.ARRAY ->
                    _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState575
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_562 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WITH ->
                let _menhir_stack = MenhirCell1_SUBTYPES (_menhir_stack, _menhir_s, _startpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
                let _menhir_s = MenhirState564 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_565 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.SEMI_COLON ->
                let (_startpos__1_, _1, _endpos__2_, _2) = (_startpos, (), _endpos, _v) in
                let _v = _menhir_action_304 _1 _2 in
                _menhir_goto_subtype _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _startpos__1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_565 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState565 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_566 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.RBRACE ->
            let _v = _menhir_action_310 () in
            _menhir_goto_tclist0_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_566 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _menhir_s = MenhirState566 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist0_typed_identifier_ : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_197 x in
        _menhir_goto_fields _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
    
    and _menhir_goto_fields : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState588 ->
            _menhir_run_578 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState579 ->
            _menhir_run_578 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState576 ->
            _menhir_run_578 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState564 ->
            _menhir_run_574 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_578 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_199 _1 in
        _menhir_goto_fields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_fields_opt : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState588 ->
            _menhir_run_589 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState579 ->
            _menhir_run_580 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState576 ->
            _menhir_run_577 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_589 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_COLLECTION -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_COLLECTION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_l_, l) = ((), _endpos, _v) in
        let _v = _menhir_action_334 _1 _endpos_l_ _startpos__1_ l in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_ty_decl : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty_decl (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.SUBTYPES ->
            _menhir_run_562 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState590
        | Tokens.SEMI_COLON ->
            let _v = _menhir_action_306 () in
            _menhir_goto_subtype_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_subtype_opt : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ty_decl (_menhir_stack, _, t) = _menhir_stack in
            let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _3, subtype_opt) = ((), _endpos, (), _v) in
            let _v = _menhir_action_079 _1 _3 _endpos__2_ _startpos__1_ subtype_opt t x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_580 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_EXCEPTION -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_EXCEPTION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_l_, l) = ((), _endpos, _v) in
        let _v = _menhir_action_333 _1 _endpos_l_ _startpos__1_ l in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_577 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_RECORD -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RECORD (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_l_, l) = ((), _endpos, _v) in
        let _v = _menhir_action_332 _1 _endpos_l_ _startpos__1_ l in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_574 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_SUBTYPES _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, _2, _, _) = _menhir_stack in
        let MenhirCell1_SUBTYPES (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_fields_, fields, _1_inlined1) = ((), _endpos, _v, ()) in
        let _v = _menhir_action_305 _1 _1_inlined1 _2 fields in
        _menhir_goto_subtype _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_fields_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_subtype : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState561 ->
            _menhir_run_595 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState590 ->
            _menhir_run_593 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_595 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_TYPE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_0, _endpos, _startpos, _v) in
            let _v = _menhir_action_080 _1 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_593 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ty_decl -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let x = _v in
        let _v = _menhir_action_307 x in
        _menhir_goto_subtype_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
    
    and _menhir_run_101 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_name_, _startpos_name_, name) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_328 _endpos_name_ _startpos_name_ name in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_name_ _v _menhir_s _tok
    
    and _menhir_run_583 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v_0 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                _menhir_run_583 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState584
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_313 x x_inlined1 in
                _menhir_goto_tclist1_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_312 x in
            _menhir_goto_tclist1_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist1_IDENTIFIER_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState582 ->
            _menhir_run_586 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState584 ->
            _menhir_run_585 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_586 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF, _menhir_box_spec) _menhir_cell1_ENUMERATION _menhir_cell0_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LBRACE (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_ENUMERATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, x, _endpos__3_) = ((), _v, _endpos) in
        let _v = _menhir_action_331 _1 _endpos__3_ _startpos__1_ x in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_585 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_314 _2 h t in
        _menhir_goto_tclist1_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_102 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_323 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_103 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BITS (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState104 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_114 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_325 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_run_115 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARRAY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_s = MenhirState116 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_597 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState598
            | Tokens.REAL_LIT _v_2 ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState598
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.INT_LIT _v_3 ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState598
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.IDENTIFIER _v_4 ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState598
            | Tokens.BOOL_LIT _v_5 ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState598
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.BITVECTOR_LIT _v_6 ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState598
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState598
            | Tokens.SEMI_COLON ->
                let _v_7 = _menhir_action_027 () in
                _menhir_run_599 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_599 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_085 _1 _endpos__2_ _startpos__1_ e x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_601 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_201 _1 in
        _menhir_goto_global_let_or_constant _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_global_let_or_constant : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_global_let_or_constant (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_551 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState742
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_552 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState742
        | _ ->
            _eRR ()
    
    and _menhir_run_602 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.FUNC ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_FUNC (_menhir_stack, _startpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_1, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState604
                | Tokens.LPAR ->
                    let _v_2 = _menhir_action_239 () in
                    _menhir_run_613 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState604 _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.ACCESSOR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                let _startpos_3 = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_3, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState638
                | Tokens.LPAR ->
                    let _v_4 = _menhir_action_239 () in
                    _menhir_run_639 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState638 _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_605 : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState605 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_606 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_606 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState606
        | Tokens.COMMA | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_229 x in
            _menhir_goto_opt_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_opt_typed_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_opt_typed_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState609 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_606 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_047 x in
            _menhir_goto_clist1_opt_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_opt_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState605 ->
            _menhir_run_611 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState609 ->
            _menhir_run_610 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_611 : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_240 x in
        _menhir_goto_params_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_params_opt : type  ttv_stack. (ttv_stack _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState730 ->
            _menhir_run_731 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState709 ->
            _menhir_run_710 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState699 ->
            _menhir_run_700 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState685 ->
            _menhir_run_686 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState638 ->
            _menhir_run_639 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState604 ->
            _menhir_run_613 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_731 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState731
        | _ ->
            _eRR ()
    
    and _menhir_run_614 : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState614 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_566 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.RPAR ->
            let _v = _menhir_action_031 () in
            _menhir_goto_clist0_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist0_typed_identifier_ : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_200 x in
        _menhir_goto_func_args _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_func_args : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState731 ->
            _menhir_run_732 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState710 ->
            _menhir_run_711 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState700 ->
            _menhir_run_701 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState686 ->
            _menhir_run_687 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState639 ->
            _menhir_run_640 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState613 ->
            _menhir_run_621 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_732 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BIARROW ->
            let _menhir_s = MenhirState733 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_711 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_BEGIN (_menhir_stack, MenhirState711) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.IDENTIFIER _v_0 ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState712
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState712
            | Tokens.END ->
                let _v_1 =
                  let _endpos__0_ = _endpos in
                  _menhir_action_225 _endpos__0_
                in
                _menhir_run_713 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
            | _ ->
                _eRR ())
        | Tokens.ARROW ->
            _menhir_run_627 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState711
        | _ ->
            _eRR ()
    
    and _menhir_run_354 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.RPAR ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.SEMI_COLON ->
                    let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    let (_startpos__1_, _1, _endpos__2_, _3, _2_inlined1) = (_startpos, (), _endpos_1, (), ()) in
                    let _v = _menhir_action_297 _1 _2_inlined1 _3 _endpos__2_ _startpos__1_ in
                    _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_stmt : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState000 ->
            _menhir_run_546 _menhir_stack _v _tok
        | MenhirState751 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState718 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState712 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState694 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState688 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState676 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState668 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState662 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState647 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState622 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState532 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState510 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState438 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState496 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState492 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState457 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState452 ->
            _menhir_run_457 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_457 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.VAR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.UNREACHABLE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.TRY ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.THROW ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.RETURN ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.REPEAT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.PRINTLN ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.PRINT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.PRAGMA ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.PASS ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.LET ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.IDENTIFIER _v_0 ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState457
        | Tokens.FOR ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.DEBUG ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.CONSTANT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.CASE ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.ASSERT ->
            let _menhir_stack = MenhirCell1_stmt (_menhir_stack, _menhir_s, _v, _endpos) in
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState457
        | Tokens.CATCH | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.OTHERWISE | Tokens.UNTIL | Tokens.WHEN ->
            let (_endpos_x_, x) = (_endpos, _v) in
            let _v = _menhir_action_218 x in
            _menhir_goto_list1_stmt_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_358 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState358 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_359 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_startpos__1_, _1, _endpos__2_) = (_startpos, (), _endpos) in
            let _v = _menhir_action_300 _1 _endpos__2_ _startpos__1_ in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState359
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState359
        | _ ->
            _eRR ()
    
    and _menhir_run_363 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_RETURN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState363 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.SEMI_COLON ->
            let _v = _menhir_action_233 () in
            _menhir_goto_option_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_option_expr_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_RETURN -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_RETURN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, _2_inlined1) = ((), _endpos, _v) in
        let _v = _menhir_action_280 _1 _2_inlined1 _endpos__2_ _startpos__1_ in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_367 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_REPEAT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState367 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_368 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRINTLN (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState369
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState369
            | Tokens.RPAR ->
                let _v = _menhir_action_027 () in
                _menhir_run_370 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_370 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRINTLN _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_PRINTLN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, x) = ((), _endpos_0, _v) in
                let _v = _menhir_action_294 _1 _endpos__2_ _startpos__1_ x in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_373 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRINT (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState374
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState374
            | Tokens.RPAR ->
                let _v = _menhir_action_027 () in
                _menhir_run_375 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_375 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRINT _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_PRINT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, x) = ((), _endpos_0, _v) in
                let _v = _menhir_action_295 _1 _endpos__2_ _startpos__1_ x in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_378 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState379
            | Tokens.REAL_LIT _v_2 ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState379
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.INT_LIT _v_3 ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState379
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.IDENTIFIER _v_4 ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState379
            | Tokens.BOOL_LIT _v_5 ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState379
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.BITVECTOR_LIT _v_6 ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState379
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState379
            | Tokens.SEMI_COLON ->
                let _v_7 = _menhir_action_027 () in
                _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_380 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_PRAGMA _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
            let MenhirCell1_PRAGMA (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_301 _1 _endpos__2_ _startpos__1_ e x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_382 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_startpos__1_, _1, _endpos__2_) = (_startpos, (), _endpos) in
            let _v = _menhir_action_279 _1 _endpos__2_ _startpos__1_ in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_384 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
        let _v = _menhir_action_209 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_lexpr (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState472 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_385 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState385 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_386 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_386 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_100 _1 in
        _menhir_goto_discard_or_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_sliced_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState400 ->
            _menhir_run_399 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState398 ->
            _menhir_run_399 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState385 ->
            _menhir_run_397 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_399 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState400 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_386 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_039 x in
            _menhir_goto_clist1_discard_or_sliced_basic_lexpr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_387 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState387
        | Tokens.DOT ->
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState387
        | Tokens.COMMA | Tokens.LBRACKET | Tokens.RPAR ->
            let _v_0 = _menhir_action_226 () in
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_388 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState388 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_391 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_392 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DOT -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.DOT ->
            _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState392
        | Tokens.COMMA | Tokens.EQ | Tokens.LBRACKET | Tokens.RPAR ->
            let _v_0 = _menhir_action_226 () in
            _menhir_run_393 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_393 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DOT _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let MenhirCell1_DOT (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, t) = ((), _v) in
        let _v = _menhir_action_227 _1 _endpos_desc_ _startpos_desc_ desc t in
        _menhir_goto_nested_fields _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_nested_fields : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState411 ->
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState387 ->
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState390 ->
            _menhir_run_394 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState392 ->
            _menhir_run_393 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_395 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let fields = _v in
        let _v = _menhir_action_012 _endpos_desc_ _startpos_desc_ desc fields in
        _menhir_goto_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_goto_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_basic_lexpr (_menhir_stack, _menhir_s, _v, _startpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState402
        | Tokens.COMMA | Tokens.EQ | Tokens.RPAR ->
            let (_startpos_basic_lexpr_, basic_lexpr) = (_startpos, _v) in
            let _v = _menhir_action_263 basic_lexpr in
            _menhir_goto_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_basic_lexpr_ _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_033 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState033 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.COLON ->
            _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_034 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState034 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_sliced_basic_lexpr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState751 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState718 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState712 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState694 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState688 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState676 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState668 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState662 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState647 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState622 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState532 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState510 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState438 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState496 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState492 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState452 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState457 ->
            _menhir_run_458 _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok
        | MenhirState400 ->
            _menhir_run_396 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState398 ->
            _menhir_run_396 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState385 ->
            _menhir_run_396 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_458 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let (_startpos_sliced_basic_lexpr_, sliced_basic_lexpr) = (_startpos, _v) in
        let _v = _menhir_action_207 sliced_basic_lexpr in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_sliced_basic_lexpr_ _v _menhir_s _tok
    
    and _menhir_run_396 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let sliced_basic_lexpr = _v in
        let _v = _menhir_action_101 sliced_basic_lexpr in
        _menhir_goto_discard_or_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_394 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LLBRACKET, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RRBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_RRBRACKET (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, idx, _, _) = _menhir_stack in
        let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (_2, fields, _4) = ((), _v, ()) in
        let _v = _menhir_action_013 _2 _4 _endpos_desc_ _startpos_desc_ desc fields idx in
        _menhir_goto_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_goto_clist1_discard_or_sliced_basic_lexpr_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState398 ->
            _menhir_run_404 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState400 ->
            _menhir_run_401 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_404 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _, x) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_060 _2 li x in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_208 _endpos__3_ _startpos__1_ x in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_401 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr, ttv_result) _menhir_cell1_discard_or_sliced_basic_lexpr -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_040 _2 h t in
        _menhir_goto_clist1_discard_or_sliced_basic_lexpr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_397 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_sliced_basic_lexpr (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState398 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_386 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_387 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_407 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_220 _1 in
        _menhir_goto_local_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_local_decl_keyword_non_var : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, _v, _startpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState459
        | Tokens.LPAR ->
            _menhir_run_314 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState459
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_315 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState459
        | _ ->
            _eRR ()
    
    and _menhir_run_314 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState314 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_315 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_315 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = _v in
        let _v = _menhir_action_099 _1 in
        _menhir_goto_discard_or_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_408 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState408 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_411 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState411
        | Tokens.LLBRACKET ->
            _menhir_run_388 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState411
        | Tokens.LBRACE ->
            let _menhir_s = MenhirState411 in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            let _menhir_s = MenhirState412 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, MenhirState411) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState414 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.MINUS ->
                    _menhir_run_415 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.LBRACKET ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState425 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_426 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.IDENTIFIER _v ->
                _menhir_run_392 _menhir_stack _menhir_lexbuf _menhir_lexer _v
            | _ ->
                _eRR ())
        | Tokens.EQ | Tokens.LBRACKET ->
            let _v_12 = _menhir_action_226 () in
            _menhir_run_395 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_415 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        let _1 = () in
        let _v = _menhir_action_096 _1 in
        _menhir_goto_discard_or_field _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_discard_or_field : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState420 ->
            _menhir_run_419 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState418 ->
            _menhir_run_419 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState414 ->
            _menhir_run_417 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_419 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState420 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_415 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_035 x in
            _menhir_goto_clist1_discard_or_field_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_416 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos_desc_, _startpos_desc_, desc) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_097 _endpos_desc_ _startpos_desc_ desc in
        _menhir_goto_discard_or_field _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_clist1_discard_or_field_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState418 ->
            _menhir_run_422 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState420 ->
            _menhir_run_421 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_422 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LPAR, ttv_result) _menhir_cell1_discard_or_field -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_field (_menhir_stack, _, x) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_058 _2 li x in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (_2, x, _endpos__3_) = ((), _v, _endpos) in
        let _v = _menhir_action_211 _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_421 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_field, ttv_result) _menhir_cell1_discard_or_field -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_036 _2 h t in
        _menhir_goto_clist1_discard_or_field_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_417 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_field (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState418 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_415 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_416 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_426 : type  ttv_stack ttv_result. (ttv_stack _menhir_cell0_LBRACKET as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            _menhir_run_325 _menhir_stack _menhir_lexbuf _menhir_lexer
        | _ ->
            _eRR ()
    
    and _menhir_run_325 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState325 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            _menhir_run_044 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_044 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState045 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_044 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON | Tokens.RBRACKET ->
            let x = _v in
            let _v = _menhir_action_033 x in
            _menhir_goto_clist1_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_IDENTIFIER_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState325 ->
            _menhir_run_326 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState227 ->
            _menhir_run_228 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState043 ->
            _menhir_run_047 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState045 ->
            _menhir_run_046 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_326 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_057 _2 li x in
        _menhir_goto_clist2_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_clist2_IDENTIFIER_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState481 ->
            _menhir_run_482 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState425 ->
            _menhir_run_427 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState312 ->
            _menhir_run_351 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_482 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_LBRACKET as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_RBRACKET (_menhir_stack, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.EQ ->
                let _menhir_s = MenhirState484 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_427 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_210 _2 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
            _menhir_goto_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_desc_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_351 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState351
        | _ ->
            _eRR ()
    
    and _menhir_run_228 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_186 _2 _endpos__3_ _startpos_e_ e x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_047 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DOT _menhir_cell0_LBRACKET -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACKET ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_DOT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_141 _2 _endpos__3_ _startpos_e_ e x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_046 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_034 _2 h t in
        _menhir_goto_clist1_IDENTIFIER_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_429 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.EQ ->
                let _menhir_s = MenhirState431 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_439 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_DEBUG (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState440
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState440
            | Tokens.RPAR ->
                let _v = _menhir_action_027 () in
                _menhir_run_441 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_441 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_DEBUG _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_DEBUG (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, x) = ((), _endpos_0, _v) in
                let _v = _menhir_action_296 _1 _endpos__2_ _startpos__1_ x in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_444 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_221 _1 in
        _menhir_goto_local_decl_keyword_non_var _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_445 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState445 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_453 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ASSERT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState453 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_list1_stmt_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState751 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState718 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState712 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState694 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState688 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState662 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState676 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState668 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState647 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState622 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState532 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState510 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState438 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState496 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState492 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState452 ->
            _menhir_run_491 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState457 ->
            _menhir_run_470 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_491 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_302 _1 in
        _menhir_goto_stmt_list _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_stmt_list : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState751 ->
            _menhir_run_753 _menhir_stack _v _tok
        | MenhirState718 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState712 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState694 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState688 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState676 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState668 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState662 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState647 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState622 ->
            _menhir_run_623 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_543 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState536 ->
            _menhir_run_537 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState532 ->
            _menhir_run_533 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState529 ->
            _menhir_run_530 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState358 ->
            _menhir_run_523 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_518 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_513 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState510 ->
            _menhir_run_511 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_507 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok
        | MenhirState438 ->
            _menhir_run_504 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState496 ->
            _menhir_run_497 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState492 ->
            _menhir_run_493 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState452 ->
            _menhir_run_456 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_623 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_224 _1 in
        _menhir_goto_maybe_empty_stmt_list _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_maybe_empty_stmt_list : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState718 ->
            _menhir_run_719 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState712 ->
            _menhir_run_713 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState694 ->
            _menhir_run_695 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState688 ->
            _menhir_run_689 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState676 ->
            _menhir_run_677 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState668 ->
            _menhir_run_669 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState662 ->
            _menhir_run_663 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_658 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState652 ->
            _menhir_run_653 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState647 ->
            _menhir_run_648 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_634 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState622 ->
            _menhir_run_624 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_719 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__2_) = _menhir_stack in
                let (_2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined1) = ((), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_067 _1_inlined1 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__2_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__2_) = _menhir_stack in
                let (_2, x, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = ((), _v, _endpos, _startpos, ()) in
                let _v = _menhir_action_068 _1_inlined1 _2 _endpos__1_inlined1_ _startpos__1_inlined1_ _startpos__2_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_713 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__2_) = _menhir_stack in
                let (_2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined1) = ((), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_073 _1_inlined1 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__2_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos__2_) = _menhir_stack in
                let (_2, x, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1) = ((), _v, _endpos, _startpos, ()) in
                let _v = _menhir_action_074 _1_inlined1 _2 _endpos__1_inlined1_ _startpos__1_inlined1_ _startpos__2_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_695 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined2) = ((), (), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_069 _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_) = ((), (), _v, (), _endpos, _startpos) in
                let _v = _menhir_action_070 _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_689 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined2) = ((), (), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_075 _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_) = ((), (), _v, (), _endpos, _startpos) in
                let _v = _menhir_action_076 _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_677 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_SEMI_COLON _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell0_SEMI_COLON (_menhir_stack, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell1_GETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _3, _2_inlined1, _4, x_inlined1, _1_inlined2, _1_inlined4) = ((), (), (), (), (), _v, (), ()) in
                let _v = _menhir_action_003 _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _3 _4 setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.END ->
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell0_SEMI_COLON (_menhir_stack, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell1_GETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _3, _4, x_inlined1, _1_inlined2, _endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4) = ((), (), (), (), _v, (), _endpos, _startpos, ()) in
                let _v = _menhir_action_004 _1 _1_inlined2 _1_inlined4 _2 _3 _4 _endpos__1_inlined4_ _startpos__1_inlined4_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_accessors : type  ttv_stack. ((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState735 ->
            _menhir_run_736 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState704 ->
            _menhir_run_705 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState643 ->
            _menhir_run_680 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_736 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_ACCESSOR _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_ACCESSOR (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _2_inlined1, _1_inlined1, accessors, _6) = ((), (), (), (), _v, ()) in
                let _v = _menhir_action_086 _1 _1_inlined1 _2 _2_inlined1 _6 accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_ACCESSOR (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, accessors, _6) = ((), (), _endpos, _startpos, (), _v, ()) in
                let _v = _menhir_action_087 _1 _1_inlined1 _2 _6 _endpos__1_inlined1_ _startpos__1_inlined1_ accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_705 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _) = _menhir_stack in
                let (_1, _2, _2_inlined1, _1_inlined1, _1_inlined2, accessors, _6) = ((), (), (), (), (), _v, ()) in
                let _v = _menhir_action_088 _1 _1_inlined1 _1_inlined2 _2 _2_inlined1 _6 accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _) = _menhir_stack in
                let (_1, _2, _1_inlined1, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_, accessors, _6) = ((), (), (), (), _endpos, _startpos, _v, ()) in
                let _v = _menhir_action_089 _1 _1_inlined1 _1_inlined2 _2 _6 _endpos__1_inlined2_ _startpos__1_inlined2_ accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_680 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _) = _menhir_stack in
                let (_1, _2, _2_inlined1, _1_inlined1, _1_inlined2, accessors, _6) = ((), (), (), (), (), _v, ()) in
                let _v = _menhir_action_090 _1 _1_inlined1 _1_inlined2 _2 _2_inlined1 _6 accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _) = _menhir_stack in
                let (_1, _2, _1_inlined1, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_, accessors, _6) = ((), (), (), (), _endpos, _startpos, _v, ()) in
                let _v = _menhir_action_091 _1 _1_inlined1 _1_inlined2 _2 _6 _endpos__1_inlined2_ _startpos__1_inlined2_ accessors func_args name params_opt ty in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_669 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _startpos__1_inlined2_, _endpos__1_inlined2_) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell1_GETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _3, _4, x_inlined1, _1_inlined2, _1_inlined4) = ((), (), (), (), _v, (), ()) in
                let _v = _menhir_action_005 _1 _1_inlined2 _1_inlined4 _2 _3 _4 _endpos__1_inlined2_ _startpos__1_inlined2_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.END ->
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _startpos__1_inlined2_, _endpos__1_inlined2_) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell1_GETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _3, _4, x_inlined1, _1_inlined2, _endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4) = ((), (), (), _v, (), _endpos, _startpos, ()) in
                let _v = _menhir_action_006 _1 _1_inlined2 _1_inlined4 _3 _4 _endpos__1_inlined2_ _endpos__1_inlined4_ _startpos__1_inlined2_ _startpos__1_inlined4_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_663 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_GETTER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_END (_menhir_stack, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SETTER ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.EQ ->
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.IDENTIFIER _v_0 ->
                        let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
                        let _endpos_2 = _menhir_lexbuf.Lexing.lex_curr_p in
                        let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_0, _startpos_1, _endpos_2) in
                        let _tok = _menhir_lexer _menhir_lexbuf in
                        (match (_tok : MenhirBasics.token) with
                        | Tokens.BEGIN ->
                            let _endpos_3 = _menhir_lexbuf.Lexing.lex_curr_p in
                            let _tok = _menhir_lexer _menhir_lexbuf in
                            (match (_tok : MenhirBasics.token) with
                            | Tokens.WHILE ->
                                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.VAR ->
                                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.UNREACHABLE ->
                                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.TRY ->
                                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.THROW ->
                                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.RETURN ->
                                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.REPEAT ->
                                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.PRINTLN ->
                                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.PRINT ->
                                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.PRAGMA ->
                                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.PASS ->
                                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.MINUS ->
                                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.LPAR ->
                                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.LET ->
                                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.IF ->
                                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.IDENTIFIER _v_4 ->
                                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState668
                            | Tokens.FOR ->
                                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.DEBUG ->
                                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.CONSTANT ->
                                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.CASE ->
                                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.ASSERT ->
                                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState668
                            | Tokens.END ->
                                let _v_5 =
                                  let _endpos__0_ = _endpos_3 in
                                  _menhir_action_225 _endpos__0_
                                in
                                _menhir_run_669 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 _tok
                            | _ ->
                                _eRR ())
                        | _ ->
                            _eRR ())
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | Tokens.SEMI_COLON ->
                let _endpos_6 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_SEMI_COLON (_menhir_stack, _endpos_6) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.SETTER ->
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.EQ ->
                        let _tok = _menhir_lexer _menhir_lexbuf in
                        (match (_tok : MenhirBasics.token) with
                        | Tokens.IDENTIFIER _v_7 ->
                            let _startpos_8 = _menhir_lexbuf.Lexing.lex_start_p in
                            let _endpos_9 = _menhir_lexbuf.Lexing.lex_curr_p in
                            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_7, _startpos_8, _endpos_9) in
                            let _tok = _menhir_lexer _menhir_lexbuf in
                            (match (_tok : MenhirBasics.token) with
                            | Tokens.BEGIN ->
                                let _endpos_10 = _menhir_lexbuf.Lexing.lex_curr_p in
                                let _tok = _menhir_lexer _menhir_lexbuf in
                                (match (_tok : MenhirBasics.token) with
                                | Tokens.WHILE ->
                                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.VAR ->
                                    _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.UNREACHABLE ->
                                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.TRY ->
                                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.THROW ->
                                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.RETURN ->
                                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.REPEAT ->
                                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.PRINTLN ->
                                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.PRINT ->
                                    _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.PRAGMA ->
                                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.PASS ->
                                    _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.MINUS ->
                                    _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.LPAR ->
                                    _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.LET ->
                                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.IF ->
                                    _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.IDENTIFIER _v_11 ->
                                    _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11 MenhirState676
                                | Tokens.FOR ->
                                    _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.DEBUG ->
                                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.CONSTANT ->
                                    _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.CASE ->
                                    _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.ASSERT ->
                                    _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState676
                                | Tokens.END ->
                                    let _v_12 =
                                      let _endpos__0_ = _endpos_10 in
                                      _menhir_action_225 _endpos__0_
                                    in
                                    _menhir_run_677 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 _tok
                                | _ ->
                                    _eRR ())
                            | _ ->
                                _eRR ())
                        | _ ->
                            _eRR ())
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_658 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_END (_menhir_stack, _startpos__1_inlined2_, _endpos__1_inlined2_) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell1_SETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _2_inlined1, x_inlined1, _5, _1_inlined2, _1_inlined4) = ((), (), (), _v, (), (), ()) in
                let _v = _menhir_action_009 _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _5 _endpos__1_inlined2_ _startpos__1_inlined2_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.END ->
                let MenhirCell0_END (_menhir_stack, _startpos__1_inlined2_, _endpos__1_inlined2_) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell1_SETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, x_inlined1, _5, _1_inlined2, _endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4) = ((), (), _v, (), (), _endpos, _startpos, ()) in
                let _v = _menhir_action_010 _1 _1_inlined2 _1_inlined4 _2 _5 _endpos__1_inlined2_ _endpos__1_inlined4_ _startpos__1_inlined2_ _startpos__1_inlined4_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_653 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_maybe_empty_stmt_list _menhir_cell0_END _menhir_cell0_SEMI_COLON -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell0_SEMI_COLON (_menhir_stack, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell1_SETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _2_inlined1, x_inlined1, _5, _1_inlined2, _1_inlined4, _2_inlined2) = ((), (), (), _v, (), (), (), ()) in
                let _v = _menhir_action_007 _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _2_inlined2 _5 setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.END ->
                let MenhirCell0_SEMI_COLON (_menhir_stack, _) = _menhir_stack in
                let MenhirCell0_END (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _, x) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, setter_arg, _, _) = _menhir_stack in
                let MenhirCell1_SETTER (_menhir_stack, _menhir_s) = _menhir_stack in
                let (_1, _2, _2_inlined1, x_inlined1, _5, _1_inlined2, _endpos__1_inlined4_, _startpos__1_inlined4_, _1_inlined4) = ((), (), (), _v, (), (), _endpos, _startpos, ()) in
                let _v = _menhir_action_008 _1 _1_inlined2 _1_inlined4 _2 _2_inlined1 _5 _endpos__1_inlined4_ _startpos__1_inlined4_ setter_arg x x_inlined1 in
                _menhir_goto_accessors _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_648 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty, _menhir_box_spec) _menhir_cell1_SETTER _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_maybe_empty_stmt_list (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_END (_menhir_stack, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_SEMI_COLON (_menhir_stack, _endpos_0) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.GETTER ->
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.BEGIN ->
                        let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                        let _tok = _menhir_lexer _menhir_lexbuf in
                        (match (_tok : MenhirBasics.token) with
                        | Tokens.WHILE ->
                            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.VAR ->
                            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.UNREACHABLE ->
                            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.TRY ->
                            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.THROW ->
                            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.RETURN ->
                            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.REPEAT ->
                            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.PRINTLN ->
                            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.PRINT ->
                            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.PRAGMA ->
                            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.PASS ->
                            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.MINUS ->
                            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.LPAR ->
                            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.LET ->
                            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.IF ->
                            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.IDENTIFIER _v_2 ->
                            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState652
                        | Tokens.FOR ->
                            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.DEBUG ->
                            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.CONSTANT ->
                            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.CASE ->
                            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.ASSERT ->
                            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState652
                        | Tokens.END ->
                            let _v_3 =
                              let _endpos__0_ = _endpos_1 in
                              _menhir_action_225 _endpos__0_
                            in
                            _menhir_run_653 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 _tok
                        | _ ->
                            _eRR ())
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | Tokens.GETTER ->
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.BEGIN ->
                    let _endpos_4 = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.WHILE ->
                        _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.VAR ->
                        _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.UNREACHABLE ->
                        _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.TRY ->
                        _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.THROW ->
                        _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.RETURN ->
                        _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.REPEAT ->
                        _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.PRINTLN ->
                        _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.PRINT ->
                        _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.PRAGMA ->
                        _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.PASS ->
                        _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.MINUS ->
                        _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.LPAR ->
                        _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.LET ->
                        _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.IF ->
                        _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.IDENTIFIER _v_5 ->
                        _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState657
                    | Tokens.FOR ->
                        _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.DEBUG ->
                        _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.CONSTANT ->
                        _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.CASE ->
                        _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.ASSERT ->
                        _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState657
                    | Tokens.END ->
                        let _v_6 =
                          let _endpos__0_ = _endpos_4 in
                          _menhir_action_225 _endpos__0_
                        in
                        _menhir_run_658 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 _tok
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_634 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_recurse_limit -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined2) = ((), (), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_071 _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_recurse_limit (_menhir_stack, _, recurse_limit) = _menhir_stack in
                let MenhirCell1_return_type (_menhir_stack, _, return_type) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_) = ((), (), _v, (), _endpos, _startpos) in
                let _v = _menhir_action_072 _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt recurse_limit return_type x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_624 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_BEGIN -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _endpos__2_inlined1_, _2_inlined1, _1_inlined2) = ((), (), _v, _endpos_0, (), ()) in
                let _v = _menhir_action_077 _1 _1_inlined2 _2 _2_inlined1 _endpos__2_inlined1_ _startpos__1_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | Tokens.ACCESSOR | Tokens.CONFIG | Tokens.CONSTANT | Tokens.EOF | Tokens.FUNC | Tokens.IMPDEF | Tokens.IMPLEMENTATION | Tokens.LET | Tokens.PRAGMA | Tokens.TYPE | Tokens.VAR ->
                let MenhirCell1_BEGIN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_func_args (_menhir_stack, _, func_args) = _menhir_stack in
                let MenhirCell1_params_opt (_menhir_stack, _, params_opt) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, name, _, _) = _menhir_stack in
                let MenhirCell0_FUNC (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_IMPLEMENTATION (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _2, x, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_) = ((), (), _v, (), _endpos, _startpos) in
                let _v = _menhir_action_078 _1 _1_inlined2 _2 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ func_args name params_opt x in
                _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_543 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHILE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_loop_limit -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_loop_limit (_menhir_stack, _, loop_limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _4, _1_inlined1, stmt_list) = ((), _endpos_1, (), (), (), _v) in
                let _v = _menhir_action_273 _1 _1_inlined1 _2 _4 _endpos__2_ _startpos__1_ expr loop_limit stmt_list in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_loop_limit (_menhir_stack, _, loop_limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_WHILE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _4, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, stmt_list) = ((), (), _endpos_0, _startpos, (), _v) in
                let _v = _menhir_action_274 _1 _1_inlined1 _4 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ expr loop_limit stmt_list in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_537 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_list1_catcher_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let (_1, _2, _3) = ((), (), _v) in
        let _v = _menhir_action_238 _1 _2 _3 in
        _menhir_goto_otherwise_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
    
    and _menhir_goto_otherwise_opt : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_list1_catcher_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_list1_catcher_ (_menhir_stack, _, c) = _menhir_stack in
                let MenhirCell1_stmt_list (_menhir_stack, _, s, _) = _menhir_stack in
                let MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, o) = ((), _endpos_0, (), (), (), _v) in
                let _v = _menhir_action_277 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ c o s in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_list1_catcher_ (_menhir_stack, _, c) = _menhir_stack in
                let MenhirCell1_stmt_list (_menhir_stack, _, s, _) = _menhir_stack in
                let MenhirCell1_TRY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, o) = ((), (), _endpos, _startpos, (), _v) in
                let _v = _menhir_action_278 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ c o s in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_533 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (_1, _4, stmt_list) = ((), (), _v) in
        let _v = _menhir_action_025 _1 _4 stmt_list ty in
        _menhir_goto_catcher _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_catcher : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHEN ->
            let _menhir_stack = MenhirCell1_catcher (_menhir_stack, _menhir_s, _v) in
            _menhir_run_525 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState541
        | Tokens.END | Tokens.OTHERWISE ->
            let x = _v in
            let _v = _menhir_action_216 x in
            _menhir_goto_list1_catcher_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_525 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COLON ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, MenhirState525, _v, _startpos_0, _endpos) in
                let _menhir_s = MenhirState527 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOLEAN ->
                    _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITS ->
                    _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BIT ->
                    _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARRAY ->
                    _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.ARROW ->
                let _v =
                  let (_endpos_name_, _startpos_name_, name) = (_endpos, _startpos_0, _v) in
                  _menhir_action_328 _endpos_name_ _startpos_name_ name
                in
                _menhir_run_531 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v MenhirState525 _tok
            | _ ->
                _eRR ())
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState525
        | _ ->
            _eRR ()
    
    and _menhir_run_531 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ARROW ->
            let _menhir_s = MenhirState532 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_list1_catcher_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState541 ->
            _menhir_run_542 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState524 ->
            _menhir_run_534 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_542 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_catcher -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_catcher (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_217 l x in
        _menhir_goto_list1_catcher_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_534 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_TRY, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_list1_catcher_ (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.OTHERWISE ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.ARROW ->
                let _menhir_s = MenhirState536 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.WHILE ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.VAR ->
                    _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.UNREACHABLE ->
                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.TRY ->
                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.THROW ->
                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.RETURN ->
                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REPEAT ->
                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINTLN ->
                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINT ->
                    _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRAGMA ->
                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PASS ->
                    _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LET ->
                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IF ->
                    _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.FOR ->
                    _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.DEBUG ->
                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CONSTANT ->
                    _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CASE ->
                    _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ASSERT ->
                    _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.END ->
            let _v = _menhir_action_237 () in
            _menhir_goto_otherwise_opt _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_530 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ty (_menhir_stack, _, ty, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _, x, _, _) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (_1, _4, stmt_list) = ((), (), _v) in
        let _v = _menhir_action_026 _1 _4 stmt_list ty x in
        _menhir_goto_catcher _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_523 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_TRY as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.CATCH ->
            let _menhir_s = MenhirState524 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHEN ->
                _menhir_run_525 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_518 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_REPEAT as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.UNTIL ->
            let _menhir_s = MenhirState519 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_513 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSE -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ELSE (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _endpos__2_, _2) = ((), _endpos, _v) in
        let _v = _menhir_action_257 _1 _2 in
        _menhir_goto_s_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_goto_s_else : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState507 ->
            _menhir_run_515 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState511 ->
            _menhir_run_514 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_515 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
                let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
                let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, s2) = ((), _endpos_1, (), (), (), _v) in
                let _v = _menhir_action_267 _1 _1_inlined1 _2 _3 _endpos__2_ _startpos__1_ e s1 s2 in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
                let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
                let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, s2) = ((), (), _endpos_0, _startpos, (), _v) in
                let _v = _menhir_action_268 _1 _1_inlined1 _3 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ e s1 s2 in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_514 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_stmt_list (_menhir_stack, _, s1, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_s2_, s2) = ((), (), _endpos, _v) in
        let _v = _menhir_action_255 _1 _3 _endpos_s2_ _startpos__1_ e s1 s2 in
        _menhir_goto_s_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_s2_ _v _menhir_s _tok
    
    and _menhir_run_511 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ELSIF ->
            _menhir_run_508 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState511
        | Tokens.ELSE ->
            _menhir_run_512 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState511
        | Tokens.END ->
            let _v_0 =
              let _endpos__0_ = _endpos in
              _menhir_action_256 _endpos__0_
            in
            _menhir_run_514 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_508 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState508 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_512 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ELSE (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState512 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_507 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_stmt_list (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ELSIF ->
            _menhir_run_508 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState507
        | Tokens.ELSE ->
            _menhir_run_512 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState507
        | Tokens.END ->
            let _v_0 =
              let _endpos__0_ = _endpos in
              _menhir_action_256 _endpos__0_
            in
            _menhir_run_515 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_504 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_loop_limit -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_loop_limit (_menhir_stack, _, limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, end_e, _, _) = _menhir_stack in
                let MenhirCell1_direction (_menhir_stack, _, dir) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, start_e, _, _) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, index_name, _, _) = _menhir_stack in
                let MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, body, _1_inlined1, _8) = ((), _endpos_1, (), (), _v, (), ()) in
                let _v = _menhir_action_275 _1 _1_inlined1 _2 _3 _8 _endpos__2_ _startpos__1_ body dir end_e index_name limit start_e in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_loop_limit (_menhir_stack, _, limit) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, end_e, _, _) = _menhir_stack in
                let MenhirCell1_direction (_menhir_stack, _, dir) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, start_e, _, _) = _menhir_stack in
                let MenhirCell0_IDENTIFIER (_menhir_stack, index_name, _, _) = _menhir_stack in
                let MenhirCell1_FOR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, body, _endpos__1_inlined1_, _startpos__1_inlined1_, _1_inlined1, _8) = ((), (), _v, _endpos_0, _startpos, (), ()) in
                let _v = _menhir_action_276 _1 _1_inlined1 _3 _8 _endpos__1_inlined1_ _startpos__1_ _startpos__1_inlined1_ body dir end_e index_name limit start_e in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined1_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_497 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_CASE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OF, ttv_result) _menhir_cell1_list1_case_alt_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_list1_case_alt_ (_menhir_stack, _, _1_inlined1) = _menhir_stack in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _2_inlined1, _1_inlined3, otherwise_stmt, _1_inlined2) = ((), _endpos_1, (), (), (), (), _v, ()) in
                let _v = _menhir_action_271 _1 _1_inlined1 _1_inlined2 _1_inlined3 _2 _2_inlined1 _3 _endpos__2_ _startpos__1_ expr otherwise_stmt in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_list1_case_alt_ (_menhir_stack, _, _1_inlined1) = _menhir_stack in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _2_inlined1, _endpos__1_inlined3_, _startpos__1_inlined3_, _1_inlined3, otherwise_stmt, _1_inlined2) = ((), (), (), _endpos_0, _startpos, (), _v, ()) in
                let _v = _menhir_action_272 _1 _1_inlined1 _1_inlined2 _1_inlined3 _2_inlined1 _3 _endpos__1_inlined3_ _startpos__1_ _startpos__1_inlined3_ expr otherwise_stmt in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined3_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_493 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_pattern_list (_menhir_stack, _, pattern) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _4, _endpos_stmt_, stmt) = ((), (), _endpos, _v) in
        let _v = _menhir_action_023 _1 _4 _endpos_stmt_ _startpos__1_ pattern stmt in
        _menhir_goto_case_alt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_case_alt : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.WHEN ->
            let _menhir_stack = MenhirCell1_case_alt (_menhir_stack, _menhir_s, _v) in
            _menhir_run_448 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState502
        | Tokens.END | Tokens.OTHERWISE ->
            let x = _v in
            let _v = _menhir_action_214 x in
            _menhir_goto_list1_case_alt_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_448 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState448 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_040 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState040 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_052 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState052
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState052
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, ()) in
            let _v = _menhir_action_243 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_pattern : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState055 ->
            _menhir_run_253 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState448 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_179 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_253 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState254 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MASK_LIT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.LPAR ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LEQ ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LBRACE ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.GEQ ->
                _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_054 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_247 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_055 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState055 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_056 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState056 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_039 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState039 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_131 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState131 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_143 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState143
        | Tokens.LBRACE ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState144
            | Tokens.REAL_LIT _v_2 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState144
            | Tokens.RBRACE ->
                let _endpos_3 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let (_2, _endpos__3_, _3, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_3, (), _endpos, _startpos, _v) in
                let _v = _menhir_action_194 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
                _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | Tokens.INT_LIT _v_4 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState144
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | Tokens.IDENTIFIER _v_5 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState144
            | Tokens.BOOL_LIT _v_6 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState144
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | Tokens.BITVECTOR_LIT _v_7 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 MenhirState144
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState143, _startpos_0) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState144
            | _ ->
                _eRR ())
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_153 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_146 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState146
        | Tokens.LBRACE ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_147 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState146
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState146
        | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.SHL | Tokens.SHR | Tokens.XOR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_108 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_147 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_2, _endpos__3_, _3) = ((), _endpos, ()) in
            let _v = _menhir_action_149 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState147
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState147
        | _ ->
            _eRR ()
    
    and _menhir_run_152 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_EQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState152 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_168 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState168 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_170 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState170 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_171 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_171 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState171 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MASK_LIT _v ->
            _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.LPAR ->
            _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LEQ ->
            _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.GEQ ->
            _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_172 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.COLON ->
            let _menhir_s = MenhirState173 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_179 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState180 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_040 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_052 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MASK_LIT _v ->
                _menhir_run_054 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.LPAR ->
                _menhir_run_055 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LEQ ->
                _menhir_run_056 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LBRACE ->
                _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_131 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_143 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.GEQ ->
                _menhir_run_168 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_170 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_172 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.ARROW | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos_x_, _startpos_x_, x) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_049 x in
            _menhir_goto_clist1_pattern_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_x_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_pattern_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState254 ->
            _menhir_run_255 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState448 ->
            _menhir_run_251 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_251 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_251 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_249 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_255 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_pattern -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_pattern (_menhir_stack, _, x, _, _) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_062 _2 li x in
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_246 _endpos__3_ _startpos__1_ x in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_251 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_249 _1 _endpos__1_ _startpos__1_ in
        _menhir_goto_pattern_list _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_pattern_list : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState448 ->
            _menhir_run_449 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_260 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState171 ->
            _menhir_run_177 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_449 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHEN as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_pattern_list (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHERE ->
            let _menhir_s = MenhirState450 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.ARROW ->
            let _menhir_s = MenhirState492 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_260 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_250 x in
            _menhir_goto_pattern_set _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_pattern_set : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState038 ->
            _menhir_run_262 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState214 ->
            _menhir_run_216 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState448 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_176 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_262 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_IN -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_pattern_set_, pattern_set) = ((), _endpos, _v) in
        let _v = _menhir_action_144 _2 _endpos_pattern_set_ _startpos_expr_ expr pattern_set in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_pattern_set_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_run_216 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_IN -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_IN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_pattern_set_, pattern_set) = ((), _endpos, _v) in
        let _v = _menhir_action_189 _2 _endpos_pattern_set_ _startpos_expr_pattern_ expr_pattern pattern_set in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_pattern_set_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_176 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_248 _1 in
        _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_177 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, x, _endpos__3_) = ((), _v, _endpos) in
            let _v = _menhir_action_251 _1 _endpos__3_ _startpos__1_ x in
            _menhir_goto_pattern_set _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_249 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_pattern -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_pattern (_menhir_stack, _menhir_s, h, _startpos_h_, _) = _menhir_stack in
        let (_2, _endpos_t_, t) = ((), _endpos, _v) in
        let _v = _menhir_action_050 _2 h t in
        _menhir_goto_clist1_pattern_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _startpos_h_ _v _menhir_s _tok
    
    and _menhir_goto_list1_case_alt_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState502 ->
            _menhir_run_503 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState447 ->
            _menhir_run_494 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_503 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_case_alt -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_case_alt (_menhir_stack, _menhir_s, x) = _menhir_stack in
        let l = _v in
        let _v = _menhir_action_215 l x in
        _menhir_goto_list1_case_alt_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_494 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_CASE, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.OTHERWISE ->
            let _menhir_stack = MenhirCell1_list1_case_alt_ (_menhir_stack, _menhir_s, _v) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.ARROW ->
                let _menhir_s = MenhirState496 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.WHILE ->
                    _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.VAR ->
                    _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.UNREACHABLE ->
                    _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.TRY ->
                    _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.THROW ->
                    _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.RETURN ->
                    _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REPEAT ->
                    _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINTLN ->
                    _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRINT ->
                    _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PRAGMA ->
                    _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.PASS ->
                    _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LET ->
                    _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IF ->
                    _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.FOR ->
                    _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.DEBUG ->
                    _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CONSTANT ->
                    _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.CASE ->
                    _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ASSERT ->
                    _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.END ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SEMI_COLON ->
                let _endpos_1 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _endpos__2_, _2, _3, _1_inlined1, _1_inlined2) = ((), _endpos_1, (), (), _v, ()) in
                let _v = _menhir_action_269 _1 _1_inlined1 _1_inlined2 _2 _3 _endpos__2_ _startpos__1_ expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
            | Tokens.ASSERT | Tokens.CASE | Tokens.CATCH | Tokens.CONSTANT | Tokens.DEBUG | Tokens.ELSE | Tokens.ELSIF | Tokens.END | Tokens.EOF | Tokens.FOR | Tokens.IDENTIFIER _ | Tokens.IF | Tokens.LET | Tokens.LPAR | Tokens.MINUS | Tokens.OTHERWISE | Tokens.PASS | Tokens.PRAGMA | Tokens.PRINT | Tokens.PRINTLN | Tokens.REPEAT | Tokens.RETURN | Tokens.THROW | Tokens.TRY | Tokens.UNREACHABLE | Tokens.UNTIL | Tokens.VAR | Tokens.WHEN | Tokens.WHILE ->
                let MenhirCell1_OF (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
                let MenhirCell1_CASE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
                let (_1, _3, _1_inlined1, _1_inlined2, _endpos__1_inlined2_, _startpos__1_inlined2_) = ((), (), _v, (), _endpos, _startpos) in
                let _v = _menhir_action_270 _1 _1_inlined1 _1_inlined2 _3 _endpos__1_inlined2_ _startpos__1_ _startpos__1_inlined2_ expr in
                _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_inlined2_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_456 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ARROW -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARROW (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, _2, _, _) = _menhir_stack in
        let MenhirCell1_pattern_list (_menhir_stack, _, pattern) = _menhir_stack in
        let MenhirCell1_WHEN (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _4, _1_inlined1, _endpos_stmt_, stmt) = ((), (), (), _endpos, _v) in
        let _v = _menhir_action_024 _1 _1_inlined1 _2 _4 _endpos_stmt_ _startpos__1_ pattern stmt in
        _menhir_goto_case_alt _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_470 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_stmt -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_stmt (_menhir_stack, _menhir_s, x, _) = _menhir_stack in
        let (_endpos_l_, l) = (_endpos, _v) in
        let _v = _menhir_action_219 l x in
        _menhir_goto_list1_stmt_ _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_l_ _v _menhir_s _tok
    
    and _menhir_run_627 : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ARROW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState627 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_701 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BIARROW ->
            let _menhir_s = MenhirState702 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_687 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_BEGIN (_menhir_stack, MenhirState687) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.IDENTIFIER _v_0 ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState688
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState688
            | Tokens.END ->
                let _v_1 =
                  let _endpos__0_ = _endpos in
                  _menhir_action_225 _endpos__0_
                in
                _menhir_run_689 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
            | _ ->
                _eRR ())
        | Tokens.ARROW ->
            _menhir_run_627 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState687
        | _ ->
            _eRR ()
    
    and _menhir_run_640 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BIARROW ->
            let _menhir_s = MenhirState641 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_621 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_func_args (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_BEGIN (_menhir_stack, MenhirState621) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.IDENTIFIER _v_0 ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState622
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState622
            | Tokens.END ->
                let _v_1 =
                  let _endpos__0_ = _endpos in
                  _menhir_action_225 _endpos__0_
                in
                _menhir_run_624 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
            | _ ->
                _eRR ())
        | Tokens.ARROW ->
            _menhir_run_627 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState621
        | _ ->
            _eRR ()
    
    and _menhir_run_710 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState710
        | _ ->
            _eRR ()
    
    and _menhir_run_700 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState700
        | _ ->
            _eRR ()
    
    and _menhir_run_686 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState686
        | _ ->
            _eRR ()
    
    and _menhir_run_639 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState639
        | _ ->
            _eRR ()
    
    and _menhir_run_613 : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_params_opt (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            _menhir_run_614 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState613
        | _ ->
            _eRR ()
    
    and _menhir_run_610 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_opt_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_opt_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_048 _2 h t in
        _menhir_goto_clist1_opt_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_683 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_IMPDEF (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.FUNC ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_FUNC (_menhir_stack, _startpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_1, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState685
                | Tokens.LPAR ->
                    let _v_2 = _menhir_action_239 () in
                    _menhir_run_686 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState685 _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.ACCESSOR ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                let _startpos_3 = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_3, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LBRACE ->
                    _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState699
                | Tokens.LPAR ->
                    let _v_4 = _menhir_action_239 () in
                    _menhir_run_700 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState699 _tok
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_708 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_FUNC (_menhir_stack, _menhir_s, _startpos) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos_0, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState709
            | Tokens.LPAR ->
                let _v_1 = _menhir_action_239 () in
                _menhir_run_710 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState709 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_722 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let (_startpos__1_, _1) = (_startpos, ()) in
        let _v = _menhir_action_202 _1 in
        _menhir_goto_global_let_or_constant _menhir_stack _menhir_lexbuf _menhir_lexer _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_723 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_CONFIG (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState723 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.MINUS ->
            _menhir_run_551 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_552 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_729 : type  ttv_stack. ttv_stack -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ACCESSOR (_menhir_stack, _menhir_s) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_605 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState730
            | Tokens.LPAR ->
                let _v_0 = _menhir_action_239 () in
                _menhir_run_731 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState730 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_058 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_RDIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState058 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_031 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_POW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState031 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_060 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_PLUS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState060 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_070 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_OR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState070 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_072 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_146 _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_e_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState072
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState072
        | _ ->
            _eRR ()
    
    and _menhir_run_062 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MUL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState062 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_064 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MOD (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState064 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_075 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState075 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_077 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState077 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_025 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState025 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_083 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState083 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_038 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IN (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState038 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BNOT ->
            _menhir_run_215 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_215 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IN as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState215 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_171 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_089 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IMPL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState089 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_085 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_GT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState085 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_087 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState087 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_091 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_145 _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_ expr in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_expr_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState091
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState091
        | _ ->
            _eRR ()
    
    and _menhir_run_042 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState043 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_044 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.IDENTIFIER _v ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, _endpos_x_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_140 _2 _endpos_x_ _startpos_e_ e x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_066 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIVRM (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState066 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_068 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState068 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_079 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState079 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_094 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState094 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_096 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BEQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState096 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_098 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BAND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState098 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_100 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState100 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_081 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState081 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_727 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_as_ty as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_as_ty (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_CONFIG (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, x, _4) = ((), _endpos_0, _v, ()) in
            let _v = _menhir_action_082 _1 _4 _endpos__2_ _startpos__1_ name ty x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState727
        | _ ->
            _eRR ()
    
    and _menhir_run_631 : type  ttv_stack. (((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type, _menhir_box_spec) _menhir_cell1_RECURSELIMIT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState631
        | Tokens.BEGIN ->
            let MenhirCell1_RECURSELIMIT (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_253 _1 _2 in
            _menhir_goto_recurse_limit _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_recurse_limit : type  ttv_stack. ((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState716 ->
            _menhir_run_717 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | MenhirState692 ->
            _menhir_run_693 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | MenhirState629 ->
            _menhir_run_632 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_717 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _menhir_stack = MenhirCell1_recurse_limit (_menhir_stack, _menhir_s, _v) in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState718
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState718
        | Tokens.END ->
            let _v_1 =
              let _endpos__0_ = _endpos in
              _menhir_action_225 _endpos__0_
            in
            _menhir_run_719 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_693 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _menhir_stack = MenhirCell1_recurse_limit (_menhir_stack, _menhir_s, _v) in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState694
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState694
        | Tokens.END ->
            let _v_1 =
              let _endpos__0_ = _endpos in
              _menhir_action_225 _endpos__0_
            in
            _menhir_run_695 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_632 : type  ttv_stack. (((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _menhir_stack = MenhirCell1_recurse_limit (_menhir_stack, _menhir_s, _v) in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.IDENTIFIER _v_0 ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState633
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState633
        | Tokens.END ->
            let _v_1 =
              let _endpos__0_ = _endpos in
              _menhir_action_225 _endpos__0_
            in
            _menhir_run_634 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_556 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier, _menhir_box_spec) _menhir_cell1_option_as_ty_ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_option_as_ty_ (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, x, _4) = ((), _endpos_0, _v, ()) in
            let _v = _menhir_action_083 _1 _4 _endpos__2_ _startpos__1_ name ty x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState556
        | _ ->
            _eRR ()
    
    and _menhir_run_520 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_REPEAT, ttv_result) _menhir_cell1_stmt_list as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LOOPLIMIT ->
            _menhir_run_308 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState520
        | Tokens.SEMI_COLON ->
            let _v_0 = _menhir_action_222 () in
            _menhir_run_521 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_308 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LOOPLIMIT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState308 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_521 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_REPEAT, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr (_menhir_stack, _, expr, _, _) = _menhir_stack in
            let MenhirCell1_stmt_list (_menhir_stack, _, stmt_list, _) = _menhir_stack in
            let MenhirCell1_REPEAT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _3, loop_limit) = ((), _endpos, (), _v) in
            let _v = _menhir_action_298 _1 _3 _endpos__2_ _startpos__1_ expr loop_limit stmt_list in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_509 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_stmt_list, ttv_result) _menhir_cell1_ELSIF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState509) in
            let _menhir_s = MenhirState510 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState509
        | _ ->
            _eRR ()
    
    and _menhir_run_489 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_IDENTIFIER (_menhir_stack, fld, _, _) = _menhir_stack in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, _4, rhs) = (_endpos_0, (), (), _v) in
            let _v = _menhir_action_287 _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc fld rhs in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState489
        | _ ->
            _eRR ()
    
    and _menhir_run_485 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_call _menhir_cell0_LBRACKET, ttv_result) _menhir_cell1_clist2_IDENTIFIER_ _menhir_cell0_RBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_RBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _, x) = _menhir_stack in
            let MenhirCell0_LBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, _4, rhs) = (_endpos_0, (), (), _v) in
            let _v = _menhir_action_288 _2_inlined1 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState485
        | _ ->
            _eRR ()
    
    and _menhir_run_478 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_call as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_call (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, rhs) = (_endpos_0, (), _v) in
            let _v = _menhir_action_286 _2_inlined1 _endpos__2_ _endpos_desc_ _startpos_desc_ desc rhs in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState478
        | _ ->
            _eRR ()
    
    and _menhir_run_473 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_lexpr as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_lexpr (_menhir_stack, _menhir_s, le, _startpos_le_) = _menhir_stack in
            let (_endpos__2_, _2_inlined1, e) = (_endpos_0, (), _v) in
            let _v = _menhir_action_285 _2_inlined1 _endpos__2_ _startpos_le_ e le in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState473
        | _ ->
            _eRR ()
    
    and _menhir_run_466 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_as_ty (_menhir_stack, _, x) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, decl_item) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, local_decl_keyword_non_var, _startpos_local_decl_keyword_non_var_) = _menhir_stack in
            let (_endpos__2_, _4, x_inlined1) = (_endpos_0, (), _v) in
            let _v = _menhir_action_284 _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x x_inlined1 in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState466
        | _ ->
            _eRR ()
    
    and _menhir_run_462 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, decl_item) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, local_decl_keyword_non_var, _startpos_local_decl_keyword_non_var_) = _menhir_stack in
            let (_endpos__2_, x, _4) = (_endpos_0, _v, ()) in
            let _v = _menhir_action_283 _4 _endpos__2_ _startpos_local_decl_keyword_non_var_ decl_item local_decl_keyword_non_var x in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState462
        | _ ->
            _eRR ()
    
    and _menhir_run_454 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_ASSERT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ASSERT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos_0, _v) in
            let _v = _menhir_action_282 _1 _endpos__2_ _startpos__1_ e in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState454
        | _ ->
            _eRR ()
    
    and _menhir_run_451 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_pattern_list as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | Tokens.ARROW ->
            let _menhir_stack = MenhirCell1_ARROW (_menhir_stack, MenhirState451) in
            let _menhir_s = MenhirState452 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState451
        | _ ->
            _eRR ()
    
    and _menhir_run_446 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_CASE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.OF ->
            let _menhir_stack = MenhirCell1_OF (_menhir_stack, MenhirState446) in
            let _menhir_s = MenhirState447 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHEN ->
                _menhir_run_448 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState446
        | _ ->
            _eRR ()
    
    and _menhir_run_436 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.LOOPLIMIT ->
            _menhir_run_308 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState436
        | Tokens.DO ->
            let _v_0 = _menhir_action_222 () in
            _menhir_run_437 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState436 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_437 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_direction, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_loop_limit (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.DO ->
            let _menhir_s = MenhirState438 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_432 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.TO ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_menhir_s, _1) = (MenhirState432, ()) in
            let _v = _menhir_action_094 _1 in
            _menhir_goto_direction _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DOWNTO ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_menhir_s, _1) = (MenhirState432, ()) in
            let _v = _menhir_action_095 _1 in
            _menhir_goto_direction _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState432
        | _ ->
            _eRR ()
    
    and _menhir_goto_direction : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_FOR _menhir_cell0_IDENTIFIER, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_direction (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v_0 ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState435
        | Tokens.REAL_LIT _v_1 ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState435
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | Tokens.INT_LIT _v_2 ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState435
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | Tokens.IDENTIFIER _v_3 ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState435
        | Tokens.BOOL_LIT _v_4 ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState435
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | Tokens.BITVECTOR_LIT _v_5 ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState435
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState435
        | _ ->
            _eRR ()
    
    and _menhir_run_409 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState409) in
            let _menhir_s = MenhirState410 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState409
        | _ ->
            _eRR ()
    
    and _menhir_run_389 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RRBRACKET (_menhir_stack, MenhirState389, _endpos_0) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.DOT ->
                _menhir_run_391 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState390
            | Tokens.COMMA | Tokens.EQ | Tokens.LBRACKET | Tokens.RPAR ->
                let _v_1 = _menhir_action_226 () in
                _menhir_run_394 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 _tok
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState389
        | _ ->
            _eRR ()
    
    and _menhir_run_366 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_RETURN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState366
        | Tokens.SEMI_COLON ->
            let x = _v in
            let _v = _menhir_action_234 x in
            _menhir_goto_option_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_run_361 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_THROW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_THROW (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, e) = ((), _endpos_0, _v) in
            let _v = _menhir_action_299 _1 _endpos__2_ _startpos__1_ e in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState361
        | _ ->
            _eRR ()
    
    and _menhir_run_330 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState330
        | Tokens.SEMI_COLON ->
            let MenhirCell1_EQ (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_236 _1 _2 in
            _menhir_goto_option_eq_expr _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_option_eq_expr : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState333 ->
            _menhir_run_349 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState328 ->
            _menhir_run_331 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_349 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_as_ty (_menhir_stack, _, x) = _menhir_stack in
        let MenhirCell1_decl_item (_menhir_stack, _, ldi) = _menhir_stack in
        let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, e) = ((), _endpos, _v) in
        let _v = _menhir_action_291 _1 _endpos__2_ _startpos__1_ e ldi x in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_331 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_decl_item (_menhir_stack, _, ldi) = _menhir_stack in
        let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos__2_, e) = ((), _endpos, _v) in
        let _v = _menhir_action_290 _1 _endpos__2_ _startpos__1_ e ldi in
        _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
    
    and _menhir_run_309 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LOOPLIMIT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState309
        | Tokens.DO | Tokens.SEMI_COLON ->
            let MenhirCell1_LOOPLIMIT (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _2) = ((), _v) in
            let _v = _menhir_action_223 _1 _2 in
            _menhir_goto_loop_limit _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_loop_limit : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState520 ->
            _menhir_run_521 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState436 ->
            _menhir_run_437 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState307 ->
            _menhir_run_310 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_310 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHILE, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_loop_limit (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.DO ->
            let _menhir_s = MenhirState311 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_307 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_WHILE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.LOOPLIMIT ->
            _menhir_run_308 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState307
        | Tokens.DO ->
            let _v_0 = _menhir_action_222 () in
            _menhir_run_310 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState307 _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_306 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_NOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState306
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState306
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState306
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState306
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_135 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_300 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.RPAR ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos_0) in
            let _v = _menhir_action_148 _endpos__3_ _startpos__1_ x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState300) in
            let _menhir_s = MenhirState302 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState300
        | _ ->
            _eRR ()
    
    and _menhir_run_298 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.ELSIF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.ELSE ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState298
        | _ ->
            _eRR ()
    
    and _menhir_run_135 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState135 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_139 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_ELSE (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState139 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_296 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState296) in
            let _menhir_s = MenhirState297 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState296
        | _ ->
            _eRR ()
    
    and _menhir_run_293 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState293
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_133 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_280 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SLICING as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState280
        | Tokens.COMMA | Tokens.RBRACE ->
            let MenhirCell1_SLICING (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_206 _2 e1 e2 in
            _menhir_goto_int_constraint _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_int_constraint : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_int_constraint (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState277 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_045 x in
            _menhir_goto_clist1_int_constraint_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_clist1_int_constraint_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState020 ->
            _menhir_run_282 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState277 ->
            _menhir_run_281 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_282 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_063 x in
        _menhir_goto_constraint_kind _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_281 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_int_constraint -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_int_constraint (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_046 _2 h t in
        _menhir_goto_clist1_int_constraint_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_278 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_SLICING (_menhir_stack, MenhirState278) in
            let _menhir_s = MenhirState279 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState278
        | Tokens.COMMA | Tokens.RBRACE ->
            let expr = _v in
            let _v = _menhir_action_205 expr in
            _menhir_goto_int_constraint _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_271 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState271
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_259 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_slice : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_slice (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState264 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.COLON ->
                _menhir_run_034 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACKET ->
            let x = _v in
            let _v = _menhir_action_051 x in
            _menhir_goto_clist1_slice_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _menhir_fail ()
    
    and _menhir_goto_clist1_slice_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState033 ->
            _menhir_run_273 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState264 ->
            _menhir_run_272 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_273 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACKET -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACKET (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_265 x in
        _menhir_goto_slices _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_goto_slices : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState402 ->
            _menhir_run_403 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState256 ->
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState181 ->
            _menhir_run_248 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState125 ->
            _menhir_run_111 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState107 ->
            _menhir_run_111 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState746 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState727 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState631 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState556 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState520 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState409 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState509 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState446 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState451 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState489 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState485 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState478 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState473 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState466 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState462 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState454 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState436 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState432 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState389 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState366 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState361 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState330 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState307 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState309 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState306 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState300 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState296 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState298 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState293 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState278 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState280 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState024 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState026 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState028 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState030 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState032 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState265 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState271 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState269 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState267 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState035 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState037 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState252 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState247 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState242 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState240 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState238 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState236 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState234 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState232 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState225 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState222 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState220 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState218 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState213 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState210 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState208 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState206 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState204 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState202 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState200 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState197 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState195 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState193 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState191 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState189 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState187 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState185 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState183 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState169 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState155 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState153 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState132 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState134 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState136 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState138 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState140 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState057 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState105 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState117 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState099 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState097 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState095 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState090 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState093 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState074 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState088 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState086 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState084 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState078 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState082 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState080 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState076 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState071 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState061 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState069 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState067 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState065 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState063 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState059 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState053 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState041 ->
            _menhir_run_050 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_403 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_basic_lexpr -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        let MenhirCell1_basic_lexpr (_menhir_stack, _menhir_s, basic, _startpos_basic_) = _menhir_stack in
        let (_endpos_desc_, _startpos_desc_, desc) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_264 _endpos_desc_ _startpos_desc_ basic desc in
        _menhir_goto_sliced_basic_lexpr _menhir_stack _menhir_lexbuf _menhir_lexer _startpos_basic_ _v _menhir_s _tok
    
    and _menhir_run_248 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_expr_pattern -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
        let (_endpos_slices_, slices) = (_endpos, _v) in
        let _v = _menhir_action_183 _endpos_slices_ _startpos_e_ e slices in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_slices_ _startpos_e_ _v _menhir_s _tok
    
    and _menhir_run_111 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.IDENTIFIER _v_0 ->
            let _startpos_1 = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos_2 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                let _menhir_stack = MenhirCell1_slices (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_0, _startpos_1, _endpos_2) in
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState112
            | Tokens.COLON ->
                let _menhir_stack = MenhirCell1_slices (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v_0, _startpos_1, _endpos_2) in
                _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState112
            | Tokens.COMMA | Tokens.RBRACE ->
                let (s, x) = (_v, _v_0) in
                let _v = _menhir_action_014 s x in
                _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_107 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState107 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RBRACE ->
            let _v = _menhir_action_308 () in
            _menhir_goto_tclist0_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist0_bitfield_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_017 x in
        _menhir_goto_bitfields _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
    
    and _menhir_goto_bitfields : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState106 ->
            _menhir_run_128 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState112 ->
            _menhir_run_122 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_128 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let (_endpos__1_, _1) = (_endpos, _v) in
        let _v = _menhir_action_019 _1 in
        _menhir_goto_bitfields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _v _tok
    
    and _menhir_goto_bitfields_opt : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RPAR -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RPAR (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
        let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_BITS (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _endpos_bitfields_opt_, bitfields_opt) = ((), _endpos, _v) in
        let _v = _menhir_action_326 _1 _endpos_bitfields_opt_ _startpos__1_ bitfields_opt x in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_bitfields_opt_ _v _menhir_s _tok
    
    and _menhir_run_122 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slices _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
        let MenhirCell1_slices (_menhir_stack, _menhir_s, s, _, _) = _menhir_stack in
        let bf = _v in
        let _v = _menhir_action_015 bf s x in
        _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_bitfield : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACKET ->
                let _menhir_stack = MenhirCell1_bitfield (_menhir_stack, _menhir_s, _v) in
                _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState125
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_316 x x_inlined1 in
                _menhir_goto_tclist1_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_315 x in
            _menhir_goto_tclist1_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist1_bitfield_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState125 ->
            _menhir_run_126 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState107 ->
            _menhir_run_108 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_126 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_bitfield -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_bitfield (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_317 _2 h t in
        _menhir_goto_tclist1_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_108 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_309 _1 in
        _menhir_goto_tclist0_bitfield_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_050 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
        let (_endpos_slices_, slices) = (_endpos, _v) in
        let _v = _menhir_action_138 _endpos_slices_ _startpos_e_ e slices in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_slices_ _startpos_e_ _v _menhir_s _tok
    
    and _menhir_run_272 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slice -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_slice (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_052 _2 h t in
        _menhir_goto_clist1_slice_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_269 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_PLUS_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState269
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_PLUS_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_260 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_267 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_STAR_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState267
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_STAR_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _, _) = _menhir_stack in
            let (_2, e2) = ((), _v) in
            let _v = _menhir_action_262 _2 e1 e2 in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_265 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.STAR_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_STAR_COLON (_menhir_stack, MenhirState265) in
            let _menhir_s = MenhirState266 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.PLUS_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_PLUS_COLON (_menhir_stack, MenhirState265) in
            let _menhir_s = MenhirState268 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COLON (_menhir_stack, MenhirState265) in
            let _menhir_s = MenhirState270 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState265
        | Tokens.COMMA | Tokens.RBRACKET ->
            let expr = _v in
            let _v = _menhir_action_258 expr in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_252 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BNOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState252
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState252
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState252
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState252
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_BNOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_178 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_247 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState247
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_AND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_154 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_242 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BAND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState242
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BAND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_155 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_240 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState240
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_157 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_238 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_BOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState238
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_BOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_156 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_236 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_COLON_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState236
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_COLON_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_177 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_234 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState234
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState234
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState234
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState234
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState234
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_DIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_158 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_232 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_DIVRM as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState232
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState232
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState232
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState232
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState232
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_DIVRM (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_159 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_225 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_EQ_OP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState225
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_EQ_OP (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_161 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_222 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState222
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_164 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_220 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState220
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_GT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_163 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_218 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_IMPL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState218
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.IMPL | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_IMPL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_165 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_213 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState213
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_167 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_210 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos__4_, _4, e2) = ((), _endpos_0, (), _v) in
            let _v = _menhir_action_184 _2 _4 _endpos__4_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__4_ _startpos_e1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState210
        | _ ->
            _eRR ()
    
    and _menhir_run_208 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState208
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_LT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_166 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_206 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState206
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MINUS (_menhir_stack, _, _, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_169 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_204 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState204
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MOD (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_170 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_202 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_MUL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState202
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MUL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_171 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_200 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_NEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState200
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COMMA | Tokens.EQ_OP | Tokens.IMPL | Tokens.NEQ | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE ->
            let MenhirCell1_NEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_162 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_197 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState197
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_OR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_172 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_195 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState195
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_PLUS (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_168 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_193 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_POW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState193
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_POW (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_176 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_191 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_RDIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState191
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_RDIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_173 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_189 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SHL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState189
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_SHL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_174 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_187 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SHR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState187
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_SHR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_175 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_185 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_SLICING as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState185
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_SLICING (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_242 _2 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_183 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_XOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState183
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.RBRACE | Tokens.RPAR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_XOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_160 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_169 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState169
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_expr_, expr) = ((), _endpos, _v) in
            let _v = _menhir_action_245 _1 _endpos_expr_ _startpos__1_ expr in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_155 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState155) in
            let _menhir_s = MenhirState156 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState155
        | Tokens.RBRACE | Tokens.RPAR | Tokens.SEMI_COLON ->
            let x = _v in
            let _v = _menhir_action_041 x in
            _menhir_goto_clist1_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_expr_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState341 ->
            _menhir_run_342 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_303 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState598 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState440 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState379 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState374 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState369 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState344 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_163 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState412 ->
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState336 ->
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState147 ->
            _menhir_run_160 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_157 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_342 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_COMMA as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_clist1_expr_ (_menhir_stack, _menhir_s, _v) in
                let _menhir_stack = MenhirCell0_RBRACE (_menhir_stack, _endpos) in
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v_0 ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState344
                | Tokens.REAL_LIT _v_1 ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState344
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.INT_LIT _v_2 ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState344
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.IDENTIFIER _v_3 ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState344
                | Tokens.BOOL_LIT _v_4 ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState344
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.BITVECTOR_LIT _v_5 ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState344
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState344
                | Tokens.RPAR ->
                    let _v_6 = _menhir_action_027 () in
                    _menhir_run_345 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 _tok
                | _ ->
                    _eRR ())
            | Tokens.SEMI_COLON ->
                let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
                let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
                let (_2, _3, _endpos__5_, _5, params) = ((), (), _endpos, (), _v) in
                let _v = _menhir_action_105 _2 _3 _5 name params in
                _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__5_ _startpos_name_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_345 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_COMMA, ttv_result) _menhir_cell1_clist1_expr_ _menhir_cell0_RBRACE _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell0_RBRACE (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_clist1_expr_ (_menhir_stack, _, params) = _menhir_stack in
            let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
            let (_2, x, _3, _endpos__3_inlined1_, _5) = ((), _v, (), _endpos, ()) in
            let _v = _menhir_action_106 _2 _3 _5 name params x in
            _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_inlined1_ _startpos_name_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_elided_param_call : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState465 ->
            _menhir_run_468 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | MenhirState334 ->
            _menhir_run_347 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_468 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_as_ty (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, lhs) = _menhir_stack in
            let MenhirCell1_local_decl_keyword_non_var (_menhir_stack, _menhir_s, ldk, _startpos_ldk_) = _menhir_stack in
            let (_endpos__2_, _endpos_desc_, _startpos_desc_, desc, _4) = (_endpos_0, _endpos, _startpos, _v, ()) in
            let _v = _menhir_action_289 _4 _endpos__2_ _endpos_desc_ _startpos_desc_ _startpos_ldk_ desc ldk lhs ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_347 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item, ttv_result) _menhir_cell1_as_ty, ttv_result) _menhir_cell1_EQ -> _ -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_as_ty (_menhir_stack, _, ty) = _menhir_stack in
            let MenhirCell1_decl_item (_menhir_stack, _, lhs) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, _endpos_desc_, _startpos_desc_, desc, _4) = ((), _endpos_0, _endpos, _startpos, _v, ()) in
            let _v = _menhir_action_293 _1 _4 _endpos__2_ _endpos_desc_ _startpos__1_ _startpos_desc_ desc lhs ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_303 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COMMA -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, x, _, _) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_061 _2 li x in
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_151 _endpos__3_ _startpos__1_ x in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_163 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_028 _1 in
        _menhir_goto_clist0_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_clist0_expr_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState598 ->
            _menhir_run_599 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState440 ->
            _menhir_run_441 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState379 ->
            _menhir_run_380 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState374 ->
            _menhir_run_375 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState369 ->
            _menhir_run_370 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState344 ->
            _menhir_run_345 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState338 ->
            _menhir_run_339 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState010 ->
            _menhir_run_294 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState162 ->
            _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_339 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_RBRACE _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_RBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
            let (_2, x, _3, _endpos__3_inlined1_) = ((), _v, (), _endpos) in
            let _v = _menhir_action_104 _2 _3 name x in
            _menhir_goto_elided_param_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_inlined1_ _startpos_name_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_294 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LPAR (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
            let (x, _endpos__3_) = (_v, _endpos) in
            let _v = _menhir_action_020 name x in
            _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_name_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_call : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState751 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState718 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState712 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState694 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState688 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState662 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState676 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState668 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState647 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState657 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState652 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState633 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState622 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState000 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState311 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState358 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState536 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState532 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState529 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState367 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState410 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState510 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState512 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState438 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState496 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState492 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState452 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState457 ->
            _menhir_run_475 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState448 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState039 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState055 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState254 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState171 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState180 ->
            _menhir_run_250 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState745 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState726 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState630 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState598 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState555 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState001 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState519 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState408 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState508 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState431 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState435 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState445 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState450 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState488 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState484 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState477 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState472 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState465 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState461 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState453 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState440 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState412 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState388 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState379 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState374 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState369 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState363 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState359 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState334 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState336 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState341 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState344 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState338 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState329 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState308 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState004 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState005 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState006 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState302 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState008 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState297 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState010 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState012 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState020 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState277 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState279 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState021 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState025 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState027 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState029 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState031 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState033 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState264 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState270 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState268 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState266 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState034 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState036 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState170 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState246 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState241 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState239 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState237 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState235 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState233 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState231 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState223 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState221 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState219 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState217 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState212 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState209 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState207 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState205 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState203 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState201 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState198 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState196 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState194 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState192 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState190 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState188 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState186 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState184 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState182 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState168 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState144 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState147 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState162 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState156 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState152 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState131 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState133 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState135 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState137 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState139 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState056 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState104 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState116 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState098 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState096 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState094 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState089 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState091 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState072 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState087 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState085 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState083 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState077 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState081 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState079 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState075 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState070 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState060 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState068 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState066 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState064 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState062 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState058 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState052 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | MenhirState040 ->
            _menhir_run_051 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_475 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let (_endpos__2_, _startpos_call_, call) = (_endpos_0, _startpos, _v) in
            let _v = _menhir_action_281 _endpos__2_ _startpos_call_ call in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_call (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _menhir_s = MenhirState477 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_call (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACKET ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
                let _menhir_s = MenhirState481 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_426 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.IDENTIFIER _v ->
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                let _menhir_stack = MenhirCell0_IDENTIFIER (_menhir_stack, _v, _startpos, _endpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.EQ ->
                    let _menhir_s = MenhirState488 in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.STRING_LIT _v ->
                        _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.REAL_LIT _v ->
                        _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.NOT ->
                        _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.MINUS ->
                        _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.LPAR ->
                        _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.INT_LIT _v ->
                        _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.IF ->
                        _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.IDENTIFIER _v ->
                        _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.BOOL_LIT _v ->
                        _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.BNOT ->
                        _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | Tokens.BITVECTOR_LIT _v ->
                        _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                    | Tokens.ARBITRARY ->
                        _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_250 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_call_, _startpos_call_, call) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_182 _endpos_call_ _startpos_call_ call in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_call_ _startpos_call_ _v _menhir_s _tok
    
    and _menhir_run_051 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_call_, _startpos_call_, call) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_137 _endpos_call_ _startpos_call_ call in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_call_ _startpos_call_ _v _menhir_s _tok
    
    and _menhir_run_164 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE, ttv_result) _menhir_cell1_clist1_expr_ _menhir_cell0_RBRACE _menhir_cell0_LPAR -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RPAR ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell0_LPAR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell0_RBRACE (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_clist1_expr_ (_menhir_stack, _, x) = _menhir_stack in
            let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
            let (x_inlined1, _endpos__3_inlined1_) = (_v, _endpos) in
            let _v = _menhir_action_022 name x x_inlined1 in
            _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_inlined1_ _startpos_name_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_160 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.RBRACE ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_clist1_expr_ (_menhir_stack, _menhir_s, _v) in
                let _menhir_stack = MenhirCell0_RBRACE (_menhir_stack, _endpos) in
                let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos) in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v_0 ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState162
                | Tokens.REAL_LIT _v_1 ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState162
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.INT_LIT _v_2 ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState162
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.IDENTIFIER _v_3 ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_3 MenhirState162
                | Tokens.BOOL_LIT _v_4 ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_4 MenhirState162
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.BITVECTOR_LIT _v_5 ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState162
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState162
                | Tokens.RPAR ->
                    let _v_6 = _menhir_action_027 () in
                    _menhir_run_164 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 _tok
                | _ ->
                    _eRR ())
            | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
                let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
                let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, name, _startpos_name_, _) = _menhir_stack in
                let (x, _endpos__3_) = (_v, _endpos) in
                let _v = _menhir_action_021 name x in
                _menhir_goto_call _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_name_ _v _menhir_s _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_157 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COMMA -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_COMMA (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, h, _, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_042 _2 h t in
        _menhir_goto_clist1_expr_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_153 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_EQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState153
        | Tokens.COMMA | Tokens.RBRACE ->
            let MenhirCell1_EQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
            let y = _v in
            let _v = _menhir_action_196 x y in
            (match (_tok : MenhirBasics.token) with
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_field_assign (_menhir_stack, _menhir_s, _v) in
                let _menhir_s = MenhirState150 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.IDENTIFIER _v ->
                    let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
                    let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
                    let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                    let _menhir_s = MenhirState151 in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.EQ ->
                        _menhir_run_152 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                    | _ ->
                        _eRR ())
                | _ ->
                    _eRR ())
            | Tokens.RBRACE ->
                let x = _v in
                let _v = _menhir_action_043 x in
                _menhir_goto_clist1_field_assign_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _menhir_fail ())
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_field_assign_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState144 ->
            _menhir_run_166 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState336 ->
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState147 ->
            _menhir_run_158 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState150 ->
            _menhir_run_154 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_166 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_195 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_158 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IDENTIFIER, ttv_result) _menhir_cell1_LBRACE -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LBRACE (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, desc, _startpos_desc_, _endpos_desc_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_150 _endpos__3_ _endpos_desc_ _startpos_desc_ desc x in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
    
    and _menhir_run_154 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_field_assign -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_field_assign (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_044 _2 h t in
        _menhir_goto_clist1_field_assign_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_140 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSE as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState140
        | Tokens.ARROW | Tokens.BEGIN | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_ELSE (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, _endpos__2_, _2) = ((), _endpos, _v) in
            let _v = _menhir_action_102 _1 _2 in
            _menhir_goto_e_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_e_else : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState298 ->
            _menhir_run_299 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState134 ->
            _menhir_run_142 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | MenhirState138 ->
            _menhir_run_141 _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_299 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e2, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e1, _, _) = _menhir_stack in
        let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_136 _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_142 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e2, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e1, _, _) = _menhir_stack in
        let MenhirCell1_IF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_181 _1 _3 _endpos_e_else_ _startpos__1_ e1 e2 e_else in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_141 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_THEN (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, c, _, _) = _menhir_stack in
        let MenhirCell1_ELSIF (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _3, _endpos_e_else_, e_else) = ((), (), _endpos, _v) in
        let _v = _menhir_action_103 _1 _3 _endpos_e_else_ _startpos__1_ c e e_else in
        _menhir_goto_e_else _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_else_ _v _menhir_s _tok
    
    and _menhir_run_138 : type  ttv_stack ttv_result. (((((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.ELSIF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.ELSE ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState138
        | _ ->
            _eRR ()
    
    and _menhir_run_136 : type  ttv_stack ttv_result. (((((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_ELSIF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState136) in
            let _menhir_s = MenhirState137 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState136
        | _ ->
            _eRR ()
    
    and _menhir_run_134 : type  ttv_stack ttv_result. ((((ttv_stack, ttv_result) _menhir_cell1_IF, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_THEN as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.ELSIF ->
            _menhir_run_135 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.ELSE ->
            _menhir_run_139 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState134
        | _ ->
            _eRR ()
    
    and _menhir_run_132 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_IF as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.THEN ->
            let _menhir_stack = MenhirCell1_THEN (_menhir_stack, MenhirState132) in
            let _menhir_s = MenhirState133 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState132
        | _ ->
            _eRR ()
    
    and _menhir_run_117 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_ARRAY as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.RRBRACKET ->
            let _menhir_s = MenhirState117 in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RRBRACKET (_menhir_stack, _menhir_s, _endpos) in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.OF ->
                let _menhir_s = MenhirState119 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING ->
                    _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.REAL ->
                    _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INTEGER ->
                    _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOLEAN ->
                    _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITS ->
                    _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BIT ->
                    _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.ARRAY ->
                    _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState117
        | _ ->
            _eRR ()
    
    and _menhir_run_105 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_BITS _menhir_cell0_LPAR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.SHR ->
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.SHL ->
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.RPAR ->
            let _menhir_s = MenhirState105 in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _menhir_stack = MenhirCell1_RPAR (_menhir_stack, _menhir_s, _endpos) in
            let _menhir_s = MenhirState106 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.LBRACE ->
                _menhir_run_107 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOT | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RECURSELIMIT | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.SUBTYPES | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
                let _v = _menhir_action_018 () in
                _menhir_goto_bitfields_opt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok
            | _ ->
                _eRR ())
        | Tokens.RDIV ->
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.POW ->
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.PLUS ->
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.OR ->
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.NEQ ->
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.MUL ->
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.MOD ->
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.MINUS ->
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.LT ->
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.LLBRACKET ->
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.LEQ ->
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.LBRACKET ->
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.IN ->
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.IMPL ->
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.GT ->
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.GEQ ->
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.EQ_OP ->
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.DOT ->
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.DIVRM ->
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.DIV ->
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.COLON_COLON ->
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.BOR ->
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.BEQ ->
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.BAND ->
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.AS ->
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | Tokens.AND ->
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState105
        | _ ->
            _eRR ()
    
    and _menhir_run_099 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BAND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState099
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BAND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_110 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_097 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState097
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_112 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_095 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_BOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState095
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_BOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_111 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_093 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_EQ_OP as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState093
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_EQ_OP (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_116 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_090 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_IMPL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState090
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_IMPL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_120 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_088 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_GEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState088
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_GEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_119 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_086 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_GT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState086
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_GT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_118 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_084 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState084
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_122 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_082 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AND as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState082
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_AND (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_109 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_080 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_COLON_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState080
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_COLON_COLON (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_132 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_078 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState078
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_LT (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_121 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_076 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState076
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MINUS (_menhir_stack, _, _, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_124 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_074 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_NEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState074
        | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.IMPL | Tokens.LOOPLIMIT | Tokens.NEQ | Tokens.OF | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE ->
            let MenhirCell1_NEQ (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_117 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_071 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_OR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState071
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_OR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_127 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_069 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState069
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState069
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState069
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState069
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState069
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_DIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_113 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_067 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_DIVRM as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState067
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState067
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState067
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState067
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState067
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_DIVRM (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_114 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_065 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MOD as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState065
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState065
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState065
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState065
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState065
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MOD (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_125 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_063 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_MUL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState063
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState063
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState063
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState063
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState063
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MUL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_126 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_061 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_PLUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState061
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_PLUS (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_123 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_059 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RDIV as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState059
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState059
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState059
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState059
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState059
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_RDIV (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_128 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_057 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LEQ as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState057
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_expr_, expr) = ((), _endpos, _v) in
            let _v = _menhir_action_244 _1 _endpos_expr_ _startpos__1_ expr in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_053 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState053
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState053
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState053
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState053
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos__1_, _) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_179 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_041 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_NOT as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState041
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState041
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState041
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState041
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RBRACE | Tokens.RDIV | Tokens.RPAR | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_NOT (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_180 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_037 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SHL as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState037
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState037
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState037
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState037
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState037
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_SHL (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_129 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_035 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_COLON as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState035
        | Tokens.COMMA | Tokens.RBRACKET ->
            let MenhirCell1_COLON (_menhir_stack, _menhir_s) = _menhir_stack in
            let (_1, e) = ((), _v) in
            let _v = _menhir_action_261 _1 e in
            _menhir_goto_slice _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_032 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_POW as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState032
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState032
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState032
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState032
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_POW (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_131 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_030 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_SHR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState030
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState030
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState030
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState030
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState030
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_SHR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_130 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_028 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_XOR as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState028
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_XOR (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_1, _endpos_e2_, e2) = ((), _endpos, _v) in
            let _v = _menhir_action_115 _1 _endpos_e2_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e2_ _startpos_e1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_026 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_LLBRACKET as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_027 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_029 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_036 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.RRBRACKET ->
            let _endpos_0 = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_LLBRACKET (_menhir_stack, _) = _menhir_stack in
            let MenhirCell1_expr (_menhir_stack, _menhir_s, e1, _startpos_e1_, _) = _menhir_stack in
            let (_2, _endpos__4_, _4, e2) = ((), _endpos_0, (), _v) in
            let _v = _menhir_action_139 _2 _4 _endpos__4_ _startpos_e1_ e1 e2 in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__4_ _startpos_e1_ _v _menhir_s _tok
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_058 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_031 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_060 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_070 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_072 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_062 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_064 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_075 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_077 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_083 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_089 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_085 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_087 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_091 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_066 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_068 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_079 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_094 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_096 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_098 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_100 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_081 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState026
        | _ ->
            _eRR ()
    
    and _menhir_run_024 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_MINUS as 'stack) -> _ -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_025 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState024
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState024
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_038 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState024
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_042 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState024
        | Tokens.AND | Tokens.ARROW | Tokens.AS | Tokens.BAND | Tokens.BEGIN | Tokens.BEQ | Tokens.BOR | Tokens.COLON | Tokens.COLON_COLON | Tokens.COMMA | Tokens.DIV | Tokens.DIVRM | Tokens.DO | Tokens.DOWNTO | Tokens.ELSE | Tokens.ELSIF | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.LEQ | Tokens.LOOPLIMIT | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OF | Tokens.OR | Tokens.PLUS | Tokens.PLUS_COLON | Tokens.POW | Tokens.RBRACE | Tokens.RBRACKET | Tokens.RDIV | Tokens.RPAR | Tokens.RRBRACKET | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.SLICING | Tokens.STAR_COLON | Tokens.THEN | Tokens.TO | Tokens.WHERE | Tokens.XOR ->
            let MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos__1_, _) = _menhir_stack in
            let (_1, _endpos_e_, e) = ((), _endpos, _v) in
            let _v = _menhir_action_134 _1 _endpos_e_ _startpos__1_ e in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_e_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist0_ty_ : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (x, _endpos__3_) = (_v, _endpos) in
        let _v = _menhir_action_327 _endpos__3_ _startpos__1_ x in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _v _menhir_s _tok
    
    and _menhir_run_328 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_decl_item (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState328) in
            let _menhir_s = MenhirState329 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.COLON ->
            _menhir_run_113 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState328
        | Tokens.SEMI_COLON ->
            let _v_6 = _menhir_action_235 () in
            _menhir_run_331 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | _ ->
            _eRR ()
    
    and _menhir_run_318 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState319 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_315 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_037 x in
            _menhir_goto_clist1_discard_or_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_discard_or_identifier_ : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState317 ->
            _menhir_run_321 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState319 ->
            _menhir_run_320 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_321 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR, ttv_result) _menhir_cell1_discard_or_identifier -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_identifier (_menhir_stack, _, x) = _menhir_stack in
        let (_2, li) = ((), _v) in
        let _v = _menhir_action_059 _2 li x in
        let _tok = _menhir_lexer _menhir_lexbuf in
        let MenhirCell1_LPAR (_menhir_stack, _menhir_s, _) = _menhir_stack in
        let x = _v in
        let _v = _menhir_action_093 x in
        _menhir_goto_decl_item _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_320 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_discard_or_identifier, ttv_result) _menhir_cell1_discard_or_identifier -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_038 _2 h t in
        _menhir_goto_clist1_discard_or_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_316 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_LPAR as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_discard_or_identifier (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_s = MenhirState317 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.MINUS ->
                _menhir_run_313 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_315 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_661 : type  ttv_stack. ((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ty as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_GETTER (_menhir_stack, _menhir_s) in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState662
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState662
            | Tokens.END ->
                let _v =
                  let _endpos__0_ = _endpos in
                  _menhir_action_225 _endpos__0_
                in
                _menhir_run_663 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState662 _tok
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_703 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _menhir_s = MenhirState704 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SETTER ->
                _menhir_run_644 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.GETTER ->
                _menhir_run_661 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_642 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.BEGIN ->
            let _menhir_s = MenhirState643 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.SETTER ->
                _menhir_run_644 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.GETTER ->
                _menhir_run_661 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_628 : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_ARROW -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_ARROW (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _2) = ((), _v) in
        let _v = _menhir_action_254 _1 _2 in
        _menhir_goto_return_type _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_return_type : type  ttv_stack. (((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState711 ->
            _menhir_run_716 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState687 ->
            _menhir_run_692 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState621 ->
            _menhir_run_629 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_716 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_return_type (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RECURSELIMIT ->
            _menhir_run_630 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState716
        | Tokens.BEGIN ->
            let _v_0 = _menhir_action_252 () in
            _menhir_run_717 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState716
        | _ ->
            _eRR ()
    
    and _menhir_run_630 : type  ttv_stack. ((((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args, _menhir_box_spec) _menhir_cell1_return_type as 'stack) -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_RECURSELIMIT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState630 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_692 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPDEF _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_return_type (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RECURSELIMIT ->
            _menhir_run_630 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState692
        | Tokens.BEGIN ->
            let _v_0 = _menhir_action_252 () in
            _menhir_run_693 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState692
        | _ ->
            _eRR ()
    
    and _menhir_run_629 : type  ttv_stack. ((((ttv_stack, _menhir_box_spec) _menhir_cell1_IMPLEMENTATION _menhir_cell0_FUNC _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_func_args as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_return_type (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.RECURSELIMIT ->
            _menhir_run_630 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState629
        | Tokens.BEGIN ->
            let _v_0 = _menhir_action_252 () in
            _menhir_run_632 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState629
        | _ ->
            _eRR ()
    
    and _menhir_run_594 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_TYPE _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_OF as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _1 = _v in
        let _v = _menhir_action_330 _1 in
        _menhir_goto_ty_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_528 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_WHEN, ttv_result) _menhir_cell1_IDENTIFIER as 'stack) -> _ -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
        match (_tok : MenhirBasics.token) with
        | Tokens.ARROW ->
            let _menhir_s = MenhirState529 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.WHILE ->
                _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.VAR ->
                _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.UNREACHABLE ->
                _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.TRY ->
                _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.THROW ->
                _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.RETURN ->
                _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REPEAT ->
                _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINTLN ->
                _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRINT ->
                _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PRAGMA ->
                _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.PASS ->
                _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LET ->
                _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IF ->
                _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.FOR ->
                _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.DEBUG ->
                _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CONSTANT ->
                _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.CASE ->
                _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ASSERT ->
                _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_292 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ARBITRARY -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _endpos_ty_, ty) = ((), (), _endpos, _v) in
        let _v = _menhir_action_147 _1 _2 _endpos_ty_ _startpos__1_ ty in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_286 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_ty (_menhir_stack, _menhir_s, _v, _endpos) in
            let _menhir_s = MenhirState287 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING ->
                _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.REAL ->
                _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INTEGER ->
                _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOLEAN ->
                _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITS ->
                _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BIT ->
                _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.ARRAY ->
                _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_053 x in
            _menhir_goto_clist1_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_ty_ : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState018 ->
            _menhir_run_289 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState287 ->
            _menhir_run_288 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_289 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_LPAR -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_030 _1 in
        _menhir_goto_clist0_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_288 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ty -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_ty (_menhir_stack, _menhir_s, h, _) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_054 _2 h t in
        _menhir_goto_clist1_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_244 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
        let (_2, _endpos_ty_, ty) = ((), _endpos, _v) in
        let _v = _menhir_action_187 _2 _endpos_ty_ _startpos_expr_pattern_ expr_pattern ty in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos_expr_pattern_ _v _menhir_s _tok
    
    and _menhir_run_174 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_ARBITRARY -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_ARBITRARY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _endpos_ty_, ty) = ((), (), _endpos, _v) in
        let _v = _menhir_action_192 _1 _2 _endpos_ty_ _startpos__1_ ty in
        _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos__1_ _v _menhir_s _tok
    
    and _menhir_run_129 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_AS -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_AS (_menhir_stack, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _menhir_s, expr, _startpos_expr_, _) = _menhir_stack in
        let (_2, _endpos_ty_, ty) = ((), _endpos, _v) in
        let _v = _menhir_action_142 _2 _endpos_ty_ _startpos_expr_ expr ty in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_ty_ _startpos_expr_ _v _menhir_s _tok
    
    and _menhir_run_121 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_COLON -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_COLON (_menhir_stack, _menhir_s) = _menhir_stack in
        let (_1, _2) = ((), _v) in
        let _v = _menhir_action_011 _1 _2 in
        _menhir_goto_as_ty _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_as_ty : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState743 ->
            _menhir_run_748 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState724 ->
            _menhir_run_725 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState606 ->
            _menhir_run_607 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState566 ->
            _menhir_run_567 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState553 ->
            _menhir_run_558 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState460 ->
            _menhir_run_464 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState351 ->
            _menhir_run_352 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | MenhirState328 ->
            _menhir_run_333 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState112 ->
            _menhir_run_123 _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_748 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_global_let_or_constant, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let x = _v in
        let _v = _menhir_action_232 x in
        _menhir_goto_option_as_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_option_as_ty_ : type  ttv_stack. ((ttv_stack, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState743 ->
            _menhir_run_744 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState553 ->
            _menhir_run_554 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_725 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_CONFIG, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_as_ty (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState726 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_607 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let x_inlined1 = _v in
        let _v = _menhir_action_230 x x_inlined1 in
        _menhir_goto_opt_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_567 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_IDENTIFIER -> _ -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, x, _, _) = _menhir_stack in
        let y = _v in
        let _v = _menhir_action_335 x y in
        _menhir_goto_typed_identifier _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_goto_typed_identifier : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match _menhir_s with
        | MenhirState616 ->
            _menhir_run_615 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState614 ->
            _menhir_run_615 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState569 ->
            _menhir_run_568 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | MenhirState565 ->
            _menhir_run_568 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_615 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _menhir_stack = MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, _v) in
            let _menhir_s = MenhirState616 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_566 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RPAR ->
            let x = _v in
            let _v = _menhir_action_055 x in
            _menhir_goto_clist1_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_clist1_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState614 ->
            _menhir_run_618 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState616 ->
            _menhir_run_617 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_618 : type  ttv_stack. ((ttv_stack _menhir_cell0_IDENTIFIER, _menhir_box_spec) _menhir_cell1_params_opt, _menhir_box_spec) _menhir_cell1_LPAR -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_032 _1 in
        _menhir_goto_clist0_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_617 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_056 _2 h t in
        _menhir_goto_clist1_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_568 : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.COMMA ->
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v_0 ->
                let _menhir_stack = MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, _v) in
                _menhir_run_566 _menhir_stack _menhir_lexbuf _menhir_lexer _v_0 MenhirState569
            | Tokens.RBRACE ->
                let (x, x_inlined1) = (_v, ()) in
                let _v = _menhir_action_319 x x_inlined1 in
                _menhir_goto_tclist1_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.RBRACE ->
            let x = _v in
            let _v = _menhir_action_318 x in
            _menhir_goto_tclist1_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_goto_tclist1_typed_identifier_ : type  ttv_stack. ttv_stack -> _ -> _ -> _ -> (ttv_stack, _menhir_box_spec) _menhir_state -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        match _menhir_s with
        | MenhirState565 ->
            _menhir_run_571 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | MenhirState569 ->
            _menhir_run_570 _menhir_stack _menhir_lexbuf _menhir_lexer _v
        | _ ->
            _menhir_fail ()
    
    and _menhir_run_571 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_LBRACE -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let _1 = _v in
        let _v = _menhir_action_311 _1 in
        _menhir_goto_tclist0_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v
    
    and _menhir_run_570 : type  ttv_stack. (ttv_stack, _menhir_box_spec) _menhir_cell1_typed_identifier -> _ -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v ->
        let MenhirCell1_typed_identifier (_menhir_stack, _menhir_s, h) = _menhir_stack in
        let (_2, t) = ((), _v) in
        let _v = _menhir_action_320 _2 h t in
        _menhir_goto_tclist1_typed_identifier_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
    
    and _menhir_run_558 : type  ttv_stack. (((ttv_stack, _menhir_box_spec) _menhir_cell1_VAR, _menhir_box_spec) _menhir_cell1_ignored_or_identifier as 'stack) -> _ -> _ -> _ -> ('stack, _menhir_box_spec) _menhir_state -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_ignored_or_identifier (_menhir_stack, _, name) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_084 _1 _endpos__2_ _startpos__1_ name x in
            _menhir_goto_decl _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | Tokens.EQ ->
            let x = _v in
            let _v = _menhir_action_232 x in
            _menhir_goto_option_as_ty_ _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_464 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_local_decl_keyword_non_var, ttv_result) _menhir_cell1_decl_item as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_as_ty (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_s = MenhirState465 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_335 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | _ ->
            _eRR ()
    
    and _menhir_run_335 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_010 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState335
        | Tokens.LBRACE ->
            let _startpos_0 = _menhir_lexbuf.Lexing.lex_start_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v_1 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_1 MenhirState336
            | Tokens.REAL_LIT _v_2 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_2 MenhirState336
            | Tokens.RBRACE ->
                let _endpos_3 = _menhir_lexbuf.Lexing.lex_curr_p in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.LPAR ->
                    let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                    let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                    let _menhir_stack = MenhirCell1_RBRACE (_menhir_stack, MenhirState336, _endpos_3) in
                    let _startpos_4 = _menhir_lexbuf.Lexing.lex_start_p in
                    let _menhir_stack = MenhirCell0_LPAR (_menhir_stack, _startpos_4) in
                    let _tok = _menhir_lexer _menhir_lexbuf in
                    (match (_tok : MenhirBasics.token) with
                    | Tokens.STRING_LIT _v_5 ->
                        _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v_5 MenhirState338
                    | Tokens.REAL_LIT _v_6 ->
                        _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6 MenhirState338
                    | Tokens.NOT ->
                        _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.MINUS ->
                        _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.LPAR ->
                        _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.INT_LIT _v_7 ->
                        _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_7 MenhirState338
                    | Tokens.IF ->
                        _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.IDENTIFIER _v_8 ->
                        _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v_8 MenhirState338
                    | Tokens.BOOL_LIT _v_9 ->
                        _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_9 MenhirState338
                    | Tokens.BNOT ->
                        _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.BITVECTOR_LIT _v_10 ->
                        _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_10 MenhirState338
                    | Tokens.ARBITRARY ->
                        _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState338
                    | Tokens.RPAR ->
                        let _v_11 = _menhir_action_027 () in
                        _menhir_run_339 _menhir_stack _menhir_lexbuf _menhir_lexer _v_11 _tok
                    | _ ->
                        _eRR ())
                | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RDIV | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.XOR ->
                    let (_2, _endpos__3_, _3, _endpos_desc_, _startpos_desc_, desc) = ((), _endpos_3, (), _endpos, _startpos, _v) in
                    let _v = _menhir_action_149 _2 _3 _endpos__3_ _endpos_desc_ _startpos_desc_ desc in
                    _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__3_ _startpos_desc_ _v _menhir_s _tok
                | _ ->
                    _eRR ())
            | Tokens.NOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | Tokens.MINUS ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | Tokens.LPAR ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | Tokens.INT_LIT _v_12 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v_12 MenhirState336
            | Tokens.IF ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | Tokens.IDENTIFIER _v_13 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_146 _menhir_stack _menhir_lexbuf _menhir_lexer _v_13 MenhirState336
            | Tokens.COMMA ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                let _menhir_stack = MenhirCell1_COMMA (_menhir_stack, MenhirState336) in
                let _menhir_s = MenhirState341 in
                let _tok = _menhir_lexer _menhir_lexbuf in
                (match (_tok : MenhirBasics.token) with
                | Tokens.STRING_LIT _v ->
                    _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.REAL_LIT _v ->
                    _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.NOT ->
                    _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.MINUS ->
                    _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.LPAR ->
                    _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.INT_LIT _v ->
                    _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.IF ->
                    _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.IDENTIFIER _v ->
                    _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BOOL_LIT _v ->
                    _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.BNOT ->
                    _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | Tokens.BITVECTOR_LIT _v ->
                    _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
                | Tokens.ARBITRARY ->
                    _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
                | _ ->
                    _eRR ())
            | Tokens.BOOL_LIT _v_20 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v_20 MenhirState336
            | Tokens.BNOT ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | Tokens.BITVECTOR_LIT _v_21 ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v_21 MenhirState336
            | Tokens.ARBITRARY ->
                let _menhir_stack = MenhirCell1_IDENTIFIER (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
                let _menhir_stack = MenhirCell1_LBRACE (_menhir_stack, MenhirState335, _startpos_0) in
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState336
            | _ ->
                _eRR ())
        | Tokens.AND | Tokens.AS | Tokens.BAND | Tokens.BEQ | Tokens.BOR | Tokens.COLON_COLON | Tokens.DIV | Tokens.DIVRM | Tokens.DOT | Tokens.EQ_OP | Tokens.GEQ | Tokens.GT | Tokens.IMPL | Tokens.IN | Tokens.LBRACKET | Tokens.LEQ | Tokens.LLBRACKET | Tokens.LT | Tokens.MINUS | Tokens.MOD | Tokens.MUL | Tokens.NEQ | Tokens.OR | Tokens.PLUS | Tokens.POW | Tokens.RDIV | Tokens.SEMI_COLON | Tokens.SHL | Tokens.SHR | Tokens.XOR ->
            let (_endpos__1_, _startpos__1_, _1) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_108 _1 _endpos__1_ _startpos__1_ in
            _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos__1_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_352 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_clist2_IDENTIFIER_ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.SEMI_COLON ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_clist2_IDENTIFIER_ (_menhir_stack, _, _2_inlined1) = _menhir_stack in
            let MenhirCell1_VAR (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
            let (_1, _endpos__2_, as_ty) = ((), _endpos, _v) in
            let _v = _menhir_action_292 _1 _2_inlined1 _endpos__2_ _startpos__1_ as_ty in
            _menhir_goto_stmt _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__2_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_333 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_VAR, ttv_result) _menhir_cell1_decl_item as 'stack) -> _ -> _ -> _ -> ('stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok ->
        let _menhir_stack = MenhirCell1_as_ty (_menhir_stack, _menhir_s, _v) in
        match (_tok : MenhirBasics.token) with
        | Tokens.EQ ->
            let _menhir_stack = MenhirCell1_EQ (_menhir_stack, MenhirState333) in
            let _menhir_s = MenhirState334 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.STRING_LIT _v ->
                _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.REAL_LIT _v ->
                _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.NOT ->
                _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.MINUS ->
                _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.LPAR ->
                _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.INT_LIT _v ->
                _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.IF ->
                _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.IDENTIFIER _v ->
                _menhir_run_335 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BOOL_LIT _v ->
                _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.BNOT ->
                _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | Tokens.BITVECTOR_LIT _v ->
                _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | Tokens.ARBITRARY ->
                _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
            | _ ->
                _eRR ())
        | Tokens.SEMI_COLON ->
            let _v_6 = _menhir_action_235 () in
            _menhir_run_349 _menhir_stack _menhir_lexbuf _menhir_lexer _v_6
        | _ ->
            _eRR ()
    
    and _menhir_run_123 : type  ttv_stack ttv_result. (ttv_stack, ttv_result) _menhir_cell1_slices _menhir_cell0_IDENTIFIER -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _v _tok ->
        let MenhirCell0_IDENTIFIER (_menhir_stack, x, _, _) = _menhir_stack in
        let MenhirCell1_slices (_menhir_stack, _menhir_s, s, _, _) = _menhir_stack in
        let ty = _v in
        let _v = _menhir_action_016 s ty x in
        _menhir_goto_bitfield _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s _tok
    
    and _menhir_run_120 : type  ttv_stack ttv_result. (((ttv_stack, ttv_result) _menhir_cell1_ARRAY, ttv_result) _menhir_cell1_expr, ttv_result) _menhir_cell1_RRBRACKET -> _ -> _ -> _ -> _ -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _v _tok ->
        let MenhirCell1_RRBRACKET (_menhir_stack, _, _) = _menhir_stack in
        let MenhirCell1_expr (_menhir_stack, _, e, _, _) = _menhir_stack in
        let MenhirCell1_ARRAY (_menhir_stack, _menhir_s, _startpos__1_) = _menhir_stack in
        let (_1, _2, _4, _endpos_t_, t, _5) = ((), (), (), _endpos, _v, ()) in
        let _v = _menhir_action_329 _1 _2 _4 _5 _endpos_t_ _startpos__1_ e t in
        _menhir_goto_ty _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_t_ _v _menhir_s _tok
    
    and _menhir_run_184 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SLICING (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState184 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_186 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState186 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_188 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_SHL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState188 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_190 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_RDIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState190 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_192 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_POW (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState192 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_194 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_PLUS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState194 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_196 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_OR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState196 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_198 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_191 _1 _2 _endpos__1_ _startpos__1_ _startpos_e_ e in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_e_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState198
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_NEQ (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState198
        | _ ->
            _eRR ()
    
    and _menhir_run_201 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MUL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState201 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_203 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_MOD (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState203 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_205 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
        let _menhir_stack = MenhirCell1_MINUS (_menhir_stack, _menhir_s, _startpos, _endpos) in
        let _menhir_s = MenhirState205 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_207 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState207 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_209 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_LLBRACKET (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState209 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_212 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_LEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState212 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_214 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IN (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState214 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACE ->
            _menhir_run_039 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BNOT ->
            _menhir_run_215 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_217 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_IMPL (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState217 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_219 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_GT (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState219 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_221 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
        let _menhir_stack = MenhirCell1_GEQ (_menhir_stack, _menhir_s, _startpos) in
        let _menhir_s = MenhirState221 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_223 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.REAL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.NOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | Tokens.MASK_LIT _v ->
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, expr_pattern, _startpos_expr_pattern_, _) = _menhir_stack in
            let (_endpos__1_, _startpos__1_, _1, _2) = (_endpos, _startpos, _v, ()) in
            let _v = _menhir_action_190 _1 _2 _endpos__1_ _startpos__1_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos__1_ _startpos_expr_pattern_ _v _menhir_s _tok
        | Tokens.LPAR ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | Tokens.INT_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.IF ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | Tokens.IDENTIFIER _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.BOOL_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.BNOT ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | Tokens.BITVECTOR_LIT _v ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v MenhirState223
        | Tokens.ARBITRARY ->
            let _menhir_stack = MenhirCell1_EQ_OP (_menhir_stack, _menhir_s) in
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState223
        | _ ->
            _eRR ()
    
    and _menhir_run_226 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_DOT (_menhir_stack, _menhir_s) in
            let _startpos = _menhir_lexbuf.Lexing.lex_start_p in
            let _menhir_stack = MenhirCell0_LBRACKET (_menhir_stack, _startpos) in
            let _menhir_s = MenhirState227 in
            let _tok = _menhir_lexer _menhir_lexbuf in
            (match (_tok : MenhirBasics.token) with
            | Tokens.IDENTIFIER _v ->
                _menhir_run_044 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
            | _ ->
                _eRR ())
        | Tokens.IDENTIFIER _v ->
            let _endpos = _menhir_lexbuf.Lexing.lex_curr_p in
            let _tok = _menhir_lexer _menhir_lexbuf in
            let MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, e, _startpos_e_, _) = _menhir_stack in
            let (_2, _endpos_x_, x) = ((), _endpos, _v) in
            let _v = _menhir_action_185 _2 _endpos_x_ _startpos_e_ e x in
            _menhir_goto_expr_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_x_ _startpos_e_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_231 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIVRM (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState231 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_233 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_DIV (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState233 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_235 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_COLON_COLON (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState235 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_237 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BOR (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState237 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_239 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BEQ (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState239 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_241 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_BAND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState241 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_243 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AS (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState243 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING ->
            _menhir_run_016 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REAL ->
            _menhir_run_017 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_018 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LBRACE ->
            _menhir_run_020 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INTEGER ->
            _menhir_run_019 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_101 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOLEAN ->
            _menhir_run_102 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITS ->
            _menhir_run_103 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BIT ->
            _menhir_run_114 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ARRAY ->
            _menhir_run_115 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_246 : type  ttv_stack ttv_result. ((ttv_stack, ttv_result) _menhir_cell1_expr_pattern as 'stack) -> _ -> _ -> ('stack, ttv_result) _menhir_state -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s ->
        let _menhir_stack = MenhirCell1_AND (_menhir_stack, _menhir_s) in
        let _menhir_s = MenhirState246 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.STRING_LIT _v ->
            _menhir_run_002 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.REAL_LIT _v ->
            _menhir_run_003 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.NOT ->
            _menhir_run_004 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_005 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_006 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.INT_LIT _v ->
            _menhir_run_007 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.IF ->
            _menhir_run_008 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_009 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BOOL_LIT _v ->
            _menhir_run_011 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.BNOT ->
            _menhir_run_012 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.BITVECTOR_LIT _v ->
            _menhir_run_013 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.ARBITRARY ->
            _menhir_run_014 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    and _menhir_run_181 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        match (_tok : MenhirBasics.token) with
        | Tokens.XOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_182 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.SLICING ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_184 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.SHR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_186 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.SHL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_188 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.RDIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_190 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.POW ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_192 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.PLUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_194 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.OR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_196 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.NEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_198 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.MUL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_201 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.MOD ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_203 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.MINUS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_205 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.LT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_207 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.LLBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_209 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.LEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_212 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.LBRACKET ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_033 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.IN ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_214 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.IMPL ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_217 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.GT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_219 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.GEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_221 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.EQ_OP ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_223 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.DOT ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_226 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.DIVRM ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_231 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.DIV ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_233 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.COLON_COLON ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_235 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.BOR ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_237 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.BEQ ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_239 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.BAND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_241 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.AS ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_243 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.AND ->
            let _menhir_stack = MenhirCell1_expr_pattern (_menhir_stack, _menhir_s, _v, _startpos, _endpos) in
            _menhir_run_246 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState181
        | Tokens.ARROW | Tokens.COMMA | Tokens.RBRACE | Tokens.RPAR | Tokens.WHERE ->
            let (_endpos_expr_pattern_, _startpos_expr_pattern_, expr_pattern) = (_endpos, _startpos, _v) in
            let _v = _menhir_action_241 _endpos_expr_pattern_ _startpos_expr_pattern_ expr_pattern in
            _menhir_goto_pattern _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_expr_pattern_ _startpos_expr_pattern_ _v _menhir_s _tok
        | _ ->
            _eRR ()
    
    and _menhir_run_023 : type  ttv_stack ttv_result. ttv_stack -> _ -> _ -> _ -> _ -> _ -> (ttv_stack, ttv_result) _menhir_state -> _ -> ttv_result =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer _endpos _startpos _v _menhir_s _tok ->
        let (_endpos_value_, _startpos_value_, value) = (_endpos, _startpos, _v) in
        let _v = _menhir_action_107 _endpos_value_ _startpos_value_ value in
        _menhir_goto_expr _menhir_stack _menhir_lexbuf _menhir_lexer _endpos_value_ _startpos_value_ _v _menhir_s _tok
    
    let _menhir_run_000 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_opn =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState000 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
    let _menhir_run_549 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_spec =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.VAR ->
            _menhir_run_550 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.TYPE ->
            _menhir_run_560 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.PRAGMA ->
            _menhir_run_597 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.LET ->
            _menhir_run_601 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.IMPLEMENTATION ->
            _menhir_run_602 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.IMPDEF ->
            _menhir_run_683 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.FUNC ->
            _menhir_run_708 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.CONSTANT ->
            _menhir_run_722 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.CONFIG ->
            _menhir_run_723 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.ACCESSOR ->
            _menhir_run_729 _menhir_stack _menhir_lexbuf _menhir_lexer MenhirState549
        | Tokens.EOF ->
            let _v = _menhir_action_212 () in
            _menhir_run_740 _menhir_stack _v
        | _ ->
            _eRR ()
    
    let _menhir_run_751 : type  ttv_stack. ttv_stack -> _ -> _ -> _menhir_box_stmts =
      fun _menhir_stack _menhir_lexbuf _menhir_lexer ->
        let _menhir_s = MenhirState751 in
        let _tok = _menhir_lexer _menhir_lexbuf in
        match (_tok : MenhirBasics.token) with
        | Tokens.WHILE ->
            _menhir_run_001 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.VAR ->
            _menhir_run_312 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.UNREACHABLE ->
            _menhir_run_354 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.TRY ->
            _menhir_run_358 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.THROW ->
            _menhir_run_359 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.RETURN ->
            _menhir_run_363 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.REPEAT ->
            _menhir_run_367 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINTLN ->
            _menhir_run_368 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRINT ->
            _menhir_run_373 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PRAGMA ->
            _menhir_run_378 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.PASS ->
            _menhir_run_382 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.MINUS ->
            _menhir_run_384 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LPAR ->
            _menhir_run_385 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.LET ->
            _menhir_run_407 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IF ->
            _menhir_run_408 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.IDENTIFIER _v ->
            _menhir_run_411 _menhir_stack _menhir_lexbuf _menhir_lexer _v _menhir_s
        | Tokens.FOR ->
            _menhir_run_429 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.DEBUG ->
            _menhir_run_439 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CONSTANT ->
            _menhir_run_444 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.CASE ->
            _menhir_run_445 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | Tokens.ASSERT ->
            _menhir_run_453 _menhir_stack _menhir_lexbuf _menhir_lexer _menhir_s
        | _ ->
            _eRR ()
    
  end
  
  let stmts =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_stmts v = _menhir_run_751 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let spec =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_spec v = _menhir_run_549 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
  let opn =
    fun _menhir_lexer _menhir_lexbuf ->
      let _menhir_stack = () in
      let MenhirBox_opn v = _menhir_run_000 _menhir_stack _menhir_lexbuf _menhir_lexer in
      v
  
end
